/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Malloc interludes which allow for releasing of all used memory blocks.
 * Just call CmReset whenever you want to free all allocated memory.
 * This is convenient when reinitializing a program which allocates memory.
 *
 * Note: You must make sure that old stale pointers are not referenced,
 * such as static pointers which were NULL when the program started but
 * won't be later.  In other words, to be safe you should not assume any
 * variables have been initialized, even for zero values.
 */

#include <stdio.h>
#include <stdlib.h>

#include "cmalloc.h"


/*
 * Undo the definitions from the include file.
 */
#undef	malloc
#undef	realloc
#undef	calloc
#undef	free


#define	MAGIC	54928347L


typedef	struct chunk	CHUNK;

struct chunk
{
	CHUNK *	next;		/* next chunk in chain */
	CHUNK *	prev;		/* previous chunk in chain */
	long	magic;		/* magic value */
	char	data[4];	/* data storage (varying size) */
};


#define	CHUNK_SIZE(size) (sizeof(CHUNK) + (size) - 4)

#define	OFFSET(type, var) \
	(((char *) &((type *)(0))->var) - ((char *) 0))
	
#define	DPTOCP(dp) \
	((CHUNK *) (((char *)(dp)) - OFFSET(CHUNK, data[0])))


/*
 * The header of the list of chunks.
 */
static	CHUNK	head =
{
	&head,
	&head
};


/*
 * Allocate some memory.
 */
char *
CmAlloc(unsigned int size)
{
	CHUNK *	cp;

	cp = (CHUNK *) malloc((unsigned int) (CHUNK_SIZE(size)));

	if (cp == NULL)
		return NULL;

	cp->magic = MAGIC;
	cp->next = head.next;
	cp->next->prev = cp;
	cp->prev = &head;

	head.next = cp;

	return cp->data;
}


/*
 * Reallocate memory.
 */
char *
CmRealloc(
	char *		dp,	/* old data pointer */
	unsigned int	size	/* new size */
)
{
	CHUNK *	cp;

	if (dp == NULL)
		return CmAlloc(size);

	cp = DPTOCP(dp);

	if (cp->magic != MAGIC)
	{
		fprintf(stderr, "Bad realloc %p\n", (char *) dp);

		return NULL;
	}

	cp = (CHUNK *) realloc((char *) cp, (unsigned int) (CHUNK_SIZE(size)));

	if (cp == NULL)
		return NULL;

	cp->next->prev = cp;
	cp->prev->next = cp;

	return cp->data;
}


/*
 * Allocate zeroed memory.
 */
char *
CmCalloc(
	unsigned int	nelem,		/* number of elements */
	unsigned int	elsize		/* size of each element */
)
{
	CHUNK *		cp;
	unsigned long	totalsize;

	totalsize = CHUNK_SIZE(nelem * elsize);

	if (totalsize > ((unsigned int) -1))
		return NULL;

	cp = (CHUNK *) calloc(1, (unsigned int) totalsize);

	if (cp == NULL)
		return NULL;

	cp->magic = MAGIC;
	cp->next = head.next;
	cp->next->prev = cp;
	cp->prev = &head;

	head.next = cp;

	return cp->data;
}


/*
 * Free memory.
 */
void
CmFree(char * dp)
{
	CHUNK *	cp;

	if (dp == NULL)
		return;

	cp = DPTOCP(dp);

	if (cp->magic != MAGIC)
	{
		fprintf(stderr, "Bad free %p\n", dp);

		return;
	}

	cp->prev->next = cp->next;
	cp->next->prev = cp->prev;
	cp->next = NULL;
	cp->prev = NULL;
	cp->magic = 0;

	free((char *) cp);
}


/*
 * Reset the whole memory arena.
 * This frees all memory allocated since the last CmReset call.
 */
void
CmReset()
{
	CHUNK *	cp;

	while (head.next != &head)
	{
		cp = head.next;

		if (cp->magic != MAGIC)
		{
			fprintf(stderr, "Bad CmReset %p\n", (char *) cp);

			break;
		}

		head.next = cp->next;

		cp->next = NULL;
		cp->prev = NULL;
		cp->magic = 0;

		free((char *) cp);
	}

	head.next = &head;
	head.prev = &head;
}


/*
 * Actual calls to malloc, realloc, calloc, and free.
 * Users must be sure not to intermix these calls with the ones above.
 */
char *
r_malloc(unsigned int size)
{
	return (char *) malloc(size);
}


char *
r_realloc(
	char *		dp,
	unsigned int	size
)
{
	return (char *) realloc(dp, size);
}


char *
r_calloc(
	unsigned int	nelem,
	unsigned int	elsize
)
{
	return (char *) calloc(nelem, elsize);
}


void
r_free(char * dp)
{
	free(dp);
}

/* END CODE */
