/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Common definitions for talking to the fcomp module.
 * These are the variables which control its actions.
 */

#ifndef	FCOMP_H
#define	FCOMP_H


#include "bool.h"


#define	INFINITY	0x7fffffffL	/* maximum number */
#define	VERBOSE_NONE	0		/* least verbosity level */
#define	VERBOSE_DEFAULT	1		/* default verbosity level */
#define	VERBOSE_FULL	2		/* full verbosity level */


/*
 * Structure which traces an optimal D-path through the edit path
 * for the two files.  The linked list of these structures lists the
 * snakes which define the path, where a snake is a region of multiple
 * line matchings.
 */
typedef struct snake	SNAKE;

struct snake
{
	SNAKE *	next;		/* next snake in path */
	long	line1;		/* line number in file1 */
	long	line2;		/* line number in file2 */
	long	count;		/* number of matching lines */
};


/*
 * Data about a line.
 * Each line is placed on an appropriate hash chain.
 * For any two entries, the two hash values match if and only if
 * the two lines match, so that the data never needs to be examined
 * once the original comparisons are complete.  In the extremely rare
 * cases where the hash would incorrectly match, it is modified.
 */
typedef struct line	LINE;

struct line
{
	LINE *	next;		/* next line on same hash chain */
	long	hash;		/* hash value for line */
	char	data[4];	/* chars (variable length, MUST BE LAST) */
};

#define LINE_SIZE(chars) (sizeof(LINE) + (chars) - 2)


/*
 * Data about the files being compared.
 */
typedef	struct
{
	LINE **		lines;		/* pointer to data lines */
	long *		lnums;		/* line number of each saved line */
	long		lineCount;	/* number of saved lines in file */
	long		linesTotal;	/* total number of lines in file */
	const char *	name;		/* file name */
	FIO *		file;		/* file structure for input */
	char		tag;		/* tag character for file */
} FILEINFO;


/*
 * Global data for file comparisons.
 */
typedef struct
{
	/*
	 * Input arguments
	 */
	long		maxJoin;	/* maximum number of lines to join */
	long		maxChanges;	/* maximum changes to allow */
	long		maxLines;	/* maximum number of lines */
	long		context;	/* lines of context in output */
	long		fileAskip;	/* lines to skip for first file */
	long		fileBskip;	/* lines to skip for second file */
	long		fileAmodify;	/* lines to modify for first file */
	long		fileBmodify;	/* lines to modify for second file */
	long		verbosity;	/* verbosity level */
	BOOL		numFlag;	/* show line numbers */
	BOOL		spaceFlag;	/* ignore extra spaces in lines */
	BOOL		blankFlag;	/* ignore blank lines in files */
	BOOL		upcaseFlag;	/* ignore case differences in lines */
	BOOL		quickFlag;	/* only show quick summary */
	BOOL		whatFlag;	/* show both files showing changes */
	BOOL		editScriptFlag;	/* output an edit script */
	BOOL		matchFlag;	/* output lines which match */
	BOOL		debugFlag;	/* tell some debugging info */

	/*
	 * Output arguments
	 */
	long		inserts;	/* lines inserted into first file */
	long		deletes;	/* lines deleted from first file */
	long		matches;	/* lines unchanged between files */
	SNAKE *		snakeList;	/* snake list of matching lines */
	FILEINFO	fileA;		/* first file being compared */
	FILEINFO	fileB;		/* second file being compared */
} FCOMP_DATA;


extern FCOMP_DATA fc;


/*
 * Structure holding information about an edit for special character sequence
 * replacements during extracts.  This should really be in sc.h, but is here
 * so that the ModifyLine prototype is correct.
 */
typedef struct
{
	const char *	edit;		/* edit number string */
	const char *	date;		/* date string */
	const char *	user;		/* user string */
} INFO;


/*
 * Routines.
 */
extern	void		Fcomp(const char *, const char *);
extern	char *		AllocStr(unsigned long);
extern	const char *	ModifyLine(const char *, long *, const INFO *);
extern	void		DumpMatch(const char *);
extern	void		DumpNormal(const char *);
extern	void		DumpWhat(const char *);
extern	void		DumpLines(const char *);
extern	void		FcompReset(void);
extern	BOOL		IsDir(const char *);
extern	void		Error(const char *, ...);
extern	void		FileError(const FIO *, const char *, ...);

#endif

/* END CODE */
