/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * fcomp - a file comparison program
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>

#include "fio.h"
#include "fcomp.h"
#include "cmalloc.h"


static	const char *	Expand(const char *);
static	char *		DirFile(const char *, const char *);
static	void		Usage(void);


/*
 * This is the main program.
 * This performs all of the parsing of the command line, and then
 * calls a subroutine to do all the work.  This allows the subroutine
 * to be loaded with other programs and called by them easily.
 */
int
main(
	int		argc,
	const char *	argv[]
)
{
	const char *	cp;		/* current option character */
	const char *	nameA;		/* first file to be compared */
	const char *	nameB;		/* second file to be compared */
	const char *	outputFileName;	/* output fileName */
	long		value;		/* numeric value */
	int		opt;		/* option character */
	BOOL		dirA;		/* first file is a directory */
	BOOL		dirB;		/* second file is a directory */

	argv++;
	argc--;

	if (argc <= 0)
	{
		Usage();

		exit(1);
	}

	nameA = NULL;
	nameB = NULL;

	outputFileName = NULL;

	fc.verbosity = VERBOSE_DEFAULT;
	fc.maxChanges = INFINITY;

	while (argc-- > 0)
	{
		cp = *(argv++);

		if (*cp != '-')
		{
			/*
			 * A file name.
			 */
			cp = Expand(cp);

			if (nameA == NULL)
			{
				nameA = cp;
				continue;
			}

			if (nameB == NULL)
			{
				nameB = cp;
				continue;
			}

			Error("Only two input files allowed");
		}

		for (cp++; *cp; cp++)
		{
			/*
			 * An option
			 */
			opt = *cp;

			if ((opt >= 'A') && (opt <= 'Z'))
				opt += ('a' - 'A');

			switch (opt)
			{
				case 'b':	/* ignore blank lines */
					fc.blankFlag = TRUE;
					break;

				case 'd':	/* type debugging information */
					fc.debugFlag = TRUE;
					break;

				case 'e':	/* output an edit script */
					fc.editScriptFlag = TRUE;
					break;

				case 'h':	/* display help */
					Usage();
					exit(0);

				case 'm':	/* show lines which match */
					fc.matchFlag = TRUE;
					break;

				case 'n':	/* show line numbers */
					fc.numFlag = TRUE;
					break;

				case 'q':	/* display quick summary */
					fc.quickFlag = TRUE;
					break;

				case 's':	/* ignore extra spaces, tabs */
					fc.spaceFlag = TRUE;
					break;

				case 'u':	/* translate to upper case */
					fc.upcaseFlag = TRUE;
					break;

				case 'v':	/* be verbose */
					fc.verbosity = VERBOSE_FULL;
					break;

				case 'w':	/* show what changes are */
					fc.whatFlag = TRUE;
					break;

				case 'c':	/* set context value */
				case 'f':	/* set failure value */
				case 'j':	/* set join value */
					if (*++cp)
					{
						Error(
					"No options allowed after -%c", opt);
					}

					if ((argc <= 0) || (**argv == '-'))
					{
						Error(
					"Numeric value required for -%c", opt);
					}

					cp = *(argv++);

					argc--;

					value = 0;

					while ((*cp >= '0') && (*cp <= '9'))
					{
						value = (value * 10) +
							*cp++ - '0';
					}

					if (*cp || (value < 0))
					{
						Error(
					"Bad number specified for -%c", opt);
					}

					switch (opt)
					{
						case 'c':
							fc.context = value;
							break;

						case 'f':
							fc.maxChanges = value;
							break;

						case 'j':
							fc.maxJoin = value;
							break;
					}

					cp = "x";

					break;

				case 'o':	/* specify output file */
					if (cp[1])
					{
						Error(
					"No options allowed after -o");
					}

					if ((argc <= 0) || (**argv == '-'))
						Error("No output file given");

					outputFileName = Expand(*(argv++));

					argc--;

					cp = "x";

					break;

				default:
					Error("Unknown option '-%c'", opt);
			}
		}
	}

	if ((nameA == NULL) || (nameB == NULL))
		Error("Must have two input files");

	/*
	 * If just one argument is a directory, then modify it by using
	 * the other argument so as to reference a corresponding file
	 * within the directory.
	 */
	dirA = IsDir(nameA);
	dirB = IsDir(nameB);

	if (dirA && dirB)
	{
		fprintf(stderr,
			"Arguments \"%s\" and \"%s\" are both directories\n",
			nameA, nameB);

		exit(1);
	}

	if (dirA)
		nameA = DirFile(nameA, nameB);

	if (dirB)
		nameB = DirFile(nameB, nameA);

	Fcomp(nameA, nameB);

	if (fc.whatFlag)
		DumpWhat(outputFileName);
	else if (fc.matchFlag)
		DumpMatch(outputFileName);
	else if (fc.editScriptFlag)
		DumpLines(outputFileName);
	else
		DumpNormal(outputFileName);

	return 0;
}


/*
 * Expand a global variable name.
 * Returns original argument if string is not a global variable,
 * or returns copied value of global variable.
 */
static const char *
Expand(const char * cp)
{
	const char *	var;		/* variable value */
	char *		mem;

	if ((cp == NULL) || (*cp != '.'))
		return cp;

	var = getenv(cp);

	if ((var == NULL) || (*var == '\0'))
		return cp;

	mem = malloc(strlen(var) + 1);

	if (mem == NULL)
		Error("Malloc failed");

	strcpy(mem, var);

	return mem;
}


/*
 * Make a directory name into a file name by applying the last
 * component of the second name to it.  For example:
 *	DirFile("/etc", "/foo/bar")
 * results in "/etc/bar".
 */
static char *
DirFile(
	const char *	dirName,	/* name which is a directory */
	const char *	fileName	/* name which is a file name */
)
{
	const char *	endName;
	char *		cp;

	endName = strrchr(fileName, '/');

	if (endName == NULL)
		endName = fileName;

	cp = malloc(strlen(dirName) + strlen(endName) + 2);

	if (cp == NULL)
		Error("Cannot allocate file name");

	strcpy(cp, dirName);
	strcat(cp, "/");
	strcat(cp, endName);

	return cp;
}


/*
 * Determine if an argument is a directory.
 * Returns TRUE if so, or FALSE if non-existent or not.
 */
BOOL
IsDir(const char * pathName)
{
	struct stat	buf;

	if (stat(pathName, &buf) < 0)
		return FALSE;

	return ((buf.st_mode & S_IFMT) == S_IFDIR);
}


/*
 * Dummy routine to make fcomp happy.
 */
const char *
ModifyLine(
	const char *	cp,
	long *		linelen,
	const INFO *	info
)
{
	return cp;
}


/*
 * Error routine which can format the output and then exits.
 */
void
Error(
	const char *	fmt,
	...
)
{
	va_list	ap;
	char	buf[8192];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	fprintf(stderr, "%s\n", buf);

	exit(1);
}


/*
 * Error routine called for file errors.
 * This prints out the error message for a file error and exits.
 */
void
FileError(
	const FIO *	fp,
	const char *	fmt,
	...
)
{
	va_list	ap;
	char	buf[8192];

	if (fp && fp->errorCode)
		errno = fp->errorCode;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	perror(buf);

	exit(1);
}


/*
 * Help message
 */
static const char * const helpText[] =
{
	"",
	"Program to compare two files and output the differences.",
	"For convenience, one of the arguments can be a directory.",
	"",
	"usage: fcomp path1 path2 [-behmnqsuvw] [-c n] [-f n] [-j n] [-o file]",
	"   -b:  ignore blank lines",
	"   -c:  add n lines of context around changes",
	"   -e:  output an edit script which is machine-readable",
	"   -f:  fail if more than n changes are needed",
	"   -h:  display this message",
	"   -j:  join changes if separated by up to n matching lines",
	"   -m:  output matching lines instead of changed lines",
	"   -n:  show line numbers",
	"   -o:  output results to specified file",
	"   -q:  output only a quick summary of changes needed",
	"   -s:  ignore extra spaces or tabs in lines",
	"   -u:  uppercase lines before comparing",
	"   -v:  be verbose about what is going on",
	"   -w:  display all of both files showing what the changes are",
	NULL
};


/*
 * Type out the possible commands.
 */
static void
Usage()
{
	const char * const *	str;

	for (str = helpText; *str; str++)
		fprintf(stderr, "%s\n", *str);
}

/* END CODE */
