/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * fcomp - a file comparison program
 *
 * This program is based on the algorithm in:
 *	An O(ND) Difference Algorithm and Its Variations
 *	Eugene W. Myers
 *	(TR 85-6, April 10, 1985)
 *	Department of Computer Science
 *	The University of Arizona
 *	Tuscon, Arizona 85721
 *
 * Also see:
 *	A File Comparison Program
 *	Webb Miller and Eugene W. Myers
 *	Software Practice and Experience
 *	(Volume 15, No. 11, November 1985)
 *
 * Actual implementation and features by David I. Bell
 */


#include <stdio.h>
#include <string.h>

#include "fio.h"
#include "fcomp.h"
#include "cmalloc.h"


#define	SNAKE_ALLOC_SIZE	500	/* chunk size for snake allocation */
#define	CHAR_ALLOC_SIZE		10240	/* allocation size for file data */
#define	UNIQUE_ALLOC_SIZE	512	/* unique size for allocation */
#define	LINE_ALLOC_SIZE		2000	/* number of lines to allocate */
#define	DETAB_ALLOC_SIZE	128	/* buffer allocation size for detab */
#define	HASH_SIZE		2003	/* size of hash table */
#define	CHANGE_BLAB_INTERVAL	100	/* how often to talk about changes */
#define	READ_BLAB_INTERVAL	1000	/* how often to talk about reading */
#define	SNAKE_BLAB_INTERVAL	1000	/* how often to talk about snakes */
#define	TAB_SIZE		8	/* tab size */


#define	IsSpace(ch)	(((ch) == ' ') || ((ch) == '\t'))
#define	IsLower(ch)	(((ch) >= 'a') && ((ch) <= 'z'))
#define	ToUpper(ch)	((ch) + 'A' - 'a')


/*
 * Local data.
 */
static	long	tableSize;	/* needed table size */
static	long	snakeCount;	/* number of snake structures used */
static	int	availChars;	/* number of available characters */
static	char *	nextChar;	/* next allocable character storage */
static	long *	V1;		/* the row containing the last d */
static	long *	V2;		/* another row */
static	SNAKE *	nextSnake;	/* next allocable snake structure */
static	SNAKE *	lastSnake;	/* end of allocable snake structures */
static	SNAKE *	endSnake;	/* last snake structure in list */
static	LINE **	hashTable;	/* hash table for lines */

FCOMP_DATA	fc;		/* data visible to others */


/*
 * Local routines.
 */
static	SNAKE *	AllocSnake(void);
static	LINE *	AddLine(const char *);
static	int	Compare(const char *, const char *);
static	void	ReadFile(FILEINFO *, long, long);
static	void	MakeSnakes(void);
static	void	FindSnake(int, long, long, long, long);
static	void	PrintLines(FIO *, const FILEINFO *, long, long, const char *);
static	void	PrintRange(FIO *, const char *, long, long, long, long);
static	char *	Detab(const char *);
static	char *	XAlloc(unsigned long);

static	long	MidSnake(int, long, long, long, long, long *, long *,
			long *, long *);


/*
 * Routine to do the comparison of two files.
 * If either of the input filenames is NULL, then this indicates
 * that the file has already been read by an earlier call.
 */
void
Fcomp(
	const char *	nameA,
	const char *	nameB
)
{
	FIO	fileA;			/* file structure for first file */
	FIO	fileB;			/* file structure for second file */
	long	changes;		/* total number of changes */

	fc.fileA.file = &fileA;
	fc.fileB.file = &fileB;

	fc.fileA.tag = 'A';
	fc.fileB.tag = 'B';

	/*
	 * If file names have been supplied on this call, then save
	 * their names and open them.
	 */
	if (nameA)
	{
		fc.fileA.name = nameA;

		if (!FioOpen(fc.fileA.file, nameA, FIO_READ))
			FileError(fc.fileA.file, "Cannot read \"%s\"", nameA);
	}

	if (nameB)
	{
		fc.fileB.name = nameB;

		if (!FioOpen(fc.fileB.file, nameB, FIO_READ))
			FileError(fc.fileB.file, "Cannot read \"%s\"", nameB);
	}

	/*
	 * Make sure that the two filenames had actually been supplied,
	 * either now or from a previous call.
	 */
	if ((fc.fileA.name == NULL) || (fc.fileB.name == NULL))
		Error("Input file names not supplied for fcomp");

	if (hashTable == NULL)
	{
		hashTable = (LINE **) calloc(HASH_SIZE, sizeof(LINE *));

		if (hashTable == NULL)
		{
			FileError((FIO *) NULL,
				"Cannot allocate memory for hash table");
		}
	}

	/*
	 * Read the files if they are newly supplied.
	 */
	if (nameA)
		ReadFile(&fc.fileA, fc.fileAskip, fc.fileAmodify);

	if (nameB)
		ReadFile(&fc.fileB, fc.fileBskip, fc.fileBmodify);

	/*
	 * See if the maximum bound on changes is trivially exceeded.
	 */
	changes = fc.fileA.lineCount - fc.fileB.lineCount;

	if (changes < 0)
		changes = -changes;

	if (changes > fc.maxChanges)
	{
		Error("More than %ld change%s required",
			fc.maxChanges, ((fc.maxChanges == 1) ? "" : "s"));
	}

	/*
	 * OK, compare the files.
	 */
	MakeSnakes();
}


/*
 * Readfile - read in a file and remember the maximum number of lines it needs.
 * This routine squeezes out blank lines if specified.
 * The first specified number of lines is skipped first.
 * Then the specified number of lines is modified.
 */
static void
ReadFile(
	FILEINFO *	fi,		/* file information */
	long		skip,		/* lines to skip */
	long		modify		/* lines to modify */
)
{
	const char *	cp;		/* current line */
	FIO *		fp;		/* file pointer for file to read in */
	LINE **		lines;		/* pointer to lines array */
	long *		lnums;		/* line numbers within file */
	long		lineCount;	/* current number of lines */
	long		linesAvail;	/* available number of lines */
	long		lineNumber;	/* current line number */
	long		lineLen;	/* length of line read */

	if (fc.verbosity > VERBOSE_DEFAULT)
		fprintf(stderr, "[Reading \"%s\"]\n", fi->name);

	fp = fi->file;

	if (skip > 0)
		FioSkip(fp, skip);

	lineCount = 0;
	lineNumber = 0;

	linesAvail = LINE_ALLOC_SIZE;

	lines = (LINE **)
		XAlloc((unsigned long) (sizeof(LINE **) * LINE_ALLOC_SIZE));

	lnums = (long *)
		XAlloc((unsigned long) (sizeof(long) * LINE_ALLOC_SIZE));

	while (TRUE)
	{
		cp = FioReadline(fp, &lineLen, FALSE);

		if (cp == NULL)
			break;

		if (++lineNumber <= modify)
			cp = ModifyLine(cp, &lineLen, (INFO *) 0);

		if ((fc.verbosity > VERBOSE_DEFAULT) &&
			((lineNumber % READ_BLAB_INTERVAL) == 0))
		{
			fprintf(stderr, "[%ld lines]\n", lineNumber);
		}

		if (fc.blankFlag && (*cp == '\n'))
			continue;

		if (lineCount >= linesAvail)
		{
			linesAvail += LINE_ALLOC_SIZE;

			lines = (LINE **)
				realloc(lines, sizeof(LINE **) * linesAvail);

			lnums = (long *)
				realloc(lnums, sizeof(long) * linesAvail);

			if ((lines == NULL) || (lnums == NULL))
			{
				FileError((FIO *) NULL,
					"Realloc for file storage failed");
			}
		}

		lnums[lineCount] = lineNumber;
		lines[lineCount] = AddLine(cp);
		lineCount++;
	}

	fi->lnums = lnums;
	fi->lines = lines;
	fi->lineCount = lineCount;
	fi->linesTotal = lineNumber;

	if (lineCount > fc.maxLines)
		fc.maxLines = lineCount;

	if (!FioClose(fp))
		FileError(fp, "Read error on \"%s\"", fi->name);

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		if (lineCount == lineNumber)
			fprintf(stderr, "[Done, %ld lines]\n", lineNumber);
		else
		{
			fprintf(stderr, "[Done, %ld lines, %ld non-blanks]\n",
				lineNumber, lineCount);
		}
	}
}


/*
 * Subroutine to allocate a line structure of the appropriate size and
 * add it to the hash table.  If it matches an already existing line, then
 * the old structure will be returned.
 */
static LINE *
AddLine(const char * cp)
{
	LINE *		lp;		/* current line element */
	const char *	curcp;		/* current character of string */
	LINE **		hashEntry;	/* hash entry to use */
	long		hash;		/* hash value */
	long		len;		/* line length */
	int		ch;		/* current char */

	/*
	 * Compute line length and hash value.
	 * This has to take into account the space and uppercase flags.
	 */
	hash = 0;

	curcp = cp;

	if (fc.spaceFlag)
	{
		while (IsSpace(*curcp))
			curcp++;
	}

	for (; *curcp; curcp++)
	{
		ch = *curcp;

		if (fc.upcaseFlag && IsLower(ch))
			ch = ToUpper(ch);

		if (fc.spaceFlag && IsSpace(ch))
		{
			ch = ' ';

			while (IsSpace(curcp[1]))
				curcp++;

			if (curcp[1] == '\0')
				break;

			if (curcp[1] == '\n')
				ch = *++curcp;
		}

		hash += ((hash * 101) + ch);

		if (hash < 0)
			hash = (hash + 1) & INFINITY;
	}

	len = curcp - cp;

	/*
	 * Search proper hash chain for already existing line.
	 * If the hash accidentally matches, modify it and try again.
	 */
again:
	hashEntry = &hashTable[hash % HASH_SIZE];

	for (lp = *hashEntry; lp; lp = lp->next)
	{
		if (lp->hash != hash)
			continue;

		if (strcmp(cp, lp->data) == 0)
			return lp;

		if ((fc.spaceFlag || fc.upcaseFlag) &&
			(Compare(cp, lp->data) == 0))
		{
			continue;
		}

		hash = ((hash * 1234321) + 1) & INFINITY;

		goto again;
	}

	/*
	 * Line not found, allocate a new one
	 */
	lp = (LINE *) AllocStr(LINE_SIZE(len));

	strcpy(lp->data, cp);

	lp->hash = hash;
	lp->next = *hashEntry;

	*hashEntry = lp;

	return lp;
}


/*
 * Routine to generate the D-path list and compute the number
 * of insertions, deletions and matching lines found.
 */
static void
MakeSnakes()
{
	SNAKE *	sp;		/* current snake element */
	long	line1;		/* current line in file A */
	long	line2;		/* current line in file B */

	if (fc.verbosity > VERBOSE_DEFAULT)
		fprintf(stderr, "[Beginning comparisons]\n");

	tableSize = fc.maxLines * 2 + 1;

	V1 = (long *) XAlloc((unsigned long) (sizeof(long) * tableSize));
	V2 = (long *) XAlloc((unsigned long) (sizeof(long) * tableSize));

	V1 += fc.maxLines;
	V2 += fc.maxLines;

	if ((fc.fileA.lineCount > 0) && (fc.fileB.lineCount > 0))
	{
		FindSnake(0, 0L, fc.fileA.lineCount, 0L,
			fc.fileB.lineCount);
	}

	/*
	 * End the list with the lower right endpoint
	 */
	sp = AllocSnake();

	sp->line1 = fc.fileA.lineCount;
	sp->line2 = fc.fileB.lineCount;
	sp->count = 0;

	if (fc.verbosity > VERBOSE_DEFAULT)
		fprintf(stderr, "[Comparisons complete]\n");

	if (fc.debugFlag)
	{
		fprintf(stderr, " line1  line2  count\n");

		for (sp = fc.snakeList; sp; sp = sp->next)
		{
			fprintf(stderr, "%5ld %6ld %6ld\n",
				sp->line1, sp->line2, sp->count);
		}

		fprintf(stderr, "\n");
	}

	/*
	 * Scan the snake list and calculate the number of inserted,
	 * deleted, and matching lines.
	 */
	line1 = 0;
	line2 = 0;

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		fc.deletes += (sp->line1 - line1);
		fc.inserts += (sp->line2 - line2);
		fc.matches += sp->count;

		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}
}


/*
 * Recursive routine to find a minimal D-path through the edit graph
 * of the two input files.  Arguments are the beginning line numbers in
 * the files, and the number of lines to examine.  This is basically a
 * divide-and-conquer routine which finds the middle snake of an optimal
 * D-path, then calls itself to find the remainder of the path before the
 * snake and after the snake.
 */
static void
FindSnake(
	int	depth,
	long	A,
	long	N,
	long	B,
	long	M
)
{
	SNAKE *	sp;
	long	ulx;
	long	uly;
	long	lrx;
	long	lry;
	long	D;
	long	count;

	/*
	 * If more than one change needed, then call ourself for each part.
	 */
	D = MidSnake(depth, A, N, B, M, &ulx, &uly, &lrx, &lry);

	if ((fc.verbosity > VERBOSE_DEFAULT) && (depth == 0))
	{
		fprintf(stderr, "[%ld change%s total]\n[Finding edit path]\n",
			D, ((D == 1) ? "" : "s"));
	}

	if (D > 1)
	{
		if ((ulx > 0) && (uly > 0))
			FindSnake(depth + 1, A, ulx, B, uly);

		count = lrx - ulx;

		if (count > fc.maxJoin)
		{
			sp = AllocSnake();
			sp->line1 = A + ulx;
			sp->line2 = B + uly;
			sp->count = count;
		}

		N -= lrx;
		M -= lry;

		if ((N > 0) && (M > 0))
			FindSnake(depth + 1, A + lrx, N, B + lry, M);

		return;
	}

	/*
	 * Only 0 or 1 change needed, so we can compute the result directly.
	 * First compute the snake coming from the upper left corner if any.
	 */
	if (N > M)
		count = uly;
	else
		count = ulx;

	if (count > fc.maxJoin)
	{
		sp = AllocSnake();

		sp->line1 = A;
		sp->line2 = B;
		sp->count = count;
	}

	/*
	 * Finally compute the snake coming from the lower right corner if any.
	 */
	count = lrx - ulx;

	if (count > fc.maxJoin)
	{
		sp = AllocSnake();

		sp->line1 = A + ulx;
		sp->line2 = B + uly;
		sp->count = count;
	}
}


/*
 * Routine to find the middle snake of an optimial D-path spanning
 * lines A to A+N in file A to lines B to B+N in file B.  Returns the
 * length D of the D-path as a return value, and the upper left and
 * lower right relative coordinates of a snake midway through the D-path.
 */
static long
MidSnake(
	int	depth,
	long	A,
	long	N,
	long	B,
	long	M,
	long *	ulx,
	long *	uly,
	long *	lrx,
	long *	lry
)
{
	long	D;
	long	MAXD;
	long	DELTA;
	long	odd;
	long	x;
	long	y;
	long	k;
	long	changes;
	long	oldx;
	LINE **	lp1;
	LINE **	lp2;

	if (fc.debugFlag)
	{
		for (D = 0; D < depth; D++)
			fprintf(stderr, "  ");

		fprintf(stderr, "searching: %ld,%ld to %ld,%ld   midsnake: ",
			A, B, A + N, B + M);
	}

	DELTA = N - M;
	odd = DELTA & 1;
	MAXD = (M + N + 1) / 2;
	V1[1] = 0;
	V2[-1] = 0;
	changes = -odd - 2;

	/*
	 * This is the main loop for searching for the snake.
	 * D is the distance off the diagonals, and is the number
	 * of changes needed to get from the upper left to the
	 * lower right corner of the region.
	 */
	for (D = 0; D <= MAXD; D++)
	{
		changes += 2;

		if (changes > fc.maxChanges)
		{
			Error("More than %ld change%s required",
				fc.maxChanges,
				((fc.maxChanges == 1) ? "" : "s"));
		}

		if ((fc.verbosity > VERBOSE_DEFAULT) && (depth == 0) &&
			(changes > 1) &&
			((changes % CHANGE_BLAB_INTERVAL) <= 1))
		{
			fprintf(stderr, "[%ld changes]\n", changes - odd);
		}

		/*
		 * Examine all diagonals within current distance.
		 * First search from upper left to lower right,
		 * and then search from lower right to upper left.
		 */
		for (k = -D; k <= D; k += 2)
		{
			/*
			 * Find the end of the furthest forward D-path
			 * in diagonal k.
			 */
			if (k == -D || (k != D && (V1[k-1] < V1[k+1])))
				x = V1[k+1];
			else
				x = V1[k-1] + 1;

			y = x - k;

			lp1 = &fc.fileA.lines[A+x];
			lp2 = &fc.fileB.lines[B+y];

			oldx = x;

			while ((x < N) && (y < M) &&
				((*lp1)->hash == (*lp2)->hash))
			{
				x++;
				y++;
				lp1++;
				lp2++;
			}

			V1[k] = x;

			/*
			 * See if path overlaps furthest reverse D-path.
			 * If so, then we have found the snake.
			 */
			if (odd && (k >= (DELTA - (D-1))) &&
				(k <= (DELTA + (D-1))))
			{
				if ((x + V2[k-DELTA]) >= N)
				{
					*ulx = oldx;
					*uly = oldx - k;
					*lrx = x;
					*lry = y;

					if (fc.debugFlag)
					{
						fprintf(stderr,
					"%ld,%ld to %ld,%ld (odd)\n",
							*ulx, *uly, *lrx, *lry);
					}

					return changes;
				}
			}
		}

		for (k = -D; k <= D; k += 2)
		{
			/*
			 * Find the end of the furthest reaching reverse
			 * path in diagonal k+DELTA.
			 */
			if (k == D || (k != -D && (V2[k+1] < V2[k-1])))
				x = V2[k-1];
			else
				x = V2[k+1] + 1;

			y = x + k;

			lp1 = &fc.fileA.lines[A+N-x-1];
			lp2 = &fc.fileB.lines[B+M-y-1];

			oldx = x;

			while ((x < N) && (y < M) &&
				((*lp1)->hash == (*lp2)->hash))
			{
				x++;
				y++;
				lp1--;
				lp2--;
			}

			V2[k] = x;

			/*
			 * See if path overlaps furthest forward D-path.
			 * If so, then we have found the snake.
			 */
			if (!odd && (k <= D-DELTA) && (k >= -D-DELTA))
			{
				if ((x + V1[k+DELTA]) >= N)
				{
					*ulx = N - x;
					*uly = M - y;
					*lrx = N - oldx;
					*lry = *lrx + *uly - *ulx;

					if (fc.debugFlag)
					{
						fprintf(stderr,
					"%ld,%ld to %ld,%ld (even)\n",
						*ulx, *uly, *lrx, *lry);
					}

					return changes;
				}
			}
		}
	}

	Error("Middle snake procedure failed!!");

	return 0;
}


/*
 * Compare two lines of the files.
 * The upcaseFlag and spaceFlag options modify the comparison
 * by ignoring case differences or extra spaces.
 * Returns nonzero if the strings differ.
 */
static int
Compare(
	const char *	s1,
	const char *	s2
)
{
	const char *	origs1;		/* original string pointers */
	const char *	origs2;
	int		c1;		/* temporaries */
	int		c2;

	origs1 = s1;
	origs2 = s2;

	if (fc.spaceFlag)
	{
		while (IsSpace(*s1))
			s1++;

		while (IsSpace(*s2))
			s2++;
	}

	while (TRUE)
	{
		while (*s1 == *s2)
		{
			if (*s1 == '\0')
				return 0;

			s1++;
			s2++;
		}

		if (fc.spaceFlag)
		{
			/*
			 * Skip over extra spaces
			 */
			if (IsSpace(*s1) &&
				((s2 == origs2) || (*s2 == '\n') ||
					IsSpace(s2[-1])))
			{
				do
				{
					s1++;
				} while (IsSpace(*s1));

				continue;
			}

			if (IsSpace(*s2) &&
				((s1 == origs1) || (*s1 == '\n') ||
					IsSpace(s1[-1])))
			{
				do
				{
					s2++;
				} while (IsSpace(*s2));

				continue;
			}
		}

		if (fc.upcaseFlag)
		{
			/*
			 * Upper case letters
			 */
			c1 = *s1++;
			c2 = *s2++;

			if (IsLower(c1))
				c1 = ToUpper(c1);

			if (IsLower(c2))
				c2 = ToUpper(c2);

			if (c1 == c2)
				continue;
		}

		return 1;
	}
}


/*
 * Allocate a new snake structure and append it to the snake list.
 */
static SNAKE *
AllocSnake()
{
	SNAKE *	sp;		/* snake being allocated */

	sp = nextSnake;

	if (sp >= lastSnake)
	{
		sp = (SNAKE *) XAlloc(
			(unsigned long) (sizeof(SNAKE) * SNAKE_ALLOC_SIZE));

		nextSnake = sp;
		lastSnake = sp + SNAKE_ALLOC_SIZE;
	}

	nextSnake++;

	sp->next = NULL;

	if (fc.snakeList == NULL)
		fc.snakeList = sp;
	else
		endSnake->next = sp;

	endSnake = sp;

	if ((fc.verbosity > VERBOSE_DEFAULT) &&
		((++snakeCount % SNAKE_BLAB_INTERVAL) == 0))
	{
		fprintf(stderr, "[%ld snake elements]\n", snakeCount);
	}

	return sp;
}


/*
 * Allocate a string of a certain size.  This routine is used instead of
 * simply calling malloc in order to reduce the overhead.  The allocated
 * strings cannot be freed or realloced.
 */
char *
AllocStr(unsigned long len)
{
	char *	cp;

	len = (len + 3) & ~3;		/* round to multiple of 4 */

	if (len >= UNIQUE_ALLOC_SIZE)
		return XAlloc(len);

	cp = nextChar;

	if (len > availChars)
	{
		cp = XAlloc((unsigned long) CHAR_ALLOC_SIZE);

		nextChar = cp;
		availChars = CHAR_ALLOC_SIZE;
	}

	nextChar += len;
	availChars -= len;

	return cp;
}


/*
 * Routine to allocate memory and quit if failed.
 */
static char *
XAlloc(unsigned long len)
{
	char *	cp;

	cp = malloc(len);

	if (cp == NULL)
		FileError((FIO *) NULL, "Malloc failed");

	return cp;
}


/*
 * Dump lines which match, either just the line numbers or with the text.
 */
void
DumpMatch(const char * outputname)
{
	FIO *		fp;
	const SNAKE *	sp;
	FIO		outputFile;

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputname)
	{
		fprintf(stderr, "[Writing matching lines to \"%s\"]\n",
			outputname);
	}

	fp = &outputFile;

	if (!FioOpen(fp, outputname, FIO_SEQWRITE))
		FileError(fp, "Cannot open output file \"%s\"", outputname);

	if (fc.editScriptFlag)
	{
		FioFormat(fp, "A %s\nB %s\n",
			fc.fileA.name, fc.fileB.name);
	}
	else
	{
		FioFormat(fp, "FILE A: %s\nFILE B: %s\n",
			fc.fileA.name, fc.fileB.name);

		FioFormat(fp,
			"TOTALS: %ld inserted  %ld deleted  %ld matched\n",
			fc.inserts, fc.deletes, fc.matches);
	}

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		if (sp->count <= 0)
			continue;

		PrintRange(fp, "MATCH", sp->line1, sp->count,
			sp->line2, sp->count);

		if (!fc.editScriptFlag)
			PrintLines(fp, &fc.fileA, sp->line1 + 1, sp->count, "");
	}

	if (fc.editScriptFlag)
	{
		FioFormat(fp, "T %ld %ld %ld\n",
			fc.inserts, fc.deletes, fc.matches);
	}

	if (!FioClose(fp))
		FileError(fp, "Write error on \"%s\"", outputname);
}


/*
 * Dump the edit script in the normal manner, showing changed line
 * numbers along with the relevent text.
 */
void
DumpNormal(const char * outputname)
{
	FIO *		fp;		/* output file */
	const SNAKE *	sp;		/* current snake */
	long		line1;		/* current line numbers */
	long		line2;
	long		count1;		/* line counts */
	long		count2;
	FIO		outputFile;	/* output file */

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputname)
	{
		fprintf(stderr, "[Writing differences to \"%s\"]\n",
			outputname);
	}

	fp = &outputFile;

	if (!FioOpen(fp, outputname, FIO_SEQWRITE))
		FileError(fp, "Cannot open output file \"%s\"", outputname);

	if (fc.quickFlag)
	{
		FioFormat(fp, "Files \"%s\" and \"%s\":\n",
			fc.fileA.name, fc.fileB.name);

		FioFormat(fp, "%ld inserted  %ld deleted  %ld matched\n",
			fc.inserts, fc.deletes, fc.matches);

		goto done;
	}

	if ((fc.inserts == 0) && (fc.deletes == 0))
	{
		if (fc.verbosity)
		{
			FioFormat(fp,
				"Files \"%s\" and \"%s\" are identical\n",
				fc.fileA.name, fc.fileB.name);
		}

		goto done;
	}

	FioFormat(fp, "FILE A: %s\nFILE B: %s\n",
		fc.fileA.name, fc.fileB.name);

	FioFormat(fp, "TOTALS: %ld inserted  %ld deleted  %ld matched\n",
		fc.inserts, fc.deletes, fc.matches);

	line1 = 0;
	line2 = 0;

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		count1 = sp->line1 - line1;
		count2 = sp->line2 - line2;

		if (count1 && count2)
		{
			PrintRange(fp, "REPLACE", line1, count1, line2, count2);
			PrintLines(fp, &fc.fileA, line1 + 1, count1, "");
			FioStr(fp, "******************** WITH:\n");
			PrintLines(fp, &fc.fileB, line2 + 1, count2, "");
		}
		else if (count1)
		{
			PrintRange(fp, "DELETE", line1, count1, line2, 1L);
			PrintLines(fp, &fc.fileA, line1 + 1, count1, "");
		}
		else if (count2)
		{
			PrintRange(fp, "INSERT", line1, 1L, line2, count2);
			PrintLines(fp, &fc.fileB, line2 + 1, count2, "");
		}

		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}

done:
	if (!FioClose(fp))
		FileError(fp, "Write error on \"%s\"", outputname);
}


/*
 * Dump all of both files showing what happened to each line.
 */
void
DumpWhat(const char * outputname)
{
	FIO *		fp;
	const SNAKE *	sp;
	long		line1;		/* current line numbers */
	long		line2;
	long		count1;		/* line counts */
	long		count2;
	FIO		outputFile;

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputname)
	{
		fprintf(stderr, "[Writing what changes were made to \"%s\"]\n",
			outputname);
	}

	fp = &outputFile;

	if (!FioOpen(fp, outputname, FIO_SEQWRITE))
		FileError(fp, "Cannot open output file \"%s\"", outputname);

	FioFormat(fp, "FILE A: %s\nFILE B: %s\n",
		fc.fileA.name, fc.fileB.name);

	FioFormat(fp, "TOTALS: %ld inserted  %ld deleted  %ld matched\n",
		fc.inserts, fc.deletes, fc.matches);

	line1 = 0;
	line2 = 0;

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		count1 = sp->line1 - line1;
		count2 = sp->line2 - line2;

		if (count1 && count2)
		{
			PrintLines(fp, &fc.fileA, line1 + 1, count1, "|-");
			PrintLines(fp, &fc.fileB, line2 + 1, count2, "|+");
		}
		else if (count1)
			PrintLines(fp, &fc.fileA, line1 + 1, count1, "|-");
		else if (count2)
			PrintLines(fp, &fc.fileB, line2 + 1, count2, "|+");

		PrintLines(fp, &fc.fileA, sp->line1 + 1, sp->count, "  ");

		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}

	if (!FioClose(fp))
		FileError(fp, "Write error on \"%s\"", outputname);
}


/*
 * Dump only the line numbers which have changed.
 * This is useful for machine processing of the differences.
 */
void
DumpLines(const char * outputname)
{
	FIO *		fp;		/* output file */
	const SNAKE *	sp;		/* current snake */
	long		line1;		/* previous line numbers */
	long		line2;
	long		count1;		/* count of line numbers */
	long		count2;
	FIO		outputFile;	/* output file */

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputname)
	{
		fprintf(stderr, "[Writing edit script to \"%s\"]\n",
			outputname);
	}

	fp = &outputFile;

	if (!FioOpen(fp, outputname, FIO_SEQWRITE))
		FileError(fp, "Cannot open output file \"%s\"", outputname);

	line1 = 0;
	line2 = 0;

	FioFormat(fp, "A %s\nB %s\n", fc.fileA.name, fc.fileB.name);

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		count1 = sp->line1 - line1;
		count2 = sp->line2 - line2;

		if (count1 && count2)
			PrintRange(fp, "R", line1, count1, line2, count2);
		else if (count1)
			PrintRange(fp, "D", line1, count1, line2, 1L);
		else if (count2)
			PrintRange(fp, "I", line1, 1L, line2, count2);

		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}

	FioFormat(fp, "T %ld %ld %ld\n", fc.inserts, fc.deletes, fc.matches);

	if (!FioClose(fp))
		FileError(fp, "Write error on \"%s\"", outputname);
}


/*
 * Print the specified number of lines of a file beginning at the
 * specified line number.  This prints imbedded blank lines too.
 * The lines can be preceeded by line number if desired.  Context
 * lines can be added also.  Tabs are converted to the appropriate
 * number of spaces if line numbering or tags are supplied.
 */
static void
PrintLines(
	FIO *			fp,		/* output file */
	const FILEINFO *	fi,		/* file information */
	long			line,
	long			count,
	const char *		tagStr		/* line tag (if whatFlag) */
)
{
	LINE **		lines;		/* data of file */
	const long *	lnum;		/* line numbers */
	const char *	data;		/* current data line */
	long		curLineNumber;	/* current line number */

	line = line - fc.context - 1;

	count += (2 * fc.context);

	if (line < 0)
	{
		count += line;
		line = 0;
	}

	if ((line + count) > fi->lineCount)
		count = fi->lineCount - line;

	lines = fi->lines + line;
	lnum = fi->lnums + line;
	curLineNumber = *lnum;

	while (count-- > 0)
	{
		while (curLineNumber < *lnum)
		{
			if (fc.whatFlag)
				FioFormat(fp, "%s\n", tagStr);
			else if (fc.numFlag)
			{
				FioFormat(fp, "%c%ld\n", fi->tag,
					curLineNumber);
			}
			else
				FioWrite(fp, "\n", 1L);

			curLineNumber++;
		}

		if (fc.whatFlag)
			FioFormat(fp, "%s  ", tagStr);
		else if (fc.numFlag)
			FioFormat(fp, "%c%-7ld ", fi->tag, curLineNumber);

		data = (*lines)->data;

		if (fc.whatFlag)
			data = Detab(data);

		FioStr(fp, data);

		lines++;
		lnum++;
		curLineNumber++;
	}
}


/*
 * Print the specified range of line numbers for both files.
 * The given string is also typed as part of the message.
 * This routine handles both full and line number output.
 * The real line numbers in each file are typed.
 */
static void
PrintRange(
	FIO *		fp,		/* output file */
	const char *	msg,		/* string to type as part of output */
	long		line1,
	long		count1,
	long		line2,
	long		count2
)
{
	long	beg1;		/* real beginning and ending line numbers */
	long	end1;
	long	beg2;
	long	end2;
	char	buf1[20];	/* string storage */
	char	buf2[20];

	end1 = line1 + count1 - 1;
	end2 = line2 + count2 - 1;

	if (line1 >= fc.fileA.lineCount)
	{
		beg1 = fc.fileA.linesTotal + 1;
		end1 = beg1;
	}
	else
	{
		beg1 = fc.fileA.lnums[line1];
		end1 = fc.fileA.lnums[end1];
	}

	if (line2 >= fc.fileB.lineCount)
	{
		beg2 = fc.fileB.linesTotal + 1;
		end2 = beg2;
	}
	else
	{
		beg2 = fc.fileB.lnums[line2];
		end2 = fc.fileB.lnums[end2];
	}

	sprintf(buf1, ((beg1 == end1) ? "%ld" : "%ld-%ld"), beg1, end1);
	sprintf(buf2, ((beg2 == end2) ? "%ld" : "%ld-%ld"), beg2, end2);

	if (fc.editScriptFlag)
		FioFormat(fp, "%c %s %s\n", *msg, buf1, buf2);
	else
	{
		FioFormat(fp, "\n******************** %s [A%s, B%s]:\n",
			msg, buf1, buf2);
	}
}


/*
 * Translate all tabs to the appropriate number of spaces in a line.
 * This is used when the line is not output at a nice tab boundary,
 * so that the output still looks correct.  Returns the new converted
 * string, which is only valid until the next call.
 */
static char *
Detab(const char * str)
{
	int		pos;		/* current character position */
	int		ch;		/* current character */
	char *		destCp;		/* destination pointer */

	static	char *	strBuf;		/* string buffer */
	static	int	maxLen;		/* maximum size of string buffer */

	if (maxLen == 0)
	{
		maxLen = DETAB_ALLOC_SIZE;

		strBuf = malloc(maxLen + TAB_SIZE + 1);

		if (strBuf == NULL)
			Error("Cannot allocate detab string");
	}

	destCp = strBuf;
	pos = 0;

	while ((ch = *str++) != '\0')
	{
		if (pos >= maxLen)
		{

			maxLen += DETAB_ALLOC_SIZE;

			strBuf = realloc(strBuf, maxLen + TAB_SIZE + 1);

			if (strBuf == NULL)
				Error("Cannot reallocate detab string");

			destCp = &strBuf[pos];
		}

		if (ch != '\t')
		{
			*destCp++ = ch;
			pos++;

			continue;
		}

		do
		{
			*destCp++ = ' ';
			pos++;
		} while (pos % TAB_SIZE);
	}

	*destCp = '\0';

	return strBuf;
}


/*
 * Reinitialize for a new file comparison.
 * This zeroes out our static variables so that another
 * file comparison will work correctly.  Notice that memory is
 * not freed here since it is freed elsewhere by calling CmReset.
 */
void
FcompReset()
{
	tableSize = 0;
	snakeCount = 0;
	availChars = 0;

	fc.inserts = 0;
	fc.deletes = 0;
	fc.matches = 0;
	fc.snakeList = NULL;

	nextChar = NULL;
	V1 = NULL;
	V2 = NULL;

	nextSnake = NULL;
	lastSnake = NULL;
	endSnake = NULL;
	hashTable = NULL;
}

/* END CODE */
