/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to position within an ASCII file and read, write, or update it.
 * The functionality here is mostly historical, and STDIO could probably
 * replace most of it.
 */

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <unistd.h>

#include "fio.h"
#include "cmalloc.h"


/*
 * If there is no binary open flag then define a dummy one.
 */
#ifndef	O_BINARY
#define	O_BINARY	0
#endif


#define	LINE_ALLOC_SIZE	132		/* allocation size for line buffer */
#define	BLOCK_SIZE	8192		/* size of an I/O block */
#define	TTY_UNIT	-1		/* file descriptor for terminal I/O */


static	char *	lineBuffer;		/* common line buffer */
static	int	lineLength;		/* total length of line */


/*
 * Open a file in the specified mode, or a terminal.
 * The specified file structure is overwritten with the file information.
 * This reads in the first block of the file for examination.
 * The terminal is indicated by a null pointer for the name.
 * Returns true if successful.
 */
BOOL
FioOpen(
	FIO *		fp,		/* file structure to fill in */
	const char *	name,		/* file name to be opened */
	int		mode
)
{
	int	unit;		/* file unit file is opened on */
	int	count;		/* words read initially */

	fp->unit = 0;
	fp->valid = 0;
	fp->pos = 0;
	fp->offset = 0;
	fp->errorCode = 0;
	fp->dirty = 0;
	fp->eofFlag = FALSE;
	fp->canRead = TRUE;

	if (name == NULL)
	{
		fp->unit = TTY_UNIT;
		fp->data = NULL;

		return TRUE;
	}

	fp->data = malloc(BLOCK_SIZE);

	if (fp->data == NULL)
		return FALSE;

	switch (mode)
	{
		case FIO_READ:
			/*
			 * Read only
			 */
			unit = open(name, O_RDONLY | O_BINARY);
			break;

		case FIO_WRITE:
		case FIO_SEQWRITE:
			/*
			 * Write only or sequential access write only.
			 */
			fp->canRead = FALSE;

			unit = open(name,
				O_WRONLY | O_CREAT | O_TRUNC | O_BINARY,
				0666);

			break;

		case FIO_UPDATE:
			/*
			 * Read and write of existing file.
			 */
			unit = open(name, O_RDWR | O_BINARY);
			break;

		default:
			free(fp->data);

			return -1;
	}

	if (unit < 0)
	{
		fp->errorCode = errno;

		free(fp->data);

		return FALSE;
	}

	fp->unit = unit;

	count = 0;

	if (fp->canRead)
	{
		count = read(unit, fp->data, BLOCK_SIZE);

		if (count < 0)
		{
			fp->errorCode = errno;
			fp->unit = 0;

			(void) close(unit);

			free(fp->data);

			return FALSE;
		}
	}

	fp->eofFlag = (count == 0);
	fp->valid = count;

	return TRUE;
}


/*
 * Close a file.
 * This flushes the final data for the file if necessary.
 * If an error had occurred anytime after opening the file and was
 * never cleared, this will be indicated here.  Thus it can be convenient
 * to check for errors only during the close to detect any errors when
 * writing to a file.  Returns true if all operations had been successful.
 */
BOOL
FioClose(FIO * fp)
{
	if (fp->unit == TTY_UNIT)
	{
		fflush(stdout);

		return TRUE;
	}

	if (fp->dirty)
		FioFlush(fp);

	if (fp->unit && close(fp->unit))
		fp->errorCode = errno;

	free(fp->data);

	fp->data = NULL;
	fp->unit = 0;
	fp->valid = 0;
	fp->offset = 0;
	fp->pos = 0;
	fp->dirty = 0;
	fp->eofFlag = FALSE;

	if (fp->errorCode)
		return FALSE;

	return TRUE;
}


/*
 * Position absolutely within a file to the indicated byte position,
 * and read in the new block of the file for examination.  Returns true
 * if successful.  Sets or clears the end of file status flag.
 */
BOOL
FioSeek(
	FIO *	fp,		/* file to position in */
	long	pos		/* position to seek to */
)
{
	long	offset;			/* offset within block */
	int	count;

	if (fp->unit == TTY_UNIT)
		return FALSE;

	offset = pos - fp->pos;

	if ((offset >= 0) && (offset < fp->valid))
	{
		fp->offset = offset;
		fp->eofFlag = FALSE;

		return TRUE;
	}

	if (fp->dirty && !FioFlush(fp))
		return FALSE;

	fp->offset = pos % BLOCK_SIZE;
	fp->pos = pos - fp->offset;
	fp->valid = 0;

	if (lseek(fp->unit, fp->pos, 0) < 0)
	{
		fp->errorCode = errno;

		return FALSE;
	}

	count = 0;

	if (fp->canRead)
	{
		count = read(fp->unit, fp->data, BLOCK_SIZE);

		if (count < 0)
		{
			fp->errorCode = errno;

			return FALSE;
		}
	}

	fp->eofFlag = (count == 0);
	fp->valid = count;

	return TRUE;
}


/*
 * Skip forwards through a file by the specified number of lines.
 * Returns true on success.  End of file counts as an error.
 */
BOOL
FioSkip(
	FIO *	fp,		/* file to scan through */
	long	count		/* number of lines to skip */
)
{
	const char *	cp;		/* current character */
	int		cc;		/* characters left in buffer */
	int		offset;

	if (fp->unit == TTY_UNIT)
		return TRUE;

	while (TRUE)
	{
		cc = fp->valid - fp->offset;

		if (cc <= 0)
		{
			if (!FioSeek(fp, fp->pos + BLOCK_SIZE))
				return FALSE;

			if (FioEof(fp))
				return FALSE;

			cc = fp->valid - fp->offset;
		}

		cp = &fp->data[fp->offset];

		while (--cc >= 0)
		{
			if (*cp++ != '\n')
				continue;

			offset = fp->valid - cc;

			if (--count <= 0)
			{
				fp->offset = offset;

				return TRUE;
			}
		}

		fp->offset = fp->valid;
	}
}


/*
 * Routine to read in the next line from a file, no matter how long it is.
 * This returns a pointer to the string, and also indirectly its length.
 * The string is normally returned from a static string, which is reused
 * for each call.  But if keepFlag is TRUE, then the returned string belongs
 * to the caller and must later be freed.  The returned line is ended with
 * a newline (if present) and a trailing null character, but the returned
 * length does not count the null character.  Returns NULL on an error or
 * end of file, with any error stored in the fp structure.  The data in the
 * line can be modified by the caller.  The returned line is guaranteed to
 * be a buffer which is at least 4 characters in size.
 */
char *
FioReadline(
	FIO *	fp,		/* file to read line from */
	long *	retLen,		/* returned length of line if non-null */
	BOOL	keepFlag	/* TRUE if line is owned by the caller */
)
{
	char *	cp;		/* pointer to character string */
	char *	dp;		/* destination pointer */
	long	curSize;	/* current size */
	long	count;		/* length of data read from file */
	long	totalLength;	/* total length of data */

	totalLength = 0;

	if (lineLength == 0)
	{
		/*
		 * Allocate line buffer
		 */
		lineBuffer = malloc(LINE_ALLOC_SIZE + 2);

		if (lineBuffer == NULL)
		{
			fp->errorCode = errno;

			return NULL;
		}

		lineLength = LINE_ALLOC_SIZE;
	}

	cp = lineBuffer;

	curSize = lineLength;

	while (TRUE)
	{
		count = FioRead(fp, cp, curSize);

		if (count < 0)
			return NULL;

		if (count == 0)
		{
			if (totalLength > 0)
				break;

			return NULL;
		}

		totalLength += count;

		cp[count] = '\0';

		if (cp[count-1] == '\n')
			break;

		lineBuffer = realloc(lineBuffer,
			lineLength + (LINE_ALLOC_SIZE + 2));

		if (lineBuffer == NULL)
		{
			fp->errorCode = errno;

			return NULL;
		}

		cp = lineBuffer + totalLength;

		lineLength += LINE_ALLOC_SIZE;
		curSize = LINE_ALLOC_SIZE;
	}

	if (retLen != NULL)
		*retLen = totalLength;

	if (!keepFlag)
		return lineBuffer;

	/*
	 * The caller wants to keep the line.
	 * So allocate a new one and return that.
	 */
	cp = lineBuffer;

	dp = malloc(totalLength + 4);

	if (dp == NULL)
	{
		fp->errorCode = errno;

		return NULL;
	}

	memcpy(dp, cp, (int) totalLength);

	dp[totalLength] = '\0';

	return dp;
}


/*
 * Read data from the current position in a file until the end of the line
 * is found or the supplied buffer is full (including a terminating null).
 * Returns the number of characters returned (including a possible
 * terminating newline), or 0 on end of file, or -1 on an error.
 * The line is terminated by a null character.
 */
long
FioRead(
	FIO *	fp,		/* file to read from */
	char *	buf,		/* buffer to read data in */
	long	size		/* size of buffer */
)
{
	char *	cp;		/* current position in file structure */
	long	origSize;
	int	cc;
	int	ci;
	int	ch;

	if (fp->unit == TTY_UNIT)
		return 0;

	/*
	 * Make sure the buffer is large enough.
	 */
	if (size < 2)
	{
		fp->errorCode = EPERM;

		return -1;
	}

	/*
	 * Remove one character to leave room for the null.
	 */
	size--;
	origSize = size;

	/*
	 * Keep reading until the buffer is full or we see a newline
	 * or end of file.
	 */
	while (TRUE)
	{
		/*
		 * Get the number of character left in this buffer.
		 */
		cc = fp->valid - fp->offset;

		/*
		 * If there are no more characters left in this buffer
		 * then try to read the next buffer.
		 */
		if (cc <= 0)
		{
			/*
			 * If the buffer was not totally full then we
			 * had end of file so return the number of
			 * character which we had.
			 */
			if (fp->valid != BLOCK_SIZE)
			{
				*buf = '\0';

				return origSize - size;
			}

			if (!FioSeek(fp, fp->pos + BLOCK_SIZE))
				return -1;

			if (FioEof(fp))
			{
				/*
				 * Terminate the buffer.
				 */
				*buf = '\0';

				/*
				 * Return the number of characters read.
				 * (This is zero on end of file.)
				 */
				return origSize - size;
			}

			cc = fp->valid;
		}

		/*
		 * Set up to read data from the buffer.
		 */
		cp = &fp->data[fp->offset];

		if (cc > size)
			cc = size;

		/*
		 * Loop character by character looking for a newline.
		 */
		ci = 0;

		while (ci < cc)
		{
			ci++;

			ch = *cp++;
			*buf++ = ch;

			if (ch == '\n')
			{
				*buf = '\0';
				fp->offset += ci;

				return (origSize - size + ci);
			}
		}

		size -= ci;
		fp->offset += ci;

		if (size == 0)
			return origSize;
	}
}


/*
 * Write formatted data at the current position in a file.
 * This appears similar to printf.  Returns true if successful.
 * Restrictions:
 *	Maximum output string of 8191 characters.
 *	No nulls in output.
 */
BOOL
FioFormat(
	FIO *		fp,
	const char *	fmt,
	...
)
{
	va_list	ap;
	char	buf[8192];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	return FioWrite(fp, buf, (long) strlen(buf));
}


/*
 * Write a null terminated string to the current position in a file.
 * Returns true on success.
 */
BOOL
FioStr(
	FIO *		fp,		/* file to write to */
	const char *	str		/* string to be written */
)
{
	return FioWrite(fp, str, (long) strlen(str));
}


/*
 * Write a single character to the current position in a file.
 * Returns true in success.
 */
BOOL
FioChar(
	FIO *	fp,		/* file to write to */
	int	ch		/* character to be written */
)
{
	char	realchar;

	realchar = ch;

	return FioWrite(fp, &realchar, 1L);
}


/*
 * Write data at the current position in a file.
 * This buffers the data in the current block, thus flushing is needed later.
 * Returns true if successful.
 */
BOOL
FioWrite(
	FIO *		fp,		/* file to write to */
	const char *	buf,		/* buffer to write data out */
	long		size		/* amount of data to write */
)
{
	char *	cp;		/* current character address */
	int	cc;		/* character count */

	if (fp->unit == TTY_UNIT)
	{
		fwrite(buf, 1, (int) size, stdout);

		return TRUE;
	}

	cp = &fp->data[fp->offset];
	cc = BLOCK_SIZE - fp->offset;

	while (size > 0)
	{
		if (cc <= 0)
		{
			fp->dirty = BLOCK_SIZE;

			if (!FioFlush(fp))
				return FALSE;

			if (!FioSeek(fp, fp->pos + BLOCK_SIZE))
				return FALSE;

			cp = fp->data;
			cc = BLOCK_SIZE;
		}

		*cp++ = *buf++;
		size--;
		cc--;
	}

	fp->offset = BLOCK_SIZE - cc;

	if (fp->dirty < fp->offset)
		fp->dirty = fp->offset;

	fp->eofFlag = FALSE;

	return TRUE;
}


/*
 * Routine to copy data from one file to another.  Data is copied from
 * the current position of the input file to the current position of the
 * output file until end of file is reached on the input file, or until
 * the specified number of characters has been read.  Minus one means
 * there is no limit to the amount to be copied.  Line boundaries are
 * ignored.  Returns true on success.
 */
BOOL
FioCopy(
	FIO *	ip,		/* input file for copy */
	FIO *	op,		/* output file for copy */
	long	count		/* chars to copy, or -1 for infinite */
)
{
	const char *	icp;		/* input character pointer */
	char *		ocp;		/* output character pointer */
	int		cc;		/* current character count */
	int		icc;		/* input and output character counts */
	int		occ;

	if (ip->unit == TTY_UNIT)
		return FALSE;

	if (op->unit == TTY_UNIT)
		return FioType(ip);

	while (count)
	{
		icc = ip->valid - ip->offset;

		if (icc <= 0)
		{
			if (!FioSeek(ip, ip->pos + BLOCK_SIZE))
				return FALSE;

			if (FioEof(ip))
				return TRUE;

			icc = ip->valid;
		}

		occ = BLOCK_SIZE - op->offset;

		if (occ <= 0)
		{
			if (!FioSeek(op, op->pos + BLOCK_SIZE))
				return FALSE;

			occ = BLOCK_SIZE;
		}

		cc = icc;

		if (cc > occ)
			cc = occ;

		if ((count > 0) && (cc > count))
			cc = count;

		icp = &ip->data[ip->offset];
		ocp = &op->data[op->offset];

		ip->offset += cc;
		op->offset += cc;
		count -= cc;

		if (op->dirty < op->offset)
			op->dirty = op->offset;

		memcpy(ocp, icp, cc);
	}

	return TRUE;
}


/*
 * Routine to type out the rest of an input file to the terminal.
 * It is assumed that we are positioned at the beginning of a line.
 * Returns true on success.
 */
BOOL
FioType(FIO * fp)
{
	const char *	cp;		/* character pointer */
	int		ch;		/* current character */
	int		icc;		/* available number of characters */
	int		cc;		/* remaining number of characters */

	if (fp->unit == TTY_UNIT)
		return FALSE;

	while (TRUE)
	{
		icc = fp->valid - fp->offset;

		if (icc <= 0)
		{
			if (!FioSeek(fp, fp->pos + BLOCK_SIZE))
				return FALSE;

			if (FioEof(fp))
				return TRUE;

			icc = fp->valid;
		}

		cp = &fp->data[fp->offset];

		for (cc = 0; cc < icc; cc++)
		{
			ch = *cp++;
			putchar(ch);
		}

		fp->offset = fp->valid;
	}

	fflush(stdout);

	if (ferror(stdout))
		return FALSE;

	return TRUE;
}


/*
 * Routine called to flush out any data which might need writing
 * in the current buffer.  Returns true if successful.
 */
BOOL
FioFlush(FIO * fp)
{
	int	count;			/* number of bytes to write */

	if (fp->unit == TTY_UNIT)
	{
		fflush(stdout);

		return TRUE;
	}

	count = fp->dirty;

	if (count <= 0)
		return TRUE;

	fp->dirty = 0;

	if (lseek(fp->unit, fp->pos, 0) < 0)
	{
		fp->errorCode = errno;

		return FALSE;
	}

	if (write(fp->unit, fp->data, count) != count)
	{
		fp->errorCode = errno;

		return FALSE;
	}

	return TRUE;
}


/*
 * Reset file I/O static variables for restarting.
 * Notice that memory is not freed here since that is done elsewhere by
 * calling CmReset.
 */
void
FioReset()
{
	lineBuffer = NULL;
	lineLength = 0;
}

/* END CODE */
