/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Merge differences from two divergent edits.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>

#include "fio.h"
#include "fcomp.h"
#include "fmerge.h"
#include "sc.h"
#include "cmalloc.h"


/*
 * The common data for merging.
 */
FMERGE_DATA	fm;


static	const char *	Expand(const char *);
static	void		Usage(void);


/*
 * This is the main program.
 */
int
main(
	int		argc,
	const char *	argv[]
)
{
	const char *	cp;		/* current option character */
	const char *	baseName;	/* file which is the base file */
	const char *	nameA;		/* the two divergent files */
	const char *	nameB;
	const char *	outputName;	/* output merged filename */
	const char *	conflictName;	/* output conflict filename */
	BOOL		outputFlag;
	BOOL		conflictFlag;
	int		fileCount;	/* number of files specified */
	int		opt;		/* option character */

	argv++;
	argc--;

	if (argc <= 0)
	{
		Usage();

		exit(1);
	}

	baseName = NULL;
	nameA = NULL;
	nameB = NULL;
	outputName = NULL;
	conflictName = NULL;

	fc.verbosity = VERBOSE_DEFAULT;
	fc.debugFlag = FALSE;

	outputFlag = FALSE;
	conflictFlag = FALSE;
	fileCount = 0;

	fm.conflicts = 0;
	fm.ignoreFlag = FALSE;
	fm.failCount = INFINITY;
	fm.unchangeCount = INFINITY;

	while (argc-- > 0)
	{
		cp = *(argv++);

		if (*cp != '-')
		{
			/*
			 * A file name
			 */
			cp = Expand(cp);

			fileCount++;

			if (baseName == NULL)
			{
				baseName = cp;
				continue;
			}

			if (nameA == NULL)
			{
				nameA = cp;
				continue;
			}

			if (nameB == NULL)
			{
				nameB = cp;
				continue;
			}

			Error("Only three input files allowed");
		}

		for (cp++; *cp; cp++)
		{
			opt = *cp;

			if ((opt >= 'A') && (opt <= 'Z'))
				opt += ('a' - 'A');

			switch (opt)
			{
			case 'd':		/* type debugging info */
				fc.debugFlag = TRUE;
				break;

			case 'h':		/* display help */
				Usage();
				exit(0);

			case 'i':		/* ignore conflicts */
				fm.ignoreFlag = TRUE;
				break;

			case 'v':		/* be verbose */
				fc.verbosity = VERBOSE_FULL;

				if (cp[1])
					Error("No options allowed after -v");

				if ((argc <= 0) || (**argv == '-'))
					break;

				fc.verbosity = 0;

				cp = *argv++;

				while ((*cp >= '0') && (*cp <= '9'))
				{
					fc.verbosity = (fc.verbosity * 10) +
						*cp++ - '0';
				}

				if (*cp || (fc.verbosity < 0))
					Error("Bad verbosity value specified");

				argc--;
				cp = "x";
				break;

			case 'o':		/* specify output file */
				if (cp[1])
					Error("No options allowed after -o");

				outputFlag = TRUE;

				if ((argc > 0) && (**argv != '-'))
				{
					outputName = Expand(*(argv++));
					argc--;
				}

				break;

			case 'c':		/* specify conflict file */
				if (cp[1])
					Error("No options allowed after -c");

				conflictFlag = TRUE;

				if ((argc > 0) && (**argv != '-'))
				{
					conflictName = Expand(*(argv++));
					argc--;
				}

				break;

			case 'f':		/* fail if too many conflicts */
				if (cp[1])
					Error("No options allowed after -f");

				if ((argc <= 0) || (**argv == '-'))
					Error("No failure value specified");

				fm.failCount = 0;

				cp = *argv++;

				while ((*cp >= '0') && (*cp <= '9'))
				{
					fm.failCount = (fm.failCount * 10) +
						*cp++ - '0';
				}

				if (*cp || (fm.failCount < 0))
					Error("Bad failure value specified");

				argc--;
				cp = "x";
				break;

			case 'u':		/* limit unchanged lines */
				if (cp[1])
					Error("No options allowed after -u");

				if ((argc <= 0) || (**argv == '-'))
					Error("No unchange value specified");

				fm.unchangeCount = 0;

				cp = *argv++;

				while ((*cp >= '0') && (*cp <= '9'))
				{
					fm.unchangeCount =
						(fm.unchangeCount * 10) +
						*cp++ - '0';
				}

				if (*cp || (fm.unchangeCount < 0))
					Error("Bad unchange value specified");

				argc--;
				cp = "x";
				break;

			default:
				Error("Unknown option '-%c'", opt);
			}
		}
	}

	if (fileCount == 1)
	{
		/*
		 * Converting conflict file
		 */
		if (conflictFlag)
		{
			Error(
		"Conflicting options for conflict file conversion");
		}

		ConvertConflicts(baseName, outputName);

		exit(0);
	}

	if (fileCount != 3)
		Error("Must have three input files for merge");

	if (!outputFlag && !conflictFlag)
		outputFlag = TRUE;

	Fmerge(baseName, nameA, nameB);

	if (fc.verbosity)
	{
		printf("Total of %ld conflict%s\n", fm.conflicts,
			(fm.conflicts == 1) ? "" : "s");
	}

	if (outputFlag)
		DumpMergedFile(outputName);

	if (conflictFlag)
		DumpConflicts(conflictName);

	if (fm.conflicts && !conflictFlag)
		return 1;

	return 0;
}


/*
 * Expand a global variable name.
 * Returns original argument if string is not a global variable,
 * or returns copied value of global variable.
 */
static const char *
Expand(const char * cp)
{
	const char *	var;
	char *		data;

	if ((cp == NULL) || (*cp != '.'))
		return cp;

	var = getenv(cp);

	if ((var == NULL) || (*var == '\0'))
		return cp;

	data = malloc(strlen(var) + 1);

	if (data == NULL)
		Error("Malloc failed");

	strcpy(data, var);

	return data;
}


/*
 * Dummy routine to make fcomp happy.
 */
const char *
ModifyLine(
	const char *	cp,
	long *		retLen,
	const INFO *	info
)
{
	return cp;
}


/*
 * Error routine which can format the output.
 */
void
Error(
	const char *	fmt,
	...
)
{
	va_list	ap;
	char	buf[8192];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	fprintf(stderr, "%s\n", buf);

	exit(1);
}


/*
 * Error routine called for file errors.
 * This prints out the error message for a file error.
 */
void
FileError(
	const FIO *	fp,
	const char *	fmt,
	...
)
{
	va_list	ap;
	char	buf[8192];

	if (fp && fp->errorCode)
		errno = fp->errorCode;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	perror(buf);

	exit(1);
}


/*
 * Help message
 */
static const char * const helpText[] =
{
"",
"Program to merge edits from two files derived from a base file.",
"",
"Usage is one of the following:",
"    fmerge basefile file1 file2 [-v n] [-o [file]] [-c [file]] [-f n] [-u n] [-i]",
"    fmerge conflictfile [-o [outputfile]] [-v n] [-i]",
"",
"Options are:",
"   -c:  output possible conflicts to file (or terminal if NULL)",
"   -o:  output merged edits to file (or terminal if NULL)",
"   -u:  only show up to n unchanged lines around possible conflicts",
"   -f:  fail if there are more than n conflict lines",
"   -h:  display this message",
"   -i:  ignore conflicts",
"   -v:  set verbosity level about what is going on",
NULL
};


/*
 * Type out the possible commands.
 */
static void
Usage()
{
	const char * const *	str;

	for (str = helpText; *str; str++)
		fprintf(stderr, "%s\n", *str);
}

/* END CODE */
