/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Worker routine to merge differences from a base file.
 */

#include <stdio.h>
#include <stdlib.h>

#include "fio.h"
#include "fcomp.h"
#include "fmerge.h"
#include "sc.h"
#include "cmalloc.h"


#define CONFLICT_BEGIN	\
"/-/-/-/-/-/-/-/-/-/ BEGIN CONFLICT  [O%ld A%ld B%ld] /-/-/-/-/-/-/-/-/-/-/\n"

#define CONFLICT_MIDDLE	\
"/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/\n"

#define CONFLICT_END	\
"/-/-/-/-/-/-/-/-/-/-/-/-/  END CONFLICT   /-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/\n"


#define F_OLD_DATA	0x0000		/* lines the same as the old ones */
#define F_DELETE_A	0x0001		/* lines deleted by file A */
#define F_INSERT_A	0x0002		/* lines inserted by file A */
#define F_DELETE_B	0x0004		/* lines deleted by file B */
#define F_INSERT_B	0x0008		/* lines inserted by file B */
#define F_CONFLICT	0x0010		/* this line is a conflict */

#define MAXSHORT	0x7fff		/* highest positive short value */


typedef struct merge	MERGE;

struct merge
{
	long	baseLine;	/* line number in base file */
	long	fileAline;	/* line number in file A */
	long	fileBline;	/* line number in file B */
	short	baseCount;	/* lines in base file used or deleted */
	short	fileAcount;	/* number of lines inserted by file A */
	short	fileBcount;	/* number of lines inserted by file B */
	short	flags;		/* what is happening to this line */
	MERGE *	next;		/* next line */
};


/*
 * Local data.
 */
static	MERGE *	begMergelist;	/* head of merge list */
static	MERGE *	endMergelist;	/* end of merge list */
static	LINE **	baseLines;	/* lines of base file */
static	LINE **	fileAlines;	/* lines of file A */
static	LINE **	fileBlines;	/* lines of file B */


/*
 * Local routines.
 */
static	MERGE *	NewMerge(void);
static	MERGE *	WalkMerges(FIO *, MERGE *, int);


/*
 * Merge differences between two divergent files, based on a common
 * precursor file.  The resulting output is saved in memory so that it
 * can be examined or output later.
 */
void
Fmerge(
	const char *	baseName,	/* original file for both edits */
	const char *	nameA,		/* file having first set of changes */
	const char *	nameB		/* file having second set of changes */
)
{
	SNAKE *		spA;		/* snake for edits of first file */
	SNAKE *		spB;		/* snake for edits of second file */
	MERGE *		mp;		/* current merge structure */
	SNAKE *		fileAsnakes;	/* snake for first edit */
	SNAKE *		fileBsnakes;	/* snake for second edit */
	long		baseLine;	/* current line of base file */
	long		lineA;		/* current line of file A */
	long		lineB;		/* current line of file B */
	long		deletesA;	/* deletes needed for file A */
	long		insertsA;	/* inserts needed for file A */
	long		deletesB;	/* deletes needed for file B */
	long		insertsB;	/* inserts needed for file B */
	long		count;		/* number of lines being examined */
	long		elements;	/* number of elements used */
	int		switchValue;	/* value to switch on */
	BOOL		bad;		/* TRUE if bad merge detected */

	fc.maxJoin = 0;
	fc.maxChanges = INFINITY;

	Fcomp(baseName, nameA);

	baseLines = fc.fileA.lines;
	fileAlines = fc.fileB.lines;
	fc.fileB.lines = NULL;
	fileAsnakes = fc.snakeList;
	fc.snakeList = NULL;

	Fcomp(NULL, nameB);

	fileBsnakes = fc.snakeList;
	fileBlines = fc.fileB.lines;
	fc.fileB.lines = NULL;
	fc.snakeList = NULL;

	/*
	 * Now walk through the snakes and merge the lines.
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Creating merge list]\n");

	spA = fileAsnakes;
	spB = fileBsnakes;

	baseLine = 0;
	lineA = 0;
	lineB = 0;
	fm.conflicts = 0;
	elements = 0;

	begMergelist = NULL;
	endMergelist = NULL;

	while (TRUE)
	{
		deletesA = spA->line1 - baseLine;
		insertsA = spA->line2 - lineA;
		deletesB = spB->line1 - baseLine;
		insertsB = spB->line2 - lineB;

		if (deletesA > MAXSHORT)
			deletesA = MAXSHORT;

		if (insertsA > MAXSHORT)
			insertsA = MAXSHORT;

		if (deletesB > MAXSHORT)
			deletesB = MAXSHORT;

		if (insertsB > MAXSHORT)
			insertsB = MAXSHORT;

		mp = NewMerge();

		mp->baseLine = baseLine;
		mp->fileAline = lineA;
		mp->fileBline = lineB;
		mp->fileAcount = insertsA;
		mp->fileBcount = insertsB;

		if ((fc.verbosity > VERBOSE_DEFAULT) &&
			((++elements % 1000) == 0))
		{
			printf("[%ld merge elements]\n", elements);
		}

		/*
		 * Calculate a switch value based on whether there are
		 * inserts and deletes for each of the two files.
		 * This results in a number from 0 to 15, where each
		 * bit represents some count being nonzero.
		 */
#define	VALUE(iA, dA, iB, dB) \
	(((iA) != 0) + ((dA) != 0) * 2 + ((iB) != 0) * 4 + ((dB) != 0) * 8)

		switchValue = VALUE(insertsA, deletesA, insertsB, deletesB);

		/*
		 * Also determine whether there is some programming error.
		 */
		bad = ((deletesA < 0) || (insertsA < 0) ||
			(deletesB < 0) || (insertsB < 0));

		if (fc.debugFlag || bad)
		{
			fprintf(stderr,
	"SA %ld %ld %ld,  SB %ld %ld %ld,  Lines %ld %ld %ld,  Case %d\n",
				spA->line1, spA->line2, spA->count,
				spB->line1, spB->line2, spB->count,
				baseLine, lineA, lineB, switchValue);

			if (bad)
				Error("Bad merge calculation");
		}

		/*
		 * Switch on the 16 combinations of inserts and deletes
		 * for the two files as calculated above.  The VALUE macro
		 * generates the appropriate switch value for the cases.
		 * The arguments are nonzero where there are inserts for A,
		 * deletes for A, inserts for B, or deletes for B.
		 */
		switch (switchValue)
		{
		case VALUE(0,0,0,0):
			/*
			 * No inserts or deletes at all.
			 */
			if ((spA->next == NULL) && (spB->next == NULL))
			{
				if (fc.verbosity > VERBOSE_DEFAULT)
				{
					printf("[%ld merge elements used]\n",
						elements);
				}

				if (fm.ignoreFlag)
					fm.conflicts = 0;

				return;
			}

			count = spA->count;

			if (count > spB->count)
				count = spB->count;

			if (count > MAXSHORT)
				count = MAXSHORT;

			mp->flags = F_OLD_DATA;
			mp->baseCount = count;

			lineA += count;
			lineB += count;

			spA->line1 += count;
			spA->line2 += count;
			spA->count -= count;
			spB->line1 += count;
			spB->line2 += count;
			spB->count -= count;

			break;


		case VALUE(1,0,0,0):
			/*
			 * Insert for file A.
			 */
			mp->flags = F_INSERT_A;

			break;


		case VALUE(1,1,0,0):
			/*
			 * Insert and delete for file A.
			 */
			mp->flags = F_INSERT_A;

			/*
			 * FALL INTO NEXT CASE.
			 */

		case VALUE(0,1,0,0):
			/*
			 * Delete for file A
			 */
			count = deletesA;

			if (count > spB->count)
				count = spB->count;

			mp->flags |= F_DELETE_A;
			mp->baseCount = count;

			lineB += count;

			spB->line1 += count;
			spB->line2 += count;
			spB->count -= count;

			break;


		case VALUE(0,0,1,1):
			/*
			 * Insert and delete for file B.
			 */
			mp->flags = F_INSERT_B;

			/*
			 * FALL INTO NEXT CASE.
			 */

		case VALUE(0,0,0,1):
			/*
			 * Delete for file B.
			 */
			count = deletesB;

			if (count > spA->count)
				count = spA->count;

			mp->flags |= F_DELETE_B;
			mp->baseCount = count;

			lineA += count;

			spA->line1 += count;
			spA->line2 += count;
			spA->count -= count;

			break;


		case VALUE(0,1,0,1):
			/*
			 * Files A and B both delete.
			 */
			count = deletesA;

			if (count > deletesB)
				count = deletesB;

			mp->flags = (F_DELETE_A | F_DELETE_B | F_CONFLICT);
			mp->baseCount = count;

			fm.conflicts += count;

			break;


		case VALUE(1,1,0,1):
			/*
			 * File A inserts and deletes, file B deletes.
			 */
			count = deletesA;

			if (count > deletesB)
				count = deletesB;

			mp->flags = (F_INSERT_A | F_DELETE_A |
				F_DELETE_B | F_CONFLICT);

			mp->baseCount = count;

			fm.conflicts += count;

			break;


		case VALUE(0,0,1,0):
			/*
			 * Insert for file B.
			 */
			mp->flags = F_INSERT_B;

			break;


		case VALUE(1,0,1,0):
			/*
			 * File A and file B both insert.
			 */
			mp->flags = (F_INSERT_A | F_INSERT_B | F_CONFLICT);
			fm.conflicts++;

			break;


		case VALUE(1,1,1,0):
			/*
			 * Insert and delete for A, insert for B.
			 */
			mp->flags = F_INSERT_A;

			/*
			 * FALL INTO NEXT CASE.
			 */

		case VALUE(0,1,1,0):
			/*
			 * Delete for file A and insert for file B.
			 */
			mp->flags |= (F_DELETE_A | F_INSERT_B | F_CONFLICT);
			mp->baseCount = 1;

			spB->line1++;
			spB->line2++;
			spB->count--;

			lineB++;

			fm.conflicts++;

			break;


		case VALUE(0,1,1,1):
			/*
			 * File A deletes, file B inserts and deletes.
			 */
			count = deletesA;

			if (count > deletesB)
				count = deletesB;

			mp->flags = (F_DELETE_A | F_INSERT_B |
				F_DELETE_B | F_CONFLICT);

			mp->baseCount = count;

			fm.conflicts += count;

			break;


		case VALUE(1,0,1,1):
			/*
			 * Insert for A, insert and delete for B.
			 */
			mp->flags = F_INSERT_B;

			/*
			 * FALL INTO NEXT CASE.
			 */

		case VALUE(1,0,0,1):
			/*
			 * Insert for file A and delete for file B.
			 */
			mp->flags |= (F_INSERT_A | F_DELETE_B | F_CONFLICT);
			mp->baseCount = 1;

			spA->line1++;
			spA->line2++;
			spA->count--;

			lineA++;

			fm.conflicts++;

			break;


		case VALUE(1,1,1,1):
			/*
			 * Insert and delete for A and for B.
			 */
			count = deletesA;

			if (count > deletesB)
				count = deletesB;

			mp->flags = (F_INSERT_A | F_DELETE_A |
				F_INSERT_B | F_DELETE_B | F_CONFLICT);

			mp->baseCount = count;

			fm.conflicts += count;

			break;


		default:
			Error("Unknown switch value %d", switchValue);
		}

		if ((fm.conflicts > fm.failCount) && !fm.ignoreFlag)
		{
			Error("More than %ld conflict%s found during merge",
				fm.failCount, (fm.failCount == 1) ? "" : "s");
		}

		baseLine += mp->baseCount;

		lineA += mp->fileAcount;
		lineB += mp->fileBcount;

		if ((spA->count <= 0) && (spA->next))
			spA = spA->next;

		if ((spB->count <= 0) && (spB->next))
			spB = spB->next;
	}
}


/*
 * Walk through the merge list and output the resulting file.
 * Conflicts will be identified by conflict identification lines.
 */
void
DumpMergedFile(const char * outputName)
{
	MERGE *	mp;		/* current merge structure */
	LINE **	lp;		/* current line */
	FIO *	fp;		/* file to output to (or TERMINAL) */
	long	count;		/* line count */
	BOOL	conflict;	/* true if inside of a conflict */
	BOOL	newConflict;	/* new conflict value */
	FIO	outputFile;	/* output file */

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputName)
		printf("[Writing merged lines to \"%s\"]\n", outputName);

	fp = &outputFile;

	if (!FioOpen(fp, outputName, FIO_SEQWRITE))
		FileError(fp, "Cannot write output file \"%s\"", outputName);

	conflict = FALSE;

	for (mp = begMergelist; mp; mp = mp->next)
	{
		newConflict = ((mp->flags & F_CONFLICT) && !fm.ignoreFlag);

		if (newConflict != conflict)
		{
			conflict = newConflict;

			FioFormat(fp, conflict ? CONFLICT_BEGIN : CONFLICT_END,
				mp->baseLine + 1, mp->fileAline + 1,
				mp->fileBline + 1);
		}

		if (mp->flags == F_OLD_DATA)
		{
			lp = &baseLines[mp->baseLine];

			for (count = mp->baseCount; count > 0; count--)
				FioStr(fp, (*lp++)->data);

			continue;
		}

		if (mp->flags & F_INSERT_A)
		{
			lp = &fileAlines[mp->fileAline];

			for (count = mp->fileAcount; count > 0; count--)
				FioStr(fp, (*lp++)->data);
		}

		if (mp->flags & F_INSERT_B)
		{
			lp = &fileBlines[mp->fileBline];

			for (count = mp->fileBcount; count > 0; count--)
				FioStr(fp, (*lp++)->data);
		}
	}

	if (conflict)
		FioStr(fp, CONFLICT_END);

	if (!FioClose(fp))
		FileError(fp, "Error writing merged file \"%s\"", outputName);
}


/*
 * Dump out the possible conflicts to a file that indicates what both
 * divergent files did to the base file.  This is useful even if there
 * are no physical conflicts in order to manually find logical conflicts.
 */
void
DumpConflicts(const char * outputName)
{
	MERGE *	mp;		/* current merge structure */
	FIO *	fp;		/* file to output to */
	FIO	outputFile;	/* output file */

	if ((fc.verbosity > VERBOSE_DEFAULT) && outputName)
		printf("[Writing conflict lines to \"%s\"]\n", outputName);

	fp = &outputFile;

	if (!FioOpen(fp, outputName, FIO_SEQWRITE))
		FileError(fp, "Cannot write conflict file \"%s\"", outputName);

	FioFormat(fp, "T %ld conflict%s\n", fm.conflicts,
		(fm.conflicts == 1) ? "" : "s");

	mp = begMergelist;

	while (mp)
	{
		mp = WalkMerges(fp, mp, -1);	/* non-conflicts */

		if (mp == NULL)
			break;

		if (!fm.ignoreFlag)
		{
			FioWrite(fp, "X  ", 3L);

			FioFormat(fp, CONFLICT_BEGIN, mp->baseLine + 1,
				mp->fileAline + 1, mp->fileBline + 1);
		}

		/*
		 * File A conflicts
		 */
		WalkMerges(fp, mp, F_INSERT_A | F_DELETE_A);

		if (!fm.ignoreFlag)
			FioFormat(fp, "X  %s", CONFLICT_MIDDLE);

		/*
		 * File B conflicts
		 */
		mp = WalkMerges(fp, mp, F_INSERT_B | F_DELETE_B);

		if (!fm.ignoreFlag)
			FioFormat(fp, "X  %s", CONFLICT_END);
	}

	if (!FioClose(fp))
	{
		FileError(fp, "Error writing conflict file \"%s\"",
			outputName);
	}
}


/*
 * Walk down and dump out a list of merges, either for non-conflicts,
 * or for either file A or B of a conflict.  Returns the first merge
 * structure not processed due to the change of conflict.
 */
static MERGE *
WalkMerges(
	FIO *	fp,		/* file to output to */
	MERGE *	mp,		/* starting point to dump */
	int	flagMask
)
{
	LINE **	lp;		/* current line */
	int	count;		/* number of lines to handle */
	int	flags;		/* flags to be checked */

	for (;; mp = mp->next)
	{
		if ((mp == NULL) ||
			((flagMask & F_CONFLICT) == (mp->flags & F_CONFLICT)))
		{
			return mp;
		}

		flags = mp->flags;

		if (flags == F_OLD_DATA)
		{
			/*
			 * Show original lines
			 */
			lp = &baseLines[mp->baseLine];

			count = mp->baseCount;

			if (((count - 1) / 2) < fm.unchangeCount)
			{
				while (--count >= 0)
				{
					FioWrite(fp, "   ", 3L);

					FioStr(fp, (*lp++)->data);
				}

				continue;
			}

			/*
			 * Number of unchanged lines exceeds specified limit.
			 * So only show the first few and last few changes.
			 */
			for (count = fm.unchangeCount; count > 0; count--)
			{
				FioWrite(fp, "   ", 3L);

				FioStr(fp, (*lp++)->data);
			}

			FioFormat(fp, "U  %ld\n",
				mp->baseCount - (fm.unchangeCount * 2));

			lp = &baseLines[mp->baseLine + mp->baseCount -
				fm.unchangeCount];

			for (count = fm.unchangeCount; count > 0; count--)
			{
				FioWrite(fp, "   ", 3L);

				FioStr(fp, (*lp++)->data);
			}

			continue;
		}

		flags &= flagMask;

		if (flags & F_INSERT_A)
		{
			lp = &fileAlines[mp->fileAline];

			for (count = mp->fileAcount; count > 0; count--)
			{
				FioWrite(fp, "IA ", 3L);

				FioStr(fp, (*lp++)->data);
			}
		}

		if (flags & F_DELETE_A)
		{
			lp = &baseLines[mp->baseLine];

			for (count = mp->baseCount; count > 0; count--)
			{
				FioWrite(fp, "DA ", 3L);

				FioStr(fp, (*lp++)->data);
			}
		}

		if (flags & F_INSERT_B)
		{
			lp = &fileBlines[mp->fileBline];

			for (count = mp->fileBcount; count > 0; count--)
			{
				FioWrite(fp, "IB ", 3L);

				FioStr(fp, (*lp++)->data);
			}
		}

		if (flags & F_DELETE_B)
		{
			lp = &baseLines[mp->baseLine];

			for (count = mp->baseCount; count > 0; count--)
			{
				FioWrite(fp, "DB ", 3L);

				FioStr(fp, (*lp++)->data);
			}
		}
	}
}


/*
 * Allocate a new merge structure and append it to the end of the merge list.
 */
static MERGE *
NewMerge()
{
	MERGE *	mp;		/* new merge structure */

	mp = (MERGE *) malloc(sizeof(MERGE));

	if (mp == NULL)
	{
		fprintf(stderr, "No memory\n");
		exit(1);
	}

	mp->next = NULL;
	mp->flags = 0;
	mp->baseLine = 0;
	mp->fileAline = 0;
	mp->fileBline = 0;
	mp->baseCount = 0;
	mp->fileAcount = 0;
	mp->fileBcount = 0;

	if (begMergelist == NULL)
		begMergelist = mp;
	else
		endMergelist->next = mp;

	endMergelist = mp;

	return mp;
}


/*
 * Read in a conflict file and produce an output file.
 */
void
ConvertConflicts(
	const char *	inputname,	/* conflict file being read */
	const char *	outputName	/* output file (or NULL for tty) */
)
{
	const char *	cp;		/* current line */
	FIO *		ip;		/* input file */
	FIO *		op;		/* output file */
	FIO		inputFile;	/* input file structure */
	FIO		outputFile;	/* output file structure */
	long		inputLines;	/* input and output line counts */
	long		outputLines;
	long		retLen;		/* length of current line */

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf(
		"[Converting conflict file \"%s\" to output file \"%s\"]\n",
			inputname, outputName);
	}

	inputLines = 0;
	outputLines = 0;

	ip = &inputFile;
	op = &outputFile;

	if (!FioOpen(ip, inputname, FIO_READ))
		FileError(ip, "Cannot read conflict file \"%s\"", inputname);

	if (!FioOpen(op, outputName, FIO_SEQWRITE))
		FileError(op, "Cannot write output file \"%s\"", outputName);

	while (TRUE)
	{
		cp = FioReadline(ip, &retLen, FALSE);

		if (cp == NULL)
			break;

		if (((++inputLines % 1000) == 0) &&
			(fc.verbosity > VERBOSE_DEFAULT))
		{
			printf("[%ld lines]\n", inputLines);
		}

		switch (*cp)
		{
		case 'I':		/* inserted lines */
		case ' ':		/* original base lines */
			cp += 3;
			retLen -= 3;

			if (retLen <= 0)
			{
				cp = "\n";
				retLen = 1;
			}

			FioWrite(op, cp, retLen);
			outputLines++;
			break;

		case '\n':		/* original blank line */
			FioWrite(op, "\n", 1L);
			break;

		case 'D':		/* deleted lines */
		case 'T':		/* the total line */
			break;

		case 'X':
			if (fm.ignoreFlag)
				break;

			Error(
		"Unresolved conflict remaining at line %ld in file \"%s\"",
				inputLines, inputname);

		case 'U':
			Error(
		"File \"%s\" is incomplete due to -u option being used",
				inputname);

		default:
			Error("Unknown line %ld read from conflict file \"%s\"",
				inputLines, inputname);
		}
	}

	if (!FioClose(ip))
		FileError(ip, "Error reading \"%s\"", inputname);

	if (!FioClose(op))
		FileError(op, "Error writing \"%s\"", outputName);

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf("[%ld lines read, %ld lines written]\n",
			inputLines, outputLines);
	}
}

/* END CODE */
