/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Definitions common to various parts of source control program
 */

#ifndef	SC_H
#define	SC_H

#include <limits.h>
#include "bool.h"
#include "fio.h"

#ifndef	PATH_MAX
#define	PATH_MAX	1024		/* maximum size of path name */
#endif

#define	MODULE_MAX	1024		/* longest module name */
#define	MAX_MODULES	10000		/* maximum modules in one command */
#define	TMP_PATH	"/tmp"		/* temp directory */


#define	MAX_INFO	80		/* maximum size of information line */
#define	MODIFY_LINES	25		/* default lines for modifying */
#define	EDIT_ALLOC_SIZE	100		/* edit structures to allocate */
#define	HEADER_FORMAT	"%c %06ld %06ld %06ld %012ld\n"	/* header format */
#define	HEADER_LINE_LENGTH	36	/* length of the header line */
#define	DEFAULT_PATH	"FSC_EDITS"	/* default pathname for sources */
#define	OLDEST_NAME	"oldest"	/* edit name for oldest version */
#define	NEWEST_NAME	"newest"	/* edit name for newest version */
#define	NO_SEEK		((long) -1)	/* value to indicate no seeking */


/*
 * Filename extensions for files in the edit history directory.
 */
#define	EXT_SOURCE	".s"	/* latest source file */
#define	EXT_HISTORY	".e"	/* edit history file */
#define	EXT_NEW		".n"	/* temporary new file */
#define	EXT_OLD		".o"	/* temporary old file */


/*
 * Types of lines in the edit history file.
 * These letters appear as the first character of each line to identify its
 * type.  They are followed by a space if more data is present on the line.
 */
#define T_HEADER	'H'	/* header line */
#define T_BEGIN_EDIT	'B'	/* beginning of edit line */
#define T_REMARK	'R'	/* remark line (informational only) */
#define T_CHANGE	'C'	/* change lines */
#define T_TEXT		'T'	/* text being inserted line */
#define T_END_EDIT	'E'	/* end of edit line */
#define T_FILE		'F'	/* read full edit from file */
#define T_POSITION	'P'	/* position of edit line */
#define T_EOF		'Z'	/* end of file line */
#define T_NONE		-1	/* value to indicate don't check type */


/*
 * The offset into a TEXT line of the actual data for the line.
 */
#define	TEXT_DATA_OFFSET	2


/*
 * A flag character on the T_CHANGE line to indicate that the last line
 * of text in the section does not end in a newline character.
 */
#define	EOL_MISSING_FLAG	'E'


/*
 * Structure to remember the position table information
 */
typedef struct
{
	long		p_pos;		/* position of beginning of edit */
	const char *	p_names;	/* names for this edit (or NULL) */
} POS;


typedef struct
{
	FIO *		historyFile;	/* edit history file */
	FIO *		sourceFile;	/* source file being edited */
	FIO *		outputFile;	/* output file */
	const char *	remarkName;	/* name of remarks file (or NULL) */
	const char *	dirName;	/* path for edit history directory */
	const char *	moduleName;	/* module name being manipulated */
	char *		modifyBuffer;	/* buffer for modified lines */
	long		firstEdit;	/* first edit number in history file */
	long		lastEdit;	/* last edit number in history file */
	long		lastFile;	/* last file number used */
	long		tablePos;	/* offset of position table */
	long		lineLen;	/* length of line */
	long		modifyLines;	/* number of lines to modify */
	BOOL		forceUpdateFlag; /* don't ask update questions */
	BOOL		forceWriteFlag;	/* don't ask overwrite questions */
	BOOL		noWriteFlag;	/* don't overwrite existing files */
	char		baseName[PATH_MAX];	/* base name of history file */
	char		historyName[PATH_MAX];	/* full path of history file */
	char		sourceName[PATH_MAX];	/* full path of source file */
} SC_DATA;

extern SC_DATA sc;		/* all external data for sc */


/*
 * Functions
 */
extern	POS *		ReadPosTable(FIO *);
extern	const char *	GetNumber(const char *, long *);
extern	const char *	GetLine(FIO *, long, int);
extern	void		FreeLine(const char *);
extern	long		FindEditNumber(const char *);
extern	long		FindEdit(const char *);
extern	void		OpenSourceFile(void);
extern	void		OpenHistoryFile(int);
extern	void		CreateHistory(const char *, const char *);
extern	void		ExtractHistory(const char *, const char *, BOOL);
extern	void		DiffHistory(const char *, const char *,
				const char *, const char *, BOOL);
extern	void		UpdateHistory(const char *, const char *);
extern	void		CompHistory(const char *);
extern	void		ListHistory(const char *, const char *, const char *);
extern	void		NameHistory(const char *);
extern	void		CleanHistory(const char *);
extern	void		CheckHistory(const char *);
extern	void		PruneHistory(const char *);
extern	void		StartEdit(FIO *, long, char *);
extern	void		CheckEditName(const char *);
extern	void		CreateSubDirectories(const char *, const char *);
extern	BOOL		RenameFiles(const char *);
extern	BOOL		IsWriteAllowed(const char *);
extern	BOOL		Confirm(void);
extern	BOOL		IsDir(const char *);
extern	void		EnableSignals(void);
extern	void		DisableSignals(void);
extern	void		EditReset(void);
extern	void		Restart(BOOL);
extern	void		Error(const char *, ...);
extern	void		FileError(const FIO *, const char *, ...);

#endif

/* END CODE */
