/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to compare different versions of a module
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"
#include "cmalloc.h"


static	int	QuickComp(const char *, const char *, long);


/*
 * Compare two different versions of a module.
 * Count specifies how many edit numbers were supplied to the command.
 * If zero, then we compare inputName to the latest edit.  If one, then
 * we compare inputName to the specified edit number.  If two, then we
 * compare the specified edit numbers.  If matchBlab is TRUE, then we
 * output a message even if the files are identical.
 */
void
DiffHistory(
	const char *	inputName,	/* file to be compared (or NULL) */
	const char *	outputName,	/* output file (or NULL for tty) */
	const char *	editName1,	/* edit name of first file (or NULL) */
	const char *	editName2,	/* edit name of second file (or NULL) */
	BOOL		matchBlab
)
{
	const char *	temp1;			/* temporary file names */
	const char *	temp2;
	long		edit1;			/* first edit number */
	long		edit2;			/* second edit number */
	long		i;			/* temporary */
	char *		num;			/* pointer to number */
	pid_t		pid;			/* process id */
	BOOL		savedForceFlag;		/* saved force flag */
	char		nameBuf1[PATH_MAX];	/* first temporary name */
	char		nameBuf2[PATH_MAX];	/* second temporary name */
	char		numBuf[12];		/* buffer for number string */

	fc.maxChanges = INFINITY;
	fc.fileAmodify = sc.modifyLines;
	fc.fileBmodify = sc.modifyLines;

	pid = getpid();

	temp1 = getenv("TMPDIR");

	if ((temp1 == NULL) || (*temp1 == '\0'))
		temp1 = TMP_PATH;

	sprintf(nameBuf1, "%s/SC1_%d", temp1, pid);
	sprintf(nameBuf2, "%s/SC2_%d", temp1, pid);

	OpenHistoryFile(FIO_READ);

	if (editName1 == NULL)
	{
		/*
		 * Compare with latest version
		 */
		FioClose(sc.historyFile);

		if (fc.verbosity > VERBOSE_DEFAULT)
		{
			printf("[Comparing files \"%s\" and \"%s\"]\n",
				sc.sourceName, inputName);
		}

		fc.fileAskip = 1;

		Fcomp(sc.sourceName, inputName);

		if (fc.whatFlag)
			DumpWhat(outputName);
		else if (matchBlab || fc.inserts || fc.deletes)
			DumpNormal(outputName);

		return;
	}

	/*
	 * Here when we have to retrieve a previous version.
	 */
	temp1 = NULL;
	temp2 = NULL;

	savedForceFlag = sc.forceWriteFlag;
	sc.forceWriteFlag = TRUE;

	edit1 = FindEditNumber(editName1);

	if (edit1 != sc.lastEdit)
		temp1 = nameBuf1;

	edit2 = edit1;

	if (editName2)
	{
		edit2 = FindEditNumber(editName2);

		if (edit2 != sc.lastEdit)
			temp2 = nameBuf2;
	}

	FioClose(sc.historyFile);

	if (temp1)
	{
		num = &numBuf[11];
		*num = '\0';

		i = edit1;

		do
		{
			*(--num) = (i % 10) + '0';
			i /= 10;
		} while (i);

		ExtractHistory(num, temp1, VERBOSE_NONE);
	} else
		fc.fileAskip = 1;

	if (editName2 == NULL)
	{
		sc.forceWriteFlag = savedForceFlag;

		Fcomp(temp1 ? temp1 : sc.sourceName, inputName);

		if (temp1)
		{
			if (fc.verbosity > VERBOSE_DEFAULT)
			{
				printf("[Deleting temporary file \"%s\"]\n",
					temp1);
			}

			(void) unlink(temp1);
		}

		if (fc.whatFlag)
			DumpWhat(outputName);
		else if (matchBlab || fc.inserts || fc.deletes)
			DumpNormal(outputName);

		return;
	}

	/*
	 * Here if we have to compare against another edit
	 */
	if (temp2)
	{
		num = &numBuf[11];
		*num = '\0';

		i = edit2;

		do {
			*(--num) = (i % 10) + '0';
			i /= 10;
		} while (i);

		ExtractHistory(num, temp2, VERBOSE_NONE);
	} else
		fc.fileBskip = 1;

	sc.forceWriteFlag = savedForceFlag;

	Fcomp(temp1 ? temp1 : sc.sourceName, temp2 ? temp2 : sc.sourceName);

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		if (temp1 && temp2)
		{
			printf("[Deleting temp files \"%s\" and \"%s\"]\n",
				temp1, temp2);
		}
		else if (temp1 || temp2)
		{
			printf("[Deleting temporary file \"%s\"]\n",
				temp1 ? temp1 : temp2);
		}
	}

	if (temp1)
		(void) unlink(temp1);

	if (temp2)
		(void) unlink(temp2);

	if (fc.whatFlag)
		DumpWhat(outputName);
	else if (matchBlab || fc.inserts || fc.deletes)
		DumpNormal(outputName);
}


/*
 * See if a file is up to date, and if so, then delete it.
 * This is useful when cleaning up a directory.
 */
void
CleanHistory(const char * inputName)
{
	OpenHistoryFile(FIO_READ);

	FioClose(sc.historyFile);

	if (access(inputName, F_OK) && (errno = ENOENT))
	{
		if (fc.verbosity > VERBOSE_DEFAULT)
		{
			printf("File \"%s\" doesn't exist for module \"%s\"\n",
				inputName, sc.moduleName);
		}

		return;
	}

	if (QuickComp(sc.sourceName, inputName, 1L))
	{
		if (fc.verbosity > VERBOSE_DEFAULT)
		{
			printf(
		"File \"%s\" differs from latest edit of module \"%s\"\n",
				inputName, sc.moduleName);
		}

		return;
	}

	if (unlink(inputName))
	{
		FileError(FIO_NULL,
			"Cannot delete file \"%s\" for module \"%s\"",
			inputName, sc.moduleName);
	}

	if (fc.verbosity)
		printf("Deleted up to date file \"%s\"\n", inputName);
}


/*
 * Check to see if a file is up to date, and report it if not.
 */
void
CheckHistory(const char * inputName)
{
	OpenHistoryFile(FIO_READ);

	FioClose(sc.historyFile);

	if ((access(inputName, F_OK) < 0) && (errno == ENOENT))
	{
		if (fc.verbosity > VERBOSE_DEFAULT)
		{
			printf("File \"%s\" does not exist for module \"%s\"\n",
				inputName, sc.moduleName);
		}

		return;
	}

	if (QuickComp(sc.sourceName, inputName, 1L))
	{
		if (fc.quickFlag)
			printf("%s\n", sc.moduleName);
		else
		{
			printf(
		"File \"%s\" differs from latest edit of module \"%s\"\n",
				inputName, sc.moduleName);
		}

		return;
	}

	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[File \"%s\" is up to date]\n", inputName);
}


/*
 * Routine to do a quick comparison of two files.
 * Returns nonzero if the two files are different.
 */
static int
QuickComp(
	const char *	nameA,		/* file names to compare */
	const char *	nameB,
	long		skipA		/* lines to skip for file A */
)
{
	FIO *		fpA;		/* first file */
	FIO *		fpB;		/* second file */
	char *		cp;		/* current temporary line */
	const char *	cpA;		/* current line for file A */
	const char *	cpB;		/* current line for file B */
	long		lineNumber;	/* current line number */
	long		cpLen;		/* length of current temporary line */
	long		lineLenA;	/* length of first file line */
	long		lineLenB;	/* length of second file line */
	FIO		fileA;		/* file structure for first file */
	FIO		fileB;		/* file structure for second file */

	fpA = &fileA;
	fpB = &fileB;

	if (!FioOpen(fpA, nameA, FIO_READ))
		FileError(fpA, "Cannot read \"%s\"", nameA);

	if (!FioOpen(fpB, nameB, FIO_READ))
		FileError(fpB, "Cannot read \"%s\"", nameB);

	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Comparing files \"%s\" and \"%s\"]\n", nameA, nameB);

	cpLen = 132;

	cp = malloc(cpLen + 1);

	if (cp == NULL)
		FileError(FIO_NULL, "Cannot allocate memory for line");

	if (skipA > 0)
		FioSkip(fpA, skipA);

	lineNumber = 0;

	while (TRUE)
	{
		cpA = FioReadline(fpA, &lineLenA, FALSE);

		if (cpA == NULL)
		{
			cpB = FioReadline(fpB, &lineLenB, FALSE);

			break;
		}

		if (lineLenA > cpLen)
		{
			cpLen = lineLenA;

			cp = realloc(cp, cpLen + 1);

			if (cp == NULL)
			{
				FileError(FIO_NULL,
					"Cannot reallocate memory for line");
			}
		}

		strcpy(cp, cpA);

		cpA = cp;

		cpB = FioReadline(fpB, &lineLenB, FALSE);

		if (cpB == NULL)
			break;

		if (++lineNumber <= sc.modifyLines)
		{
			cpA = ModifyLine(cpA, &lineLenA, (INFO *) NULL);
			cpB = ModifyLine(cpB, &lineLenB, (INFO *) NULL);
		}

		if ((lineLenA != lineLenB) || strcmp(cpA, cpB))
			break;
	}

	if (!FioClose(fpA))
		FileError(fpA, "Read error on \"%s\"", nameA);

	if (!FioClose(fpB))
		FileError(fpB, "Read error on \"%s\"", nameB);

	free(cp);

	return (cpA || cpB);
}


/*
 * Compare a file to the newest version of a module, but output nothing.
 * Whether or not there are any differences is returned in the variables
 * inserts and deletes.  The edit list is built so that it can be
 * immediately used to perform an update.
 */
void
CompHistory(const char * inputName)
{
	OpenHistoryFile(FIO_READ);

	FioClose(sc.historyFile);

	fc.maxChanges = INFINITY;
	fc.fileAmodify = sc.modifyLines;
	fc.fileBmodify = sc.modifyLines;
	fc.fileAskip = 1;

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf("[Comparing files \"%s\" and \"%s\"]\n",
			sc.sourceName, inputName);
	}

	Fcomp(sc.sourceName, inputName);
}

/* END CODE */
