/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to extract versions of a module.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "fio.h"
#include "fcomp.h"
#include "sc.h"
#include "cmalloc.h"


/*
 * States for examining history file.
 */
#define	S_OUTSIDE	0	/* outside of an edit */
#define	S_INSIDE	1	/* just inside of an edit */
#define	S_CHANGE	2	/* saw changes in an edit */
#define	S_FILE		3	/* saw file reference in an edit */


/*
 * In-memory edit history structure.
 */
typedef struct edit EDIT;

struct edit
{
	EDIT *		next;		/* next edit history structure */
	EDIT *		prev;		/* previous edit history structure */
	long		count;		/* line count */
	const char *	text;		/* text line if any */
	long		textLen;	/* length of text line */
};


/*
 * Local data.
 */
static	EDIT *	editFreeList;	/* free list for edit structures */
static	EDIT *	editNewBegin;	/* pointer to chunk of new edit structures */
static	EDIT *	editNewEnd;	/* pointer to end of new edit structures */
static	EDIT	editHead;	/* head of edit list */
static	EDIT	editTail;	/* tail of edit list */
static	EDIT	newEditHead;	/* header of new edit list */
static	EDIT	newEditTail;	/* tail of new edit list */
static	long	curLine;	/* current line number */
static	char	fromSource;	/* dummy addr to indicate source file lines */


/*
 * Local routines.
 */
static	EDIT *	AllocEdit(EDIT *);
static	EDIT *	RemEdit(EDIT *);
static	void	ReadEdits(long);
static	void	InitList(EDIT *, EDIT *);
static	void	AddToList(FIO *, long, long, long, BOOL);
static	void	MergeLists(void);
static	void	WriteFile(const char *, const INFO *);
static	void	CopyFile(const char *, const INFO *);
static	void	PrintList(const char *, const EDIT *);
static	void	PrintOne(const char *, const EDIT *);


/*
 * Extract a revision of a module using the edit history.
 */
void
ExtractHistory(
	const char *	editName,	/* edit string (NULL for current) */
	const char *	outputName,	/* output file (NULL if tty) */
	BOOL		blabFlag	/* describe file being extracted */
)
{
	char *		str;		/* string being manipulated */
	const char *	msg;		/* part of message */
	long		editNumber;	/* edit number being extracted */
	long		i;		/* temporary value */
	INFO		info;		/* information about edit */
	char		infoLine[MAX_INFO];	/* first remark line */
	char		editStr[16];	/* edit number string */

	OpenHistoryFile(FIO_READ);

	editNumber = FindEditNumber(editName);

	/*
	 * If there is an output file, then check to see if we are overwriting
	 * an old file, and let the user know what is being extracted.
	 */
	if (outputName)
	{
		if (sc.noWriteFlag && (access(outputName, F_OK) == 0))
		{
			if (fc.verbosity)
			{
				printf(
		"Existing file \"%s\" not overwritten by module \"%s\"\n",
					outputName, sc.moduleName);
			}

			FioClose(sc.historyFile);

			return;
		}

		msg = "";

		if (editNumber == sc.firstEdit)
			msg = " initial";

		if (editNumber == sc.lastEdit)
			msg = " latest";

		if (fc.verbosity || blabFlag)
		{
			printf("%sExtracting%s edit %ld of module \"%s\"",
				(blabFlag ? "" : "["), msg, editNumber,
				sc.moduleName);

			if (strcmp(outputName, sc.moduleName))
				printf(" to file \"%s\"", outputName);

			printf("%s\n", (blabFlag ? "" : "]"));
		}
	}

	if (!IsWriteAllowed(outputName))
	{
		FioClose(sc.historyFile);

		return;
	}

	StartEdit(sc.historyFile, editNumber, infoLine);

	/*
	 * Fill in the information structure for line modifying to use.
	 */
	str = infoLine;

	while (*str == ' ')
		str++;

	info.user = str;

	while (*str && (*str != ' '))
		str++;

	*str++ = '\0';

	while (*str == ' ')
		str++;

	/*
	 * Parse and modify the date:
	 * Incoming:  Sun Sep 16 01:03:52 1998\n
	 * Outgoing:  16-Sep-1998
	 */
	str[1] = str[8];
	str[2] = str[9];
	str[3] = '-';
	str[7] = '-';
	str[8] = str[20];
	str[9] = str[21];
	str[10] = str[22];
	str[11] = str[23];
	str[12] = '\0';

	if (*++str == ' ')
		str++;

	info.date = str;

	str = &editStr[15];
	*str = '\0';

	i = editNumber;

	do
	{
		*(--str) = (i % 10) + '0';
		i /= 10;
	} while (i);

	info.edit = str;

	/*
	 * If we are extracting the latest edit, then just copy it from
	 * the source file.  Otherwise, do the work of applying edits
	 * from the history file.
	 */
	if (editNumber == sc.lastEdit)
	{
		FioClose(sc.historyFile);

		CopyFile(outputName, &info);

		return;
	}

	ReadEdits(editNumber);

	WriteFile(outputName, &info);
}


/*
 * Routine to collect edit scripts from the log file for a given edit.
 * This builds an in-memory edit list which can then be used to generate
 * the specified version of the source file.  The history file has already
 * been opened, the edit number checked, and the first line of the desired
 * edit sequence read and checked.
 */
static void
ReadEdits(long editNumber)
{
	FIO *		fp;		/* history file */
	const char *	cp;		/* line of data */
	long		fileNumber;	/* file number containing whole edit */
	long		lineNumber;
	long		insertCount;
	long		deleteCount;
	long		inscc;
	long		delcc;
	long		linecc;
	long		temp;
	BOOL		more;
	BOOL		eolMissing;
	int		state;		/* state of current edit */

	fp = sc.historyFile;

	insertCount = 0;
	deleteCount = 0;

	state = S_INSIDE;
	more = TRUE;
	lineNumber = 1;

	InitList(&editHead, &editTail);

	InitList(&newEditHead, &newEditTail);

	/*
	 * Here for each line of the edit history file.
	 * Switch on the type of each line, verifying that the
	 * sequence of types is legitimate.  Convert each change
	 * line into one or more edit structures.  After each
	 * set of changes for an edit has been read, then merge
	 * together the old and new edit lists to form the final
	 * edit list.
	 */
	while (more)
	{
		cp = FioReadline(fp, (long *) NULL, FALSE);

		if (cp == NULL)
			Error("Error reading edit");

		switch (*cp++)
		{
		case T_BEGIN_EDIT:
			/*
			 * Begin another edit.
			 */
			cp = GetNumber(cp, &temp);

			if (cp == NULL)
				Error("no number in begin edit");

			if (state != S_OUTSIDE)
			{
				Error(
			"Beginning edit %ld without ending previous one", temp);
			}

			state = S_INSIDE;

			if (temp != editNumber + 1)
				Error("Non-consecutive begin edit");

			editNumber = temp;
			insertCount = 0;
			deleteCount = 0;
			lineNumber = 1;

			InitList(&newEditHead, &newEditTail);

			break;


		case T_CHANGE:
			/*
			 * Change some lines.
			 */
			if (state == S_OUTSIDE)
				Error("Change line outside of edit");

			if ((state != S_INSIDE) && (state != S_CHANGE))
				Error("Change line not expected in edit");

			state = S_CHANGE;

			cp = GetNumber(cp, &linecc);

			if (cp == NULL)
				Error("Missing line number in change line");

			if (linecc < lineNumber)
				Error("Decreasing line number in change line");

			cp = GetNumber(cp, &inscc);
			cp = GetNumber(cp, &delcc);

			if (cp == NULL)
				Error("Bad line counts in change line");

			if ((inscc == 0) && (delcc == 0))
				Error("Change line does nothing");

			while (*cp == ' ')
				cp++;

			eolMissing = FALSE;

			while (*cp && (*cp != ' '))
			{
				if (*cp == EOL_MISSING_FLAG)
					eolMissing = TRUE;

				cp++;
			}

			AddToList(fp, linecc, inscc, delcc, eolMissing);

			insertCount += inscc;
			deleteCount += delcc;

			if (delcc == 0)
				delcc = 1;

			lineNumber = linecc + delcc;

			break;


		case T_TEXT:
			/*
			 * Text line.
			 */
			if (state != S_OUTSIDE)
				Error("Text line not within edit");

			Error("Unexpected text line");
			break;


		case T_FILE:
			/*
			 * Get results from explicit file.
			 */
			if (state == S_OUTSIDE)
				Error("File line not inside edit");

			if (state != S_INSIDE)
				Error("File line not expected in edit");

			state = S_FILE;

			cp = GetNumber(cp, &fileNumber);

			if (cp == NULL)
				Error("Missing file number in file line");

			/* NEEDS MORE WORK */
			break;


		case T_END_EDIT:
			/*
			 * End of current edit.
			 */
			if (state == S_OUTSIDE)
				Error("Ending edit not yet started");

			state = S_OUTSIDE;

			cp = GetNumber(cp, &temp);

			if (cp == NULL)
				Error("No edit number at end of edit");

			if (temp != editNumber)
			{
				Error("End of edit %ld not correct number",
					editNumber);
			}

			cp = GetNumber(cp, &inscc);
			cp = GetNumber(cp, &delcc);

			if (cp == NULL)
				Error("Bad total counts in end of edit line");

			if (inscc != insertCount)
				Error("Wrong insert count at end of edit");

			if (delcc != deleteCount)
				Error("Wrong delete count at end of edit");

			more = (editNumber != sc.lastEdit);

			MergeLists();

			break;


		case T_REMARK:
			/*
			 * Remark line.
			 */
			if (state == S_OUTSIDE)
				Error("Remark line outside of edit");
			break;


		case T_EOF:
			Error("Unexpected end of file line");


		default:
			Error("Unexpected line in edit");
		}
	}

	FioClose(fp);
}


/*
 * Initialize an edit list.
 */
static void
InitList(
	EDIT *	hp,		/* head of list */
	EDIT *	tp		/* tail of list */
)
{
	hp->count = 0;
	hp->textLen = 0;
	hp->text = NULL;
	hp->prev = NULL;
	hp->next = tp;
	tp->count = 0x40000000;		/* large count of lines */
	tp->text = &fromSource;		/* all from the source file */
	tp->textLen = 0;
	tp->next = NULL;
	tp->prev = hp;

	curLine = 1;
}


/*
 * Append a change to the new edit list.
 * The change has the specified number of inserted and deleted lines.
 * The inserted lines are read from the current edit history position.
 * If the eolMissing flag is set, then the last of the inserted lines
 * has its terminating newline removed.
 */
static void
AddToList(
	FIO *	fp,			/* edit history file */
	long	line,
	long	insertCount,
	long	deleteCount,
	BOOL	eolMissing
)
{
	EDIT *		ep;
	const char *	text;
	char *		cp;
	int		len;

	if (curLine < line)
	{
		/*
		 * Edit for reading data from source
		 */
		ep = AllocEdit(&newEditTail);

		ep->text = &fromSource;
		ep->count = line - curLine;
		curLine = line;
	}

	if (deleteCount > 0)
	{
		/*
		 * Edit for deleting from source
		 */
		ep = AllocEdit(&newEditTail);

		ep->count = deleteCount;
		curLine += deleteCount;
	}

	while (insertCount-- > 0)
	{
		/*
		 * Edits for inserting new text.
		 * Read the next line of text.
		 */
		text = GetLine(fp, NO_SEEK, T_TEXT);

		if (text == NULL)
			Error("Expected text line missing");

		len = sc.lineLen;

		/*
		 * If this is the last line of the inserts which is
		 * specified to be missing a newline, then remove it.
		 */
		if (eolMissing && (insertCount == 0) && (len > 0) &&
			(text[len - 1] == '\n'))
		{
			cp = malloc(len);

			if (cp == NULL)
				Error("Memory allocation failed");

			memcpy(cp, text, len);

			len--;
			cp[len] = '\0';

			text = cp;
		}

		ep = AllocEdit(&newEditTail);

		ep->text = text;
		ep->textLen = len;
		ep->count = 1;
	}
}


/*
 * Routine to merge the current and new edit lists together to form
 * the new current edit list containing both changes.  This works by
 * conceptually walking through each 'line' and determining the result
 * of interactions of deleting, inserting, and copying from the source.
 */
static void
MergeLists()
{
	EDIT *	op;		/* pointer to current old edit structure */
	EDIT *	np;		/* pointer to current new edit structure */
	long	count;		/* number of lines in common */
	int	selectValue;	/* select value */

	op = editHead.next;
	np = newEditHead.next;

	if (fc.debugFlag)
	{
		PrintList("old list", op);
		PrintList("new list", np);
		printf("\nMERGES:\n");
	}

	while (np != &newEditTail)
	{
		count = np->count;

		if (op->count < count)
			count = op->count;

		selectValue = 0;

		if (np->text == NULL)
			selectValue = 1;

		if (np->text == &fromSource)
			selectValue = 2;

		if (op->text == NULL)
			selectValue += 3;

		if (op->text == &fromSource)
			selectValue += 6;

		if (fc.debugFlag)
		{
			printf("case %d: ", selectValue);
			PrintOne("op:", op);
			PrintOne("np:", np);
		}

		switch (selectValue)
		{
		case 0:			/* old inserts and new inserts */
		case 2:			/* old inserts and new from source */
			if (fc.debugFlag)
				PrintOne("out:", op);

			op = op->next;
			break;

		case 1:			/* old inserts and new deletes */
		case 4:			/* old and new both delete */
		case 7:			/* old from source and new deletes */
			op = AllocEdit(op);

			op->count = np->count;

			if (fc.debugFlag)
				PrintOne("out:", op);

			op = op->next;
			np = np->next;

			break;

		case 3:			/* old deletes and new inserts */
			np->count -= count;

			if (np->count <= 0)
			{
				FreeLine(np->text);
				np = np->next;
			}

			op->count--;

			if (op->count <= 0)
				op = RemEdit(op);

			break;

		case 5:			/* old deletes and new from source */
			np->count -= count;

			if (np->count <= 0)
				np = np->next;

			if (op->count == count)
			{
				if (fc.debugFlag)
					PrintOne("out:", op);

				op = op->next;
				break;
			}

			op = AllocEdit(op);

			op->count = count;

			if (fc.debugFlag)
				PrintOne("out:", op);

			op = op->next;
			op->count -= count;

			break;

		case 6:			/* old from source and new inserts */
			op = AllocEdit(op);

			op->text = np->text;
			op->textLen = np->textLen;
			op->count = 1;

			if (fc.debugFlag)
				PrintOne("out:", op);

			op = op->next;
			np = np->next;

			if (op == &editTail)
				break;

			op->count--;

			if (op->count <= 0)
				op = RemEdit(op);

			break;

		case 8:			/* old and new both from source */
			np->count -= count;

			if (np->count <= 0)
				np = np->next;

			if (op->count <= count)
			{
				if (op != &editTail)
				{
					if (fc.debugFlag)
						PrintOne("out:", op);

					op = op->next;
				}

				break;
			}

			op = AllocEdit(op);

			op->text = &fromSource;
			op->count = count;

			if (fc.debugFlag)
				PrintOne("out:", op);

			op = op->next;

			if (op != &editTail)
				op->count -= count;

			break;
		}

		if (fc.debugFlag)
			printf("\n");
	}

	if (fc.debugFlag)
	{
		PrintList("Updated old list", editHead.next);
		printf("\n");
	}
}


/*
 * Allocate a new edit structure and insert it in front of the
 * specified edit structure in a list.
 */
static EDIT *
AllocEdit(EDIT * ep)
{
	EDIT *	np;		/* new edit structure */

	np = editFreeList;

	if (np)
		editFreeList = np->next;
	else
	{
		if (editNewBegin == editNewEnd)
		{
			np = (EDIT *) malloc(sizeof(EDIT) * EDIT_ALLOC_SIZE);

			if (np == NULL)
				Error("Cannot allocate new edit structures");

			editNewBegin = np;
			editNewEnd = np + EDIT_ALLOC_SIZE;
		}

		np = editNewBegin++;
	}

	np->text = NULL;
	np->textLen = 0;
	np->count = 0;
	np->next = ep;
	np->prev = ep->prev;
	np->prev->next = np;
	ep->prev = np;

	return np;
}


/*
 * Remove an edit from an edit list and put it onto the free list.
 * Returns the next edit structure on the list.
 */
static EDIT *
RemEdit(EDIT *ep)
{
	EDIT *	np;		/* next edit structure in list */

	np = ep->prev;
	np->next = ep->next;
	np = ep->next;
	np->prev = ep->prev;
	ep->next = editFreeList;

	editFreeList = ep;

	return np;
}


/*
 * Write out the extracted file using the information in the edit list
 * and the current source file.
 */
static void
WriteFile(
	const char *	outputName,	/* name of output file */
	const INFO *	info		/* edit information */
)
{
	FIO *		sf;		/* source file */
	FIO *		of;		/* output file */
	EDIT *		ep;		/* current edit */
	long		line;		/* current line in source file */
	long		count;
	const char *	cp;

	line = 0;

	sf = sc.sourceFile;
	of = sc.outputFile;

	OpenSourceFile();

	if (!FioOpen(of, outputName, FIO_SEQWRITE))
		FileError(of, "Cannot open output file \"%s\"", outputName);

	/*
	 * Loop over each of the edit structures applying them to the
	 * source file to produce the output file.
	 */
	for (ep = editHead.next; ep; ep = ep->next)
	{
		count = ep->count;

		/*
		 * If we are deleting lines from the source file,
		 * THEN just skip the required number of lines.
		 */
		if (ep->text == NULL)
		{
			if (!FioSkip(sf, count))
			{
				FileError(sf,
					"Premature end of file in source file");
			}

			continue;
		}

		/*
		 * If we are inserting lines from the history file,
		 * then insert them from the edit structure.
		 */
		if (ep->text != &fromSource)
		{
			if (count != 1)
			{
			Error("Edit command for text has bad count value");
			}

			cp = ep->text;

			sc.lineLen = ep->textLen;

			if (++line <= sc.modifyLines)
				cp = ModifyLine(cp, &sc.lineLen, info);

			if (!FioWrite(of, cp, sc.lineLen))
			{
				FileError(of, "Write to \"%s\" failed",
					outputName);
			}

			continue;
		}

		/*
		 * The final choice is preserving lines from the source file,
		 * so copy the lines from the source file.
		 */
		while (count-- > 0)
		{
			cp = FioReadline(sf, &sc.lineLen, FALSE);

			if (cp == NULL)
			{
				if (ep != &editTail)
				{
					Error(
				"Premature end of file from source file");
				}

				if (!FioFlush(of))
				{
					FileError(of,
						"Write to \"%s\" failed",
						outputName);
				}

				if (!FioClose(of))
				{
					FileError(of,
						"Error closing \"%s\"",
						outputName);
				}

				return;
			}

			if (++line <= sc.modifyLines)
				cp = ModifyLine(cp, &sc.lineLen, info);

			if (!FioWrite(of, cp, sc.lineLen))
			{
				FileError(of, "Write to \"%s\" failed",
					outputName);
			}
		}
	}

	Error("Unexpected end of edit list");
}


/*
 * Copy the data from one file to another.
 * This is used when extracting the latest version of a module.
 */
static void
CopyFile(
	const char *	outputName,	/* name of output file */
	const INFO *	info		/* edit information */
)
{
	FIO *		sf;		/* source file for copying */
	FIO *		of;		/* output file for copying */
	const char *	cp;		/* current line of file */
	long		lineLen;	/* length of line */
	long		line;		/* current line number */

	sf = sc.sourceFile;
	of = sc.outputFile;

	OpenSourceFile();

	if (!FioOpen(of, outputName, FIO_SEQWRITE))
		FileError(of, "Cannot create output file \"%s\"", outputName);

	for (line = 0; line < sc.modifyLines; line++)
	{
		cp = FioReadline(sf, &lineLen, FALSE);

		if (cp == NULL)
			break;

		cp = ModifyLine(cp, &lineLen, info);

		FioWrite(of, cp, lineLen);
	}

	if (!FioCopy(sf, of, -1L))
		FileError(of, "Copy failed to \"%s\"", outputName);

	if (!FioClose(of))
		FileError(of, "Close failed for \"%s\"", outputName);

	if (!FioClose(sf))
		FileError(sf, "Read error for source file");
}


/*
 * Print out an edit list
 */
static void
PrintList(
	const char *	msg,
	const EDIT *	ep
)
{
	const char *	blab;
	long		line;

	printf("\n%s\n", msg);

	line = 1;

	while (ep)
	{
		blab = "insert";

		if (ep->text == NULL)
			blab = "delete";

		if (ep->text == &fromSource)
			blab = "source";

		printf("edit at %p: line %ld: %s %ld\n",
			(char *) ep, line, blab, ep->count);

		if ((ep->text == NULL) || (ep->text == &fromSource))
			line += ep->count;

		ep = ep->next;
	}
}


static void
PrintOne(
	const char *	str,
	const EDIT *	ep
)
{
	const char *	blab;

	blab = "insert";

	if (ep->text == NULL)
		blab = "delete";

	if (ep->text == &fromSource)
		blab = "source";

	printf("  %s at %p: %s %ld", str, (char *) ep, blab, ep->count);
}


/*
 * Reset editing data for new file.
 */
void
EditReset()
{
	editFreeList = NULL;
	editNewBegin = NULL;
	editNewEnd = NULL;
}

/* END CODE */
