/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to return information about edits.
 */

#include <stdio.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"
#include "cmalloc.h"


static	void	ListOne(FIO *, long, const POS *);


/*
 * List out information about the edits.
 * If no edit names are given, then all edits are displayed.
 * If one edit name is given, then just that edit is displayed.
 * If two edit names are given, then the specified range of edits is displayed.
 */
void
ListHistory(
	const char *	editName1,	/* first edit name (NULL if none) */
	const char *	editName2,	/* second edit name (NULL if none) */
	const char *	outputName	/* output file name (NULL if tty) */
)
{
	FIO *		fp;		/* history file */
	FIO *		of;		/* output file */
	const POS *	pp;		/* current edit position entry */
	const POS *	posTable;	/* position table */
	long		begEdit;	/* beginning edit number to display */
	long		endEdit;	/* ending edit number to display */
	int		step;		/* step value for loop */

	fp = sc.historyFile;
	of = sc.outputFile;

	OpenHistoryFile(FIO_READ);

	begEdit = sc.lastEdit;
	endEdit = sc.firstEdit;

	if (editName1)
	{
		begEdit = FindEditNumber(editName1);
		endEdit = begEdit;
	}

	if (editName2)
		endEdit = FindEditNumber(editName2);

	if (!FioOpen(of, outputName, FIO_SEQWRITE))
		FileError(of, "Cannot open output file \"%s\"", outputName);

	posTable = ReadPosTable(fp);

	if (!fc.quickFlag)
		FioFormat(of, "Module \"%s\":\n", sc.moduleName);

	pp = posTable + (sc.lastEdit - begEdit + 1);

	step = 1;

	if (begEdit > endEdit)
		step = -1;

	while (TRUE)
	{
		ListOne(fp, begEdit, pp);

		if (begEdit == endEdit)
			break;

		begEdit += step;

		pp -= step;

		FioWrite(of, "\n", 1L);
	}

	free(posTable);

	if (!FioClose(fp))
	{
		FileError(fp, "Error reading history file \"%s\"",
			sc.historyName);
	}

	if (!FioClose(of))
		FileError(of, "Error writing output file \"%s\"", outputName);
}


/*
 * List information about a single edit.
 * The information is written to the usual output file.
 */
static void
ListOne(
	FIO *		fp,		/* edit history file */
	long		edit,		/* edit number */
	const POS *	pp		/* position structure for edit */
)
{
	FIO *		of;		/* output file */
	const char *	cp;		/* current line */
	long		tempEdit;	/* edit number */

	of = sc.outputFile;

	cp = GetLine(fp, pp->p_pos, T_BEGIN_EDIT);

	if (cp == NULL)
		Error("Position failed to beginning of edit %ld", edit);

	cp = GetNumber(cp, &tempEdit);

	if ((cp == NULL) || (edit != tempEdit))
		Error("Bad number at beginning of edit %ld", edit);

	cp = GetLine(fp, NO_SEEK, T_REMARK);

	if (cp == NULL)
		FileError(fp, "Cannot read identification line");

	/*
	 * If the quick flag was specified, then only list the raw
	 * edit number, otherwise list the number and the names.
	 */
	if (fc.quickFlag)
		FioFormat(of, "%ld\n", edit);
	else
	{
		FioFormat(of, "Edit %ld:   %s", edit, cp);

		if (pp->p_names)
		{
			FioStr(of, "Named: ");
			FioStr(of, pp->p_names);
			FioWrite(of, "\n", 1L);
		}
	}

	while ((cp = GetLine(fp, NO_SEEK, T_REMARK)) != 0)
		FioWrite(of, cp, sc.lineLen);
}

/* END CODE */
