/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Program to perform source control operations for modules.
 */

static char buildDate[] = "[# Date 6-Aug-2000 #]";


/*
 * Version of program.
 */
#define VERSION "10.3"


#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <setjmp.h>
#include <dirent.h>
#include <stdarg.h>
#include <string.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"
#include "cmalloc.h"


/*
 * Actions to be performed for modules
 */
#define A_CREATE	0x0001		/* want to create */
#define A_UPDATE	0x0002		/* want to update */
#define A_EXTRACT	0x0004		/* want to extract */
#define A_LIST		0x0008		/* want to list changes */
#define A_DIFF		0x0010		/* want to list differences */
#define A_TYPE		0x0020		/* want to type output */
#define A_DIFF_UPDATE	0x0040		/* show differences and update */
#define A_NAME		0x0080		/* want to set edit name */
#define A_CLEAN		0x0100		/* clean up to date copy of module */
#define A_COND_UPDATE	0x0200		/* update if there are differences */
#define A_CHECK		0x0400		/* see if a file is up to date */
#define A_PRUNE		0x0800		/* prune old edits from edithistory */


/*
 * Definitions of which actions accept various extra arguments
 */
#define	REMARK_ACTIONS	(A_CREATE | A_UPDATE | A_DIFF_UPDATE | A_COND_UPDATE)
#define	NAME_ACTIONS	(REMARK_ACTIONS | A_NAME)
#define	INPUT_ACTIONS	(REMARK_ACTIONS | A_DIFF | A_CHECK | A_CLEAN)
#define	OUTPUT_ACTIONS	(A_EXTRACT | A_LIST | A_DIFF)


static	int	exitStatus;	/* exit status */
static	jmp_buf	jmpBuf;		/* longjmp buffer */
static	char	msgBuf[8192];	/* error message buffer */
SC_DATA		sc;		/* storage for sc program */


/*
 * Routines
 */
static	void		Usage(void);
static	const char *	GetEditString(int *, const char ***);
static	long		GetValue(int *, const char ***, long);

static	void		DoAllModules(int, const char *, const char *,
				const char *, const char *, const char *);

static	void		AllModulesRecursion(char *, int, int, const char *,
				const char *, const char *, const char *,
				const char *);

static	void		DoModule(int, const char *, const char *,
				const char *, const char *, const char *,
				const char *);


int
main(int argc, const char ** argv)
{
	const char *	outputName;	/* output name for extraction */
	const char *	inputName;	/* input name for update */
	const char *	editName;	/* name for this edit */
	const char *	cp;		/* characters of option strings */
	const char *	editString;	/* edit number string */
	const char *	editString2;	/* second edit number string */
	int		moduleCount;	/* number of modules */
	int		curMod;		/* current module number */
	int		action;		/* action to be performed */
	int		lineLength;	/* line length */
	BOOL		stdinFlag;	/* TRUE if modules come from stdin */
	BOOL		allFlag;	/* TRUE if want to do all modules */
	FIO		historyFile;	/* file structure for history file */
	FIO		sourceFile;	/* file structure for source file */
	FIO		outputFile;	/* file structure for output file */
	const char *	modules[MAX_MODULES];	/* list of modules */
	char		moduleName[PATH_MAX];	/* single module name */

	if (argc <= 1)
	{
		Usage();

		return 1;
	}

	if (setjmp(jmpBuf))
		return 1;

	fc.debugFlag = FALSE;
	fc.quickFlag = FALSE;
	fc.verbosity = VERBOSE_DEFAULT;

	sc.noWriteFlag = FALSE;
	sc.forceWriteFlag = FALSE;
	sc.forceUpdateFlag = FALSE;
	sc.modifyLines = MODIFY_LINES;
	sc.firstEdit = 0;
	sc.lastEdit = 0;
	sc.tablePos = 0;
	sc.dirName = DEFAULT_PATH;
	sc.moduleName = NULL;
	sc.remarkName = NULL;

	stdinFlag = FALSE;
	allFlag = FALSE;

	moduleCount = 0;
	action = 0;

	outputName = NULL;
	inputName = NULL;
	editName = NULL;
	editString = NULL;
	editString2 = NULL;

	sc.historyFile = &historyFile;
	sc.sourceFile = &sourceFile;
	sc.outputFile = &outputFile;

	for (++argv; --argc > 0; argv++)
	{
		cp = *argv;

		if (*cp != '-')
		{
			if (moduleCount >= MAX_MODULES)
				Error("Too many modules specified");

			modules[moduleCount++] = cp;

			continue;
		}

		cp = &argv[0][1];

		/*
		 * If the option is a single character one, then
		 * do a switch on it.
		 */
		if (cp[1] == '\0')
		{
			switch (*cp)
			{
			case 'd':	/* generate differences */
				action |= A_DIFF;
				editString = GetEditString(&argc, &argv);
				editString2 = GetEditString(&argc, &argv);
				break;

			case 'e':	/* extract edit */
				action |= A_EXTRACT;
				editString = GetEditString(&argc, &argv);
				break;

			case 'h':	/* give help */
				Usage();
				exit(0);

			case 'i':	/* input file */
				argv++;

				if ((--argc <= 0) || (**argv == '-'))
					Error("No input file supplied");

				inputName = *argv;
				break;

			case 'l':	/* list edit */
				action |= A_LIST;
				editString = GetEditString(&argc, &argv);
				editString2 = GetEditString(&argc, &argv);
				break;

			case 'm':	/* set number of modified lines */
				sc.modifyLines = GetValue(&argc, &argv,
					(long) MODIFY_LINES);

				break;

			case 'o':	/* output file */
				argv++;

				if ((--argc <= 0) || (**argv == '-'))
					Error("No output file supplied");

				outputName = *argv;
				break;

			case 'p':	/* directory path for edit histories */
				argv++;

				if ((--argc <= 0) || (**argv == '-'))
					Error("No directory path supplied");

				sc.dirName = *argv;
				break;

			case 'q':	/* be quick */
				fc.quickFlag = TRUE;
				break;

			case 'r':	/* remark file */
				sc.remarkName = "";

				if ((argc > 1) && (argv[1][0] != '-'))
				{
					sc.remarkName = *(++argv);
					argc--;
				}

				break;

			case 't':	/* type edit to terminal */
				action |= A_TYPE;
				editString = GetEditString(&argc, &argv);
				break;

			case 'u':	/* update history file */
				action |= A_UPDATE;
				break;

			case 'v':	/* be verbose */
				fc.verbosity = GetValue(&argc, &argv,
					(long) VERBOSE_FULL);

				break;

			case 'w':	/* descripe what changes are */
				fc.whatFlag = TRUE;
				break;

			default:
				Error("Unknown option \"-%c\"", *cp);
			}

			continue;
		}

		/*
		 * Here for multi-character options.
		 */
		if (strcmp(cp, "debug") == 0)
		{
			fc.debugFlag = TRUE;
			continue;
		}

		if (strcmp(cp, "create") == 0)
		{
			action |= A_CREATE;
			continue;
		}

		if (strcmp(cp, "du") == 0)
		{
			action |= A_DIFF_UPDATE;
			continue;
		}

		if (strcmp(cp, "cu") == 0)
		{
			action |= A_COND_UPDATE;
			continue;
		}

		if (strcmp(cp, "clean") == 0)
		{
			action |= A_CLEAN;
			continue;
		}

		if (strcmp(cp, "check") == 0)
		{
			action |= A_CHECK;
			continue;
		}

		if (strcmp(cp, "fu") == 0)
		{
			sc.forceUpdateFlag = TRUE;
			continue;
		}

		if (strcmp(cp, "fw") == 0)
		{
			sc.forceWriteFlag = TRUE;
			continue;
		}

		if (strcmp(cp, "all") == 0)
		{
			allFlag = TRUE;
			continue;
		}

		if (strcmp(cp, "help") == 0)
		{
			Usage();
			exit(0);
		}

		if (strcmp(cp, "name") == 0)
		{
			argv++;

			if ((--argc <= 0) || (**argv == '-'))
				Error("No edit name supplied");

			editName = *argv;
			CheckEditName(editName);

			continue;
		}

		if (strcmp(cp, "nw") == 0)
		{
			sc.noWriteFlag = TRUE;
			continue;
		}

		if (strcmp(cp, "prune") == 0)
		{
			action |= A_PRUNE;
			editString = GetEditString(&argc, &argv);
			continue;
		}

		if (strcmp(cp, "stdin") == 0)
		{
			stdinFlag = TRUE;
			continue;
		}

		Error("Unknown option \"-%s\"", cp);
	}

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		buildDate[strlen(buildDate) - 3] = '\0';

		printf("[FSC version %s of %s]\n", VERSION, &buildDate[8]);
	}

	/*
	 * If no action was specified, default it.
	 */
	if (action == 0)
		action = (editName ? A_NAME : A_EXTRACT);

	/*
	 * Check for illegal combinations of options.
	 */
	if (action != (action & -action))
		Error("Multiple actions specified");

	if (sc.remarkName && ((action & REMARK_ACTIONS) == 0))
		Error("Cannot use -r with specified action");

	if (inputName && ((action & INPUT_ACTIONS) == 0))
		Error("Cannot use -i with specified action");

	if (outputName && ((action & OUTPUT_ACTIONS) == 0))
		Error("Cannot use -o with specified action");

	if (editName && ((action & NAME_ACTIONS) == 0))
		Error("Cannot use -name with specified action");

	if (sc.forceWriteFlag && sc.noWriteFlag)
		Error("Cannot specify both -fw and -nw");

	if (fc.quickFlag && fc.whatFlag)
		Error("Cannot specify both -w and -q");

	if ((moduleCount <= 0) && !stdinFlag && (action == A_CHECK))
		allFlag = TRUE;

	if ((moduleCount <= 0) && !allFlag && !stdinFlag)
		Error("No module name specified");

	if (stdinFlag && allFlag)
		Error("Cannot specify both -stdin and -all");

	if (allFlag)
	{
		if (moduleCount)
			Error("Cannot specify module names with -all");

		if (inputName && !IsDir(inputName))
			Error("Input name must be a directory for -all");

		if (outputName && !IsDir(outputName))
			Error("Output name must be a directory for -all");
	}

	if (stdinFlag)
	{
		if (moduleCount)
			Error("Cannot specify module names with -stdin");

		if (inputName && !IsDir(inputName))
			Error("Input name must be a directory for -stdin");

		if (outputName && !IsDir(outputName))
			Error("Output name must be a directory for -stdin");
	}

	if (stdinFlag && !sc.forceWriteFlag && !sc.noWriteFlag)
		Error("Must use -fw or -nw with -stdin");

	if (moduleCount > 1)
	{
		if (inputName && !IsDir(inputName))
		{
			Error(
		"Input name must be a directory for multiple modules");
		}

		if (outputName && !IsDir(outputName))
		{
			Error(
		"Output name must be a directory for multiple modules");
		}
	}

	/*
	 * Make sure that the module storage directory is accessible
	 */
	if (access(sc.dirName, R_OK) < 0)
	{
		if (errno == ENOENT)
		{
			Error(
		"Module storage directory \"%s\" needs to be created",
			sc.dirName);
		}

		FileError(FIO_NULL,
			"Cannot access module storage directory \"%s\"",
			sc.dirName);
	}

	/*
	 * If all of the modules in the module storage directory are to
	 * be acted upon, then do that.
	 */
	if (allFlag)
	{
		DoAllModules(action, inputName, outputName,
			editString, editString2, editName);

		return exitStatus;
	}

	/*
	 * If modules are to be input from standard input then do that.
	 * Each input line is a module name.
	 */
	if (stdinFlag)
	{
		while (fgets(moduleName, sizeof(moduleName), stdin) != NULL)
		{
			lineLength = strlen(moduleName);

			if (lineLength < 2)
				continue;

			if (moduleName[lineLength - 1] != '\n')
				Error("Module line from stdin is too long");

			moduleName[lineLength - 1] = '\0';

			DoModule(action, moduleName, inputName, outputName,
				editName, editString, editString2);
		}

		if (ferror(stdin))
			Error("Error reading from stdin");

		return exitStatus;
	}

	/*
	 * We will act on the module names given on the command line.
	 * Act on the specified module names only.
	 */
	for (curMod = 0; curMod < moduleCount; curMod++)
	{
		DoModule(action, modules[curMod], inputName, outputName,
			editName, editString, editString2);
	}

	return exitStatus;
}


/*
 * Perform the specified action for all modules in the edit history directory.
 * The edit history directory can contain directories, in which we search
 * recursively through them to find all of the module names within the
 * directory hierarchy.
 */
static void
DoAllModules(
	int		action,
	const char *	inputName,	/* input name (DIRECTORY) (or NULL) */
	const char *	outputName,	/* output name (DIRECTORY) (or NULL) */
	const char *	editString,	/* edit string (or NULL) */
	const char *	editString2,	/* second edit string (or NULL) */
	const char *	editName	/* edit name (or NULL) */
)
{
	char	pathBuffer[PATH_MAX];	/* current directory path */

	strcpy(pathBuffer, sc.dirName);

	AllModulesRecursion(pathBuffer, strlen(pathBuffer), action,
		inputName, outputName, editString, editString2, editName);
}


/*
 * Recursive routine to find all of the modules within a directory.
 * This does a deep traversal of the modules within the directory hierarchy.
 * Modules are regular files having the ".e" extension.  The supplied path
 * buffer is modified and must be large enough for all levels of recursion.
 */
static void
AllModulesRecursion(
	char *		path,		/* current directory path */
	int		moduleOffset,	/* offset in path of module names */
	int		action,
	const char *	inputName,	/* input name (DIRECTORY) (or NULL) */
	const char *	outputName,	/* output name (DIRECTORY) (or NULL) */
	const char *	editString,	/* edit string (or NULL) */
	const char *	editString2,	/* second edit string (or NULL) */
	const char *	editName	/* edit name (or NULL) */
)
{
	const char *	moduleName;	/* current module name */
	const char *	entryName;	/* current file name entry */
	DIR *		dirp;		/* directory handle */
	struct dirent *	dp;		/* current directory entry */
	int		entryLength;	/* length of filename entry */
	int		pathLength;	/* length of current path */
	int		modulePathLength;	/* length of module path */
	char *		endPath;	/* end of supplied path */
	char *		buffer;		/* buffer for name storage */
	unsigned long	bufSize;	/* current size of buffer */
	long		bufIndex;	/* index into buffer */

	/*
	 * Allocate an initial buffer to hold the module names.
	 */
	bufIndex = 0;
	bufSize = 2048;

	buffer = r_malloc(bufSize);

	if (buffer == NULL)
		Error("Cannot allocate memory for file names");

	/*
	 * Set up to read the file names from the current directory path.
	 */
	dirp = opendir(path);

	if (dirp == NULL)
	{
		FileError(FIO_NULL,
			"Cannot open edit history directory \"%s\"", path);
	}

	pathLength = strlen(path);

	endPath = path + pathLength;
	modulePathLength = pathLength - moduleOffset;

	/*
	 * Read all of the files in the FSC directory looking for those
	 * ending in our edit history extension.  Add all such file names
	 * to the module list (with the extension removed).
	 */
	while ((dp = readdir(dirp)) != 0)
	{
		entryName = dp->d_name;
		entryLength = strlen(entryName);

		/*
		 * Ignore empty names and the current and parent directories.
		 */
		if (entryLength == 0)
			continue;

		if ((entryLength == 1) && (entryName[0] == '.'))
			continue;

		if ((entryLength == 2) && (entryName[0] == '.') &&
			(entryName[1] == '.'))
		{
			continue;
		}

		/*
		 * Make sure that the path buffer is still big enough,
		 * taking into account the separating slash and final null.
		 */
		if (pathLength + entryLength + 2 > PATH_MAX)
			Error("Edit history path is too long");

		/*
		 * Construct the full path to this file name entry.
		 */
		*endPath = '/';

		memcpy(endPath + 1, entryName, entryLength + 1);

		/*
		 * If this is a directory then call ourself recursively
		 * to handle the modules within this directory.
		 */
		if (IsDir(path))
		{
			AllModulesRecursion(path, moduleOffset, action,
				inputName, outputName, editString,
				editString2, editName);

			continue;
		}

		/*
		 * If the file does not have the correct extension for
		 * the edit history of a module, then ignore it.
		 */
		if ((entryLength < 2) ||
			(entryName[entryLength - 2] != '.') ||
			(entryName[entryLength - 1] != 'e'))
		{
			continue;
		}

		/*
		 * This is a module edit history file name.
		 * Remove the extension characters from the length.
		 */
		entryLength -= 2;

		/*
		 * Reallocate the module name buffer if necessary.
		 */
		if ((bufIndex + modulePathLength + entryLength + 3) > bufSize)
		{
			bufSize = bufIndex + modulePathLength
				+ entryLength + 8192;

			buffer = r_realloc(buffer, bufSize);

			if (buffer == NULL)
				Error("Realloc failed for file names");
		}

		/*
		 * Append the module name into the module name buffer.
		 * The module name includes a leading relative path if it
		 * was found in a subdirectory.
		 */
		if (modulePathLength > 0)
		{
			memcpy(buffer + bufIndex, path + moduleOffset + 1,
				modulePathLength);

			bufIndex += modulePathLength - 1;

			buffer[bufIndex++] = '/';
		}

		memcpy(buffer + bufIndex, entryName, entryLength);
		bufIndex += entryLength;

		buffer[bufIndex++] = '\0';
	}

	closedir(dirp);

	/*
	 * Terminate the module list with an extra null.
	 */
	buffer[bufIndex] = '\0';

	/*
	 * Now act on all of the collected module names that were found
	 * within this particular directory.  The module names are
	 * separated by single null characters, with an extra null
	 * character to end the list.
	 */
	moduleName = buffer;

	while (*moduleName)
	{
		DoModule(action, moduleName, inputName, outputName,
			editName, editString, editString2);

		moduleName += (strlen(moduleName) + 1);
	}

	r_free(buffer);

	*endPath = '\0';
}


/*
 * Perform the specified action for a particular module.
 * In case of a fatal error, a longjmp is made back to this routine,
 * when then simply returns.  The exitStatus will have been set nonzero
 * so that we will later exit with an error.
 */
static void
DoModule(
	int		action,
	const char *	moduleName,	/* module name */
	const char *	inputName,	/* input name (or NULL) */
	const char *	outputName,	/* output name (or NULL) */
	const char *	editName,	/* name of edit (or NULL) */
	const char *	editString,	/* edit string (or NULL) */
	const char *	editString2	/* second edit string (or NULL) */
)
{
	char	fullInputName[PATH_MAX];	/* input name if a directory */
	char	fullOutputName[PATH_MAX];	/* output name if a directory */

	/*
	 * Set up to catch any errors, and then reset all static data
	 * in preparation for this new module.
	 */
	if (setjmp(jmpBuf))
		return;

	Restart(FALSE);

	/*
	 * Make sure that the module name is not too long.
	 */
	if (strlen(moduleName) > MODULE_MAX)
		Error("Module name \"%s\" is too long", moduleName);

	/*
	 * Create the full path names if the input or output file names
	 * are directories.
	 */
	if (inputName && IsDir(inputName))
	{
		strcpy(fullInputName, inputName);
		strcat(fullInputName, "/");
		strcat(fullInputName, moduleName);

		inputName = fullInputName;
	}

	if (outputName && IsDir(outputName))
	{
		strcpy(fullOutputName, outputName);
		strcat(fullOutputName, "/");
		strcat(fullOutputName, moduleName);

		outputName = fullOutputName;
	}

	/*
	 * Save everything away in our structure for reference.
	 */
	sc.moduleName = moduleName;

	strcpy(sc.baseName, sc.dirName);
	strcat(sc.baseName, "/");
	strcat(sc.baseName, moduleName);
	strcpy(sc.sourceName, sc.baseName);
	strcat(sc.sourceName, EXT_SOURCE);
	strcpy(sc.historyName, sc.baseName);
	strcat(sc.historyName, EXT_HISTORY);

	/*
	 * Now execute the specified action for this module name.
	 */
	switch (action)
	{
		case A_EXTRACT:
			if (outputName == NULL)
				outputName = moduleName;

			ExtractHistory(editString, outputName,
				(fc.verbosity != 0));
			break;

		case A_CREATE:
			if (inputName == NULL)
				inputName = moduleName;

			CreateHistory(inputName, editName);
			break;

		case A_DIFF:
			if (inputName && editString2)
				Error("Conflicting options for differences");

			if (inputName == NULL)
				inputName = moduleName;

			DiffHistory(inputName, outputName, editString,
				editString2, TRUE);

			break;

		case A_DIFF_UPDATE:
			if (inputName == NULL)
				inputName = moduleName;

			DiffHistory(inputName, outputName, editString,
				editString2, FALSE);

			if ((fc.inserts == 0) && (fc.deletes == 0))
			{
				if (fc.verbosity)
				{
					printf(
				"Module \"%s\" does not need updating\n",
					moduleName);
				}

				break;
			}

			printf("\n");
			UpdateHistory(inputName, editName);
			break;

		case A_COND_UPDATE:
			if (inputName == NULL)
				inputName = moduleName;

			CompHistory(inputName);

			if ((fc.inserts == 0) && (fc.deletes == 0))
			{
				if (fc.verbosity)
				{
					printf(
				"Module \"%s\" does not need updating\n",
					moduleName);
				}

				break;
			}

			UpdateHistory(inputName, editName);
			break;

		case A_UPDATE:
			if (inputName == NULL)
				inputName = moduleName;

			UpdateHistory(inputName, editName);
			break;

		case A_LIST:
			ListHistory(editString, editString2, outputName);
			break;

		case A_NAME:
			NameHistory(editName);
			break;

		case A_TYPE:
			ExtractHistory(editString, outputName, VERBOSE_NONE);
			break;

		case A_CLEAN:
			if (inputName == NULL)
				inputName = moduleName;

			CleanHistory(inputName);
			break;

		case A_CHECK:
			if (inputName == NULL)
				inputName = moduleName;

			CheckHistory(inputName);
			break;

		case A_PRUNE:
			PruneHistory(editString);
			break;

		default:
			Error("Unknown action %x", action);
			break;
	}
}


/*
 * Routine to read an optional edit number string from the command line.
 * Returns the string, or NULL if one is not there.
 */
static const char *
GetEditString(
	int *		argcp,		/* address of argc variable */
	const char ***	argvp		/* address of argv variable */
)
{
	const char *	str;

	if (*argcp <= 1)
		return NULL;

	str = (*argvp)[1];

	if ((*str == '-') && ((str[1] <= '0') || (str[1] > '9')))
		return NULL;

	(*argcp)--;
	(*argvp)++;

	return str;
}


/*
 * Routine to read an optional non-negative number from the command line.
 * Returns the specified default value if no number is present.
 */
static long
GetValue(
	int *		argcp,		/* address of argc variable */
	const char ***	argvp,		/* address of argv variable */
	long		value		/* default value */
)
{
	const char *	cp;		/* characters of next argument */

	if (*argcp <= 1)
		return value;

	cp = (*argvp)[1];

	if (*cp == '-')
		return value;

	value = 0;

	while ((*cp >= '0') && (*cp <= '9'))
		value = (value * 10) + *cp++ - '0';

	if (*cp || (value < 0))
		Error("Bad numeric argument");

	(*argcp)--;
	(*argvp)++;

	return value;
}


/*
 * Error routine which can format the output.
 * Jumps back to top level code.
 */
void
Error(
	const char *	fmt,
	...
)
{
	va_list	ap;

	va_start(ap, fmt);
	vsprintf(msgBuf, fmt, ap);
	va_end(ap);

	fflush(stdout);
	fprintf(stderr, "%s\n", msgBuf);
	exitStatus = 1;

	Restart(TRUE);
}


/*
 * Error routine called for file errors.
 * This prints out the error message for a file error.
 * Jumps back to top level code.
 */
void
FileError(
	const FIO *	fp,
	const char *	fmt,
	...
)
{
	va_list	ap;

	if (fp && fp->errorCode)
		errno = fp->errorCode;

	va_start(ap, fmt);
	vsprintf(msgBuf, fmt, ap);
	va_end(ap);

	fflush(stdout);
	perror(msgBuf);
	exitStatus = 1;

	Restart(TRUE);
}


/*
 * Routine called after errors to restart the program for the next file.
 * This reenables breaks, closes all files, frees all allocated memory,
 * resets static variables, and then possibly longjmps back to the top
 * level code.
 */
void
Restart(BOOL jumpFlag)
{
	int	fd;		/* for closing files */

	EnableSignals();
	CmReset();
	FcompReset();
	FioReset();
	EditReset();

	sc.modifyBuffer = NULL;
	sc.firstEdit = 0;
	sc.lastEdit = 0;
	sc.tablePos = 0;

	for (fd = 3; fd < 512; fd++)
		(void) close(fd);

	if (jumpFlag)
		longjmp(jmpBuf, 1);
}


static const char * const helpText[] =
{
"",
"Program to manipulate edit histories of modules.",
"",
"Usage is one of the following:",
"sc [modules] -create   Create a new module from an initial file",
"sc [modules] -e n      Extract a saved edit for a module to a file",
"sc [modules] -t n      Type a saved edit for a module to the terminal",
"sc [modules] -u        Update a module to create a new edit",
"sc [modules] -cu       Update a module if there are any differences",
"sc [modules] -d m n    Display differences between edits of a module",
"sc [modules] -du       Display differences and then update if different",
"sc [modules] -l m n    List the edits made to a module",
"sc [modules] -prune n  Delete all edits from the edit history before edit n",
"sc [modules] -clean    Delete file if module is up to date",
"sc [modules] -check    Check to see if a file needs to be updated",
"",
"Additional options (with defaults if not specified):",
"-all                   Do all modules in module storage directory",
"-fu                    Do not check matching edit numbers when updating",
"-fw                    Do not check overwriting of existing files",
"-i filename            Input name for create or update (module name)",
"-m count               Maximum lines to modify for edit sequences (25)",
"-name editname         Specify a name for this edit (none)",
"-nw                    Do not overwrite existing files",
"-o filename            Output name for extract, display, or list",
"-p pathname            Pathname for module storage (\"FSC_EDITS\")",
"-q                     Only give a quick list of differences or edits",
"-r filename            Name of file containing remarks (none)",
"-stdin                 Do module names read line by line from stdin",
"-v level               Set verbosity level (1)",
"-w                     Show what changes are when displaying differences",
NULL
};


/*
 * Describe the program
 */
static void
Usage()
{
	const char * const *	str;

	for (str = helpText; *str; str++)
		fprintf(stderr, "%s\n", *str);
}

/* END CODE */
