/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routines to modify lines of files for special purposes.
 */

#include <stdio.h>
#include <string.h>

#include "fio.h"
#include "fcomp.h"
#include "sc.h"
#include "cmalloc.h"


#define INPUT_LINE_SIZE	200	/* longest input line */
#define OUTPUT_LINE_SIZE 300	/* longest output line */
#define MAX_VALUES	4	/* most number of values per line */
#define MIN_WORD	4	/* shortest keyword */
#define MAX_WORD	6	/* longest keyword */
#define SEQ_EDGE	3	/* size of edge of sequence */
#define MIN_SEQUENCE ((SEQ_EDGE * 2) + MIN_WORD)	/* smallest sequence */


#define V_EDIT		1	/* edit keyword */
#define V_DATE		2	/* date keyword */
#define V_USER		3	/* user name keyword */
#define	V_MODULE	4	/* module keyword */


typedef struct
{
	int		type;		/* type of value */
	const char *	str;		/* new string value */
	const char *	begVal;		/* beginning of old value */
	const char *	endVal;		/* end of old value */
} VALUE;


static	int	ScanWords(const char *, long, const INFO *, VALUE *);
static	int	ScanSequence(const char *, long, const char **, const char **);


/*
 * This routine detects lines which contain strings with the format:
 *	[# keyword value, keyword value, ..., keyword value #]
 * and changes the values to appropriate strings as specified by the keywords.
 * If the format is not correct, any keyword is unknown, or the line is too
 * long, then the line is unchanged.  Keywords can be in either lower or upper
 * case.  This routine is usually only called for a small number of lines of
 * the file in order to increase the speed of processing the rest of the file.
 * The info structure provides the values for insertion.  If the pointer is
 * NULL, then all the values will be cleared.  The returned data is only
 * valid while the input data is valid, and only until the next call.
 */
const char *
ModifyLine(
	const char *	cp,		/* line to be examined */
	long *		lineLen,	/* length of input and output lines */
	const INFO *	info		/* edit info (or NULL if none) */
)
{
	const char *	endLine;	/* ending position in line */
	const char *	str;		/* string for replacement */
	char *		op;		/* current output character */
	VALUE *		vp;		/* current value on line */
	long		len;		/* line length */
	int		count;		/* number of values on line */
	VALUE		values[MAX_VALUES];	/* information about values */

	len = *lineLen;

	if ((len < MIN_SEQUENCE) || (len > INPUT_LINE_SIZE))
		return cp;

	count = ScanWords(cp, len, info, values);

	if (count <= 0)
		return cp;

	if (sc.modifyBuffer == NULL)
	{
		sc.modifyBuffer = malloc(OUTPUT_LINE_SIZE);

		if (sc.modifyBuffer == NULL)
		{
			FileError(FIO_NULL,
				"Cannot allocate memory for modify buffer");
		}
	}

	/*
	 * Copy the old line to the new one, except replace the old
	 * values with the new ones.
	 */
	endLine = cp + len;

	op = sc.modifyBuffer;

	for (vp = values; --count >= 0; vp++)
	{
		while (cp != vp->begVal)
			*op++ = *cp++;

		str = vp->str;

		if (str)
		{
			*op++ = ' ';

			while (*str)
				*op++ = *str++;
		}

		cp = vp->endVal;
	}

	while (cp != endLine)
		*op++ = *cp++;

	*op = '\0';

	*lineLen = (op - sc.modifyBuffer);

	return sc.modifyBuffer;
}


/*
 * Routine to find keywords within special sequences on a line.
 * Returns a list of the new values for those keywords, and the locations
 * where the old values are within the line.  Return value is the number
 * of keywords found, or -1 if something invalid was found.
 */
static int
ScanWords(
	const char *	line,		/* current location in line */
	long		len,		/* length of rest of line */
	const INFO *	info,		/* information structure (or NULL) */
	VALUE *		vp		/* beginning of value table */
)
{
	const char *	cp;		/* current character */
	const char *	beg;		/* beginning of special sequence */
	const char *	end;		/* end of special sequence */
	const char *	endWord;	/* end of word */
	const char *	str;		/* string value */
	char *		op;		/* output character */
	int		count;		/* number of values found */
	int		ch;		/* current character */
	int		type;		/* type of value found */
	char		word[MAX_WORD + 2];	/* lower cased word */

	count = 0;
	endWord = &word[MAX_WORD];

	while (ScanSequence(line, len, &beg, &end))
	{
		for (cp = beg; cp != end; )
		{
			if (++count > MAX_VALUES)
				return -1;

			for (op = word; ; cp++)
			{
				ch = *cp;

				if ((ch == ' ') || (ch == ','))
					break;

				if (op == endWord)
					return -1;

				if ((ch >= 'A') && (ch <= 'Z'))
					ch += ('a' - 'A');

				*op++ = ch;
			}

			*op = '\0';

			vp->begVal = cp;

			while ((cp != end) && (*cp != ','))
				cp++;

			vp->endVal = cp;

			if (cp != end)
			{
				if ((++cp == end) || (*cp++ != ' ') ||
					(cp == end))
				{
					return -1;
				}
			}

			if (strcmp(word, "edit") == 0)
				type = V_EDIT;
			else if (strcmp(word, "date") == 0)
				type = V_DATE;
			else if (strcmp(word, "user") == 0)
				type = V_USER;
			else if (strcmp(word, "module") == 0)
				type = V_MODULE;
			else
				return -1;

			str = NULL;

			if (info)
			{
				switch (type)
				{
					case V_EDIT:
						str = info->edit;
						break;

					case V_DATE:
						str = info->date;
						break;

					case V_USER:
						str = info->user;
						break;

					case V_MODULE:
						str = sc.moduleName;
						break;
				}
			}

			vp->str = str;
			vp->type = type;

			vp++;
		}

		len -= (end - line);
		line = end;
	}

	return count;
}


/*
 * Routine to find a special sequence inside of a line, and return
 * pointers to the beginning and end of the inner part of the sequence.
 * Returns TRUE if found a potential sequence for examination.
 */
static int
ScanSequence(
	const char *	cp,		/* current location in the line */
	long		lineLen,	/* length of line */
	const char **	begSeq,		/* start of inner part of sequence */
	const char **	endSeq		/* end of inner part of sequence */
)
{
	const char *	cp2;		/* test characters */
	long		lineLen2;	/* length of line left */

	if (lineLen > INPUT_LINE_SIZE)
		return FALSE;

	while (lineLen > MIN_SEQUENCE)
	{
		if ((cp[0] != '[') || (cp[1] != '#') || (cp[2] != ' '))
		{
			cp++;
			lineLen--;

			continue;
		}

		cp += SEQ_EDGE;

		lineLen -= SEQ_EDGE;
		lineLen2 = lineLen;

		cp2 = cp;

		while ((cp2[0] != ' ') || (cp2[1] != '#') || (cp2[2] != ']'))
		{
			if (--lineLen2 <= SEQ_EDGE)
				return FALSE;

			cp2++;
		}

		*begSeq = cp;
		*endSeq = cp2;

		return TRUE;
	}

	return FALSE;
}


/*
 * Routine to search through the beginning of a file to find the edit
 * number, and return its value.  If the edit number is not found, or
 * if its value is null, then -1 is returned.
 */
long
FindEdit(const char * inputName)
{
	FIO *		fp;		/* file being read */
	const char *	cp;		/* current line */
	VALUE *		vp;		/* current value */
	long		lineLen;	/* line length */
	long		num;		/* scanned number */
	int		count;		/* number of values on line */
	long		i;		/* line counter */
	FIO		inputFile;	/* file structure */
	VALUE		values[MAX_VALUES];	/* returned values */

	fp = &inputFile;

	if (!FioOpen(fp, inputName, FIO_READ))
		FileError(fp, "Cannot open input file \"%s\"", inputName);

	for (i = sc.modifyLines; i > 0; i--)
	{
		cp = FioReadline(fp, &lineLen, FALSE);

		if (cp == NULL)
			break;

		count = ScanWords(cp, lineLen, (INFO *) NULL, values);

		for (vp = values; --count >= 0; vp++)
		{
			if (vp->type != V_EDIT)
				continue;

			cp = vp->begVal + 1;

			num = 0;

			while ((*cp >= '0') && (*cp <= '9'))
				num = (num * 10) + (*cp++ - '0');

			if ((cp == vp->endVal) && (num > 0))
			{
				if (!FioClose(fp))
				{
					FileError(fp,
						"Error reading file \"%s\"",
						inputName);
				}

				return num;
			}
		}
	}

	if (!FioClose(fp))
		FileError(fp, "Error reading file \"%s\"", inputName);

	return -1;
}

/* END CODE */
