/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to specify a name for an edit of a module.
 */

#include <stdio.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"


/*
 * Create a new name for an edit of a module.
 * Currently this only works for the latest edit.
 */
void
NameHistory(const char * editName)
{
	FIO *		fp;		/* history file */
	POS *		pp;		/* current entry in position table */
	POS *		posTable;	/* position table */
	const char *	oldNames;	/* old names for edit */
	char *		newNames;	/* new names for edit */
	int		nameLen;	/* length of new name */
	long		edit;		/* current edit number */
	int		saveError;	/* saved error code on errors */

	fp = sc.historyFile;

	OpenHistoryFile(FIO_UPDATE);

	if (fc.verbosity)
	{
		printf("Latest edit %ld of module \"%s\" being named \"%s\"\n",
			sc.lastEdit, sc.moduleName, editName);
	}

	posTable = ReadPosTable(fp);
	posTable++;			/* ignore unused entry */

	pp = posTable;

	oldNames = pp->p_names;

	if (oldNames == NULL)		/* single new name */
		pp->p_names = editName;
	else
	{
		/*
		 * Put name in front of other names
		 */
		nameLen = strlen(editName);

		newNames = AllocStr((unsigned long)
			(strlen(oldNames) + nameLen + 2));

		strcpy(newNames, editName);

		newNames[nameLen] = ' ';

		strcpy(&newNames[nameLen+1], oldNames);

		pp->p_names = newNames;
	}

	DisableSignals();

	/*
	 * Now position back to the beginning of the position table, and write
	 * back the table with the new name.  Errors will be handled later.
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Writing new position table]\n");

	if (!FioSeek(fp, sc.tablePos))
		Error("Cannot seek to position table");

	pp = posTable;

	for (edit = sc.lastEdit; edit >= sc.firstEdit; edit--, pp++)
	{
		FioFormat(fp,
			(pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
			T_POSITION, edit, pp->p_pos, pp->p_names);
	}

	FioFormat(fp, "%c %ld\n", T_EOF, FioPos(fp));

	/*
	 * Force the last bit of the output to the file, then check for errors.
	 * If an error occurred, write back the original position table without
	 * the new name.  This should succeed since it was there previously.
	 */
	FioFlush(fp);

	if (FioError(fp))
	{
		saveError = fp->errorCode;
		fp->errorCode = 0;

		if (!FioSeek(fp, sc.tablePos))
		{
			FileError(fp,
		"Seek to position table failed - history file \"%s\" damaged",
				sc.historyName);
		}

		pp = posTable;
		pp->p_names = oldNames;

		for (edit = sc.lastEdit; edit >= sc.firstEdit; edit--, pp++)
		{
			FioFormat(fp, (pp->p_names ?
				"%c %ld %ld %s\n" : "%c %ld %ld\n"),
				T_POSITION, edit, pp->p_pos, pp->p_names);
		}

		FioFormat(fp, "%c %ld\n", T_EOF, FioPos(fp));

		FioFlush(fp);

		if (FioError(fp))
		{
			FileError(fp,
	"Rewrite of position table failed - history file \"%s\" damaged",
				sc.historyName);
		}

		FioClose(fp);

		fp->errorCode = saveError;

		FileError(fp,
	"Update of history file \"%s\" failed (file still correct)",
			sc.historyName);
	}

	FioClose(fp);

	EnableSignals();
}


/*
 * Check to see if a new edit name is valid.
 * Valid names cannot contain only digits, or contain an empty string or
 * only digits following the last plus or minus sign.  They also cannot
 * be either of the two special reserved words.
 */
void
CheckEditName(const char * editName)
{
	const char *	cp;
	int		ch;

	if (*editName == '\0')
		Error("An empty edit name is illegal");

	if ((strcmp(editName, OLDEST_NAME) == 0) ||
		(strcmp(editName, NEWEST_NAME) == 0))
	{
		Error("Edit name \"%s\" is reserved", editName);
	}

	for (cp = editName + strlen(editName); cp != editName; cp--)
	{
		ch = cp[-1];

		if ((ch == '-') || (ch == '+'))
		{
			Error(
	"At least one non-digit must appear after a plus or minus sign");
		}

		if ((ch < '0') || (ch > '9'))
			return;
	}

	Error("Edit names must contain a non-digit character");
}

/* END CODE */
