/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to prune away old edits from the edit history file.
 */

#include <stdio.h>
#include <string.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"


/*
 * Prune old edits from the history file to save on disk space.
 * This removes all edits up until the specified edit number.
 */
void
PruneHistory(const char * editName)
{
	FIO *	fp;			/* input history file */
	FIO *	ofp;			/* output history file */
	POS *	pp;			/* current entry in position table */
	long	lastPrunedEdit;		/* last edit number to be pruned */
	long	firstKeepEdit;		/* first edit number to keep */
	long	pruneCount;		/* number of edits being pruned */
	long	copyPos;		/* starting position to copy from */
	long	newTablePos;		/* position of new position table */
	long	edit;			/* current edit number */
	long	shift;			/* amount to shift data up by */
	char	tempName[PATH_MAX];	/* temporary name of new edit file */

	strcpy(tempName, sc.baseName);
	strcat(tempName, EXT_NEW);

	fp = sc.historyFile;
	ofp = sc.outputFile;

	OpenHistoryFile(FIO_READ);

	firstKeepEdit = FindEditNumber(editName);
	lastPrunedEdit = firstKeepEdit - 1;

	pruneCount = firstKeepEdit - sc.firstEdit;

	if (pruneCount <= 0)
	{
		FioClose(fp);

		if (fc.verbosity)
		{
			printf("No edits pruned from module \"%s\"\n",
				sc.moduleName);
		}

		return;
	}

	if (!sc.forceUpdateFlag)
	{
		printf(
	"The edit history for module \"%s\" currently has edits %ld to %ld.\n",
			sc.moduleName, sc.firstEdit, sc.lastEdit);

		printf("Do you wish to irretrievably discard edit");

		printf((sc.firstEdit == lastPrunedEdit) ?
			" %ld? " : "s %ld to %ld? ",
			sc.firstEdit, lastPrunedEdit);

		if (!Confirm())
			return;
	}

	pp = ReadPosTable(fp);
	pp++;			/* ignore unused entry */

	if (!FioOpen(ofp, tempName, FIO_WRITE))
	{
		FileError(ofp, "Cannot open temporary history file \"%s\"",
			tempName);
	}

	FioFormat(ofp, HEADER_FORMAT, T_HEADER, 0L, 0L, 0L, 0L);

	copyPos = pp[sc.lastEdit - firstKeepEdit].p_pos;

	FioSeek(fp, copyPos);

	FioCopy(fp, ofp, sc.tablePos - copyPos);

	if (!FioClose(fp))
	{
		FileError(fp, "Read error for history file \"%s\"",
			sc.historyName);
	}

	newTablePos = FioPos(ofp);

	shift = sc.tablePos - newTablePos;

	for (edit = sc.lastEdit; edit >= firstKeepEdit; edit--, pp++)
	{
		FioFormat(ofp,
			(pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
			T_POSITION, edit, pp->p_pos - shift, pp->p_names);
	}

	FioFormat(ofp, "%c %ld\n", T_EOF, FioPos(ofp));

	FioSeek(ofp, 0L);

	FioFormat(ofp, HEADER_FORMAT, T_HEADER, firstKeepEdit, sc.lastEdit,
		0L, newTablePos);

	if (!FioClose(ofp))
	{
		FileError(ofp, "Write error for temporary history file \"%s\"",
			tempName);
	}

	DisableSignals();

	if (!RenameFiles(EXT_HISTORY))
	{
		FileError(FIO_NULL,
			"Rename failed for edit history file \"%s\"",
			sc.historyName);
	}

	if (fc.verbosity)
	{
		printf("Pruned %ld edit%s from module \"%s\"\n", pruneCount,
			(pruneCount == 1) ? "" : "s", sc.moduleName);
	}

	EnableSignals();
}

/* END CODE */
