/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Commonly used generic routines for source control program.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "fio.h"
#include "fcomp.h"
#include "sc.h"
#include "cmalloc.h"


/*
 * Routine used to open the history file and read the first line.
 * This saves the information about the first and last edit numbers,
 * and the position of the position table in variables.
 */
void
OpenHistoryFile(int mode)
{
	FIO *		fp;		/* file handle for history file */
	const char *	cp;		/* current line of file */

	fp = sc.historyFile;

	if (!FioOpen(fp, sc.historyName, mode))
	{
		if (FioError(fp) == ENOENT)
			Error("Module \"%s\" does not exist", sc.moduleName);

		FileError(fp, "Cannot open edit history file \"%s\"",
			sc.historyName);
	}

	cp = GetLine(fp, 0L, T_HEADER);

	if (cp == NULL)
		Error("Cannot read header line");

	if ((sc.lineLen + 2) != HEADER_LINE_LENGTH)
		Error("Header line has wrong length");

	cp = GetNumber(cp, &sc.firstEdit);
	cp = GetNumber(cp, &sc.lastEdit);
	cp = GetNumber(cp, &sc.lastFile);
	cp = GetNumber(cp, &sc.tablePos);

	if (cp == NULL)
		Error("Bad header line");

	if ((sc.firstEdit < 0) || (sc.lastEdit < sc.firstEdit) ||
		(sc.tablePos <= 0))
	{
		Error("Bad edit values in header line");
	}
}


/*
 * Find an edit number in the history file given the edit name string.
 * This string is either a straight number, a straight edit name, or a
 * combination of the two.  A combination is indicated by an imbedded plus
 * or minus sign.  Examples:  "45", "good2", "nice+5".  The resulting
 * edit number is returned if it is valid.
 */
long
FindEditNumber(const char * editName)
{
	FIO *		fp;		/* history file */
	const char *	cp;		/* current character */
	const char *	testName;	/* name to test with */
	int		nameLen;	/* length of edit name */
	long		edit;		/* resulting edit number */
	long		relEdit;	/* relative edit number */
	long		curEdit;	/* current edit number */
	long		tempEdit;	/* edit number just read */
	long		tempPos;	/* position number just read */
	BOOL		isNeg;		/* TRUE if negative number given */
	BOOL		hasNumber;	/* TRUE if name contains number */
	char		nameBuf[32];	/* temporary name buffer */

	if (editName == NULL)
		return sc.lastEdit;

	fp = sc.historyFile;
	edit = 0;
	relEdit = 0;
	nameLen = 0;
	isNeg = FALSE;
	hasNumber = FALSE;
	cp = editName;

	/*
	 * Scan from the end of the name backwards skipping over all
	 * trailing digits.
	 */
	cp = editName + strlen(editName);

	while ((cp != editName) && (cp[-1] >= '0') && (cp[-1] <= '9'))
		cp--;

	hasNumber = (cp == editName);

	if ((cp != editName) && ((cp[-1] == '-') || (cp[-1] == '+')))
	{
		hasNumber = TRUE;
		cp--;
	}

	if (!hasNumber)
		cp = editName + strlen(editName);

	/*
	 * First, see if there is a name present.
	 * If so, copy it and remove any trailing edit number.
	 */
	if (cp != editName)
	{
		nameLen = cp - editName;

		if (nameLen >= sizeof(nameBuf))
			Error("Edit name is too long");

		memcpy(nameBuf, editName, nameLen);
		nameBuf[nameLen] = '\0';

		editName = nameBuf;
	}

	/*
	 * Now parse the edit number offset if any.
	 * If there isn't one, the parse will just obtain zero.
	 */
	if ((*cp == '+') || (*cp == '-'))
		isNeg = (*cp++ == '-');

	while ((*cp >= '0') && (*cp <= '9'))
		relEdit = relEdit * 10 + (*cp++ - '0');

	if (relEdit < 0)
	{
		Error("Bad edit name \"%s\" given for module \"%s\"",
			editName, sc.moduleName);
	}

	if (isNeg)
		relEdit = -relEdit;

	/*
	 * If a name was specified, then search the history file for that name.
	 * Remember the corresponding edit number if found.
	 * Handle the magic 'oldest' and 'newest' edit names specially.
	 */
	if (nameLen > 0)
	{
		if (strcmp(editName, OLDEST_NAME) == 0)
		{
			edit = sc.firstEdit;

			goto foundedit;
		}

		if (strcmp(editName, NEWEST_NAME) == 0)
		{
			edit = sc.lastEdit;

			goto foundedit;
		}

		if (fc.verbosity > VERBOSE_DEFAULT)
			printf("[Searching for edit name \"%s\"]\n", editName);

		curEdit = sc.lastEdit + 1;

		FioSeek(fp, sc.tablePos);

		while (edit == 0)
		{
			if (--curEdit < sc.firstEdit)
			{
				FioClose(fp);

				Error(
			"Edit name \"%s\" does not exist for module \"%s\"",
					editName, sc.moduleName);
			}

			cp = GetLine(fp, NO_SEEK, T_POSITION);

			if (cp == NULL)
			{
				FileError(fp,
			"Cannot read position line for edit %ld",
				curEdit);
			}

			cp = GetNumber(cp, &tempEdit);
			cp = GetNumber(cp, &tempPos);

			if ((cp == NULL) || (tempEdit != curEdit))
			{
				Error("Bad position line for edit %ld",
					curEdit);
			}

			/*
			 * Search the line for the edit name
			 */
			while (*cp)
			{
				while (*cp == ' ')
					cp++;

				testName = cp;

				while (*cp && (*cp != ' ') && (*cp != '\n'))
					cp++;

				if ((cp - testName == nameLen) &&
					(memcmp(testName, editName, nameLen)
						== 0))
				{
					edit = curEdit;
				}

				cp++;
			}
		}
	}

	/*
	 * We either found the name, or else we just had a number.
	 * Combine the two values, and complain if out of range.
	 * When given just a number, negative means backwards from last one.
	 */
foundedit:
	if ((edit == 0) && (relEdit <= 0))
		edit = sc.lastEdit;

	edit += relEdit;

	if ((edit < sc.firstEdit) || (edit > sc.lastEdit))
	{
		FioClose(fp);

		Error(
	"Edit number %ld for module \"%s\" is not in the range %ld to %ld",
			edit, sc.moduleName, sc.firstEdit, sc.lastEdit);
	}

	return edit;
}


/*
 * Read in the position table from the history file and save it into memory.
 * This reserves the first entry of the table for a possible new entry.
 * The history file was already opened and the header line was read.
 * Returns a pointer to the beginning of the table.
 */
POS *
ReadPosTable(FIO * fp)
{
	const char *	cp;		/* current line of file */
	POS *		pp;		/* current position table */
	POS *		posTable;	/* beginning of position table */
	char *		names;		/* names for edit */
	long		edit;		/* current edit number */
	long		tempEdit;	/* temporary edit number */
	long		tempPos;	/* temporary position */
	long		eofPos;		/* end of file position */
	int		len;		/* length of string */

	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Saving old position table]\n");

	pp = (POS *) malloc(sizeof(POS) * (sc.lastEdit - sc.firstEdit + 2));

	if (pp == NULL)
		Error("Cannot allocate memory for position table");

	posTable = pp;

	pp->p_pos = -1;
	pp->p_names = NULL;
	pp++;

	/*
	 * Read the position table lines from the file and save the positions.
	 * Save the names also if there are any.
	 */
	if (!FioSeek(fp, sc.tablePos))
		Error("Cannot seek to position table");

	for (edit = sc.lastEdit; edit >= sc.firstEdit; edit--)
	{
		cp = GetLine(fp, NO_SEEK, T_POSITION);

		if (cp == NULL)
			Error("Cannot read position entry for edit %ld", edit);

		cp = GetNumber(cp, &tempEdit);
		cp = GetNumber(cp, &tempPos);

		if (cp == NULL)
		{
			Error("Bad numbers in position entry for edit %ld",
				edit);
		}

		if (tempEdit != edit)
		{
			Error("Wrong number in position entry for edit %ld",
				edit);
		}

		pp->p_pos = tempPos;
		pp->p_names = NULL;

		while (*cp == ' ')
			cp++;

		/*
		 * If there is a name then save it.
		 */
		if (*cp && (*cp != '\n'))
		{
			len = strlen(cp);

			names = AllocStr((unsigned long) len);

			memcpy(names, cp, len - 1);
			names[len - 1] = '\0';

			pp->p_names = names;
		}

		pp++;
	}

	/*
	 * That should have been all of the position lines.
	 * Now verify that the next line is the end of file line.
	 */
	eofPos = FioPos(fp);

	cp = GetLine(fp, NO_SEEK, T_EOF);

	if (cp == NULL)
		Error("Cannot read end of file line");

	cp = GetNumber(cp, &tempPos);

	if (cp == NULL)
		Error("No position in end of file line");

	if (eofPos != tempPos)
		Error("Wrong position %ld in end of file line", eofPos);

	return posTable;
}


/*
 * Find the position of the beginning of an edit in the history file,
 * position to it, and verify that it is correct.  Then if infoLine is
 * non-NULL, read the first remark line for the edit (which is information
 * about the edit), and return that into infoLine.  The file position is
 * left ready to read the remainder of the edit sequence.
 */
void
StartEdit(
	FIO *		fp,		/* history file */
	long		num,		/* edit number of interest */
	char *		infoLine	/* returned info line (or NULL) */
)
{
	const char *	cp;		/* current line of file */
	long		editNumber;	/* current edit number read */
	long		editPos;	/* position of beginning of edit */
	long		temp;		/* temporary value read */

	cp = GetLine(fp, sc.tablePos, T_POSITION);

	if (cp == NULL)
		Error("Cannot read initial position table line");

	cp = GetNumber(cp, &editNumber);

	if (cp == NULL)
		Error("No edit number in start of position table");

	if (editNumber != sc.lastEdit)
		Error("Wrong edit number at start of position table");

	while (editNumber != num)
	{
		editNumber--;

		cp = GetLine(fp, NO_SEEK, T_POSITION);

		if (cp == NULL)
			Error("Position table too short");

		cp = GetNumber(cp, &temp);

		if (cp == NULL)
			Error("No edit number in position entry");

		if (temp != editNumber)
			Error("Non-decreasing position table entry");
	}

	cp = GetNumber(cp, &editPos);

	if (cp == NULL)
		Error("No edit position");

	cp = GetLine(fp, editPos, T_BEGIN_EDIT);

	if (cp == NULL)
		Error("Not at start of first edit");

	cp = GetNumber(cp, &temp);

	if (temp != editNumber)
		Error("Not at beginning of correct edit");

	if (infoLine == NULL)
		return;

	cp = GetLine(fp, NO_SEEK, T_REMARK);

	if (cp == NULL)
		Error("Cannot read first remark line of edit %ld", num);

	strncpy(infoLine, cp, MAX_INFO);

	infoLine[MAX_INFO-1] = '\0';
	infoLine[MAX_INFO-2] = '\0';
}


/*
 * Open the source file and verify that it has the correct version number
 * stored in the first line of it.  Further reads from the file will then
 * read in the real lines of the source file.
 */
void
OpenSourceFile()
{
	FIO *		fp;		/* file handle for history file */
	const char *	cp;		/* first line of file */
	long		tempEdit;	/* edit number read from first line */

	fp = sc.sourceFile;

	if (!FioOpen(fp, sc.sourceName, FIO_READ))
		FileError(fp, "Cannot open source file \"%s\"", sc.sourceName);

	cp = GetLine(fp, NO_SEEK, T_BEGIN_EDIT);

	if (cp == NULL)
		Error("Bad first line in source file \"%s\"", sc.sourceName);

	if ((GetNumber(cp, &tempEdit) == NULL))
	{
		Error("No number in first line of source file \"%s\"",
			sc.sourceName);
	}

	if (tempEdit != sc.lastEdit)
	{
		Error("First line has %ld instead of %ld in file \"%s\"",
			tempEdit, sc.lastEdit, sc.sourceName);
	}
}


/*
 * Rename the temporary file with the EXT_NEW extension to the file with
 * the specified extension, deleting the destination file.  This is used
 * for atomically changing the history file (with extension EXT_HISTORY)
 * or the latest source file (with extension EXT_SOURCE).  This uses the
 * temporary file name with the extension EXT_OLD so that at all times
 * either the old or the new file is in existence.  Returns true if
 * the rename is successful.  Breaks should be off during this call.
 */
BOOL
RenameFiles(const char * extension)
{
	char	destName[PATH_MAX];	/* name of destination file */
	char	newName[PATH_MAX];	/* name of current new file */
	char	oldName[PATH_MAX];	/* name of old file */

	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Renaming files and deleting old version source]\n");

	sprintf(destName, "%s%s", sc.baseName, extension);
	sprintf(newName, "%s%s", sc.baseName, EXT_NEW);
	sprintf(oldName, "%s%s", sc.baseName, EXT_OLD);

	(void) unlink(oldName);

	if (rename(destName, oldName) < 0)
		return FALSE;

	if (rename(newName, destName) < 0)
	{
		(void) rename(oldName, destName);

		return FALSE;
	}

	(void) unlink(oldName);

	return TRUE;
}


/*
 * Create any needed subdirectories in the specified base directory to
 * contain the specified module name.  All of the directories in the
 * base path must exist.  The final component of the module path is
 * not created as a directory.
 */
void
CreateSubDirectories(const char * baseDir, const char * modulePath)
{
	const char *	cp;
	int		baseLength;
	int		subdirLength;
	char		fullPath[PATH_MAX];

	if (strlen(baseDir) + strlen(modulePath) + 2 > PATH_MAX)
		Error("Path name is too long");

	if (access(baseDir, F_OK) < 0)
		FileError(FIO_NULL, "Cannot access \"%s\"", baseDir);

	baseLength = strlen(baseDir);

	/*
	 * Look for the next directory along the module path.
	 */
	cp = modulePath;

	while ((cp = strchr(cp, '/')) != NULL)
	{
		subdirLength = cp - modulePath;

		memcpy(fullPath, baseDir, baseLength);
		fullPath[baseLength] = '/';

		memcpy(&fullPath[baseLength + 1], modulePath, subdirLength);

		fullPath[baseLength + subdirLength + 1] = '\0';

		/*
		 * If this path is not a directory then try to create it.
		 */
		if (!IsDir(fullPath))
		{
			if (mkdir(fullPath, 0777) < 0)
			{
				FileError(FIO_NULL,
					"Cannot create directory \"%s\"",
					fullPath);
			}
		}

		cp++;
	}
}


/*
 * Check whether a file name is allowed to be written.
 * If the file already exists, the user may be prompted for his opinion.
 * Returns true if writing the file is ok.
 */
BOOL
IsWriteAllowed(const char * name)
{
	/*
	 * If forced writing is specified, then return success.
	 */
	if (sc.forceWriteFlag || (name == NULL))
		return TRUE;

	/*
	 * Check if the file exists.  If not, then we can write it.
	 */
	if (access(name, F_OK) < 0)
	{
		if (errno == ENOENT)
			return TRUE;

		fprintf(stderr, "Cannot access \"%s\": %s\n",
			name, strerror(errno));

		return FALSE;
	}

	/*
	 * The file exists.  If overwriting is prohibited, then fail.
	 */
	if (sc.noWriteFlag)
		return FALSE;

	/*
	 * Ask the user whether overwriting is ok.
	 */
	printf("Output file \"%s\" already exists.  Ok to overwrite it? ",
		name);

	fflush(stdout);

	return Confirm();
}


/*
 * Read a yes-no response from the user after a question has been asked.
 * Returns whether yes was indicated.  A null response returns false.
 * The answer is read from the standard input.
 */
BOOL
Confirm()
{
	BOOL	haveAnswer;	/* whether we have the answer */
	BOOL	answer;		/* answer */
	int	ch;		/* current input character */

	haveAnswer = FALSE;
	answer = FALSE;

	while (TRUE)
	{
		ch = getchar();

		if ((ch == EOF) || (ch == '\n'))
			break;

		if (haveAnswer || (ch == ' '))
			continue;

		answer = ((ch == 'y') || (ch == 'Y'));
	}

	return answer;
}


/*
 * Routine to read a single line from the edit file, and verify that
 * its type is as desired.  If type is negative, then it is not checked
 * and the whole line is returned.  If type is nonnegative, then it is
 * checked and skipped over, and NULL is returned if it is wrong.
 * Length of the line is returned in the global variable linelen.
 * If type is text, then the line is returned fully allocated.
 */
const char *
GetLine(
	FIO *	fp,		/* file to read line from */
	long	seekPos,	/* position to seek to or NO_SEEK */
	int	type		/* type of line expected, or -1 */
)
{
	char *	cp;
	long	pos;

	if ((seekPos != NO_SEEK) && (!FioSeek(fp, seekPos)))
		return NULL;

	pos = FioPos(fp);

	cp = FioReadline(fp, &sc.lineLen, (type == T_TEXT));

	if (cp == NULL)
		return NULL;

	/*
	 * If the line is empty, then pad it out.
	 * This is allowable because FioReadline guarantees that at
	 * least four bytes is allocated for the returned string.
	 */
	if (sc.lineLen == 2)
	{
		sc.lineLen = 3;

		cp[1] = ' ';
		cp[2] = '\n';
		cp[3] = '\0';
	}

	if (fc.debugFlag)
		printf("%5ld: %s", pos, cp);

	/*
	 * If no type was specified, then return the line as is.
	 */
	if (type < 0)
		return cp;

	/*
	 * We want the type field check and then removed.
	 * Skip over it and return the remaining text on the line.
	 */
	sc.lineLen -= TEXT_DATA_OFFSET;

	if ((sc.lineLen <= 0) || (type != cp[0]) || (cp[1] != ' '))
		return NULL;

	return cp + TEXT_DATA_OFFSET;
}


/*
 * Routine to free a text line which had been allocated and returned by
 * getline.
 */
void
FreeLine(const char * line)
{
	if (line == NULL)
		return;

	/*
	 * Back up to the start of the line that had been allocated.
	 */
	line -= TEXT_DATA_OFFSET;

	free(line);
}


/*
 * Routine to read a number from a string preceeded by optional spaces and
 * followed by a space or an end of line.  If the scanning is successful,
 * it is indirectly returned and the next position in the string is returned.
 * If the number was bad or if no string is given, NULL is returned.
 */
const char *
GetNumber(
	const char *	cp,		/* string to scan */
	long *		value		/* returned value */
)
{
	BOOL	nonum;		/* TRUE if no number detected */
	BOOL	isNeg;		/* TRUE if number is negative */

	if (cp == NULL)
		return NULL;

	nonum = TRUE;
	isNeg = FALSE;

	*value = 0;

	while (*cp == ' ')
		cp++;

	isNeg = (*cp == '-');

	if (isNeg)
		cp++;

	while ((*cp >= '0') && (*cp <= '9'))
	{
		*value = (*value * 10) + *cp++ - '0';

		if (*value < 0)
			return NULL;

		nonum = FALSE;
	}

	if (isNeg)
		*value = -*value;

	if (nonum || (*cp && (*cp != ' ') && (*cp != '\n')))
		cp = NULL;

	return cp;
}


/*
 * Check to see if the specified path is a directory.
 * Returns TRUE if so, FALSE if not or non-existent.
 */
BOOL
IsDir(const char * pathName)
{
	struct stat	buf;

	if (stat(pathName, &buf) < 0)
		return FALSE;

	return ((buf.st_mode & S_IFMT) == S_IFDIR);
}


/*
 * Routine to disable signals during a critical section of code.
 * The signals are disabled until the critical section is done.
 */
void
DisableSignals()
{
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
}


/*
 * Routine to reenable signals when critical code is complete.
 */
void
EnableSignals()
{
	signal(SIGHUP, SIG_DFL);
	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
}

/* END CODE */
