/*
 * Copyright (c) 2000 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Routine to append the changes for a new edit to the history file.
 */

#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include "fio.h"
#include "sc.h"
#include "fcomp.h"
#include "cmalloc.h"


static	char *	GetRemarks(const char *, const char *, long);
static	void	AddHistory(FIO *, const char *, long);
static	void	HelpRemarks(void);
static	void	WriteSource(const char *);
static	void	WriteEdits(FIO *, long);
static	void	CompareFiles(const char *, const char *, long);
static	void	SwapFiles(void);
static	BOOL	DoCopy(FIO *, FIO *);


/*
 * Create a new module.
 * This creates the new history file, and the initial source file.
 * This will start as edit 1, with the given file as the original source.
 * If the module contains path components, then any missing directories
 * in that path are automatically created.
 */
void
CreateHistory(
	const char *	inputName,	/* name of file to input */
	const char *	editName	/* name for edit (NULL if none) */
)
{
	FIO *	fp;		/* history file */
	FIO *	ip;		/* input filename */
	FIO *	up;		/* updated source file */
	long	posPos;
	long	editPos;
	char *	remarks;	/* remarks for new file */
	FIO	inputFile;

	/*
	 * Make sure the input name is not a directory.
	 */
	if (IsDir(inputName))
		Error("Filename \"%s\" is a directory", inputName);

	/*
	 * Verify that the module doesn't already exist, and
	 * open up the initial file for reading.
	 */
	fp = sc.historyFile;

	if (!FioOpen(fp, sc.historyName, FIO_READ) == 0)
	{
		FioClose(fp);

		Error("Module \"%s\" already exists as \"%s\"",
			sc.moduleName, sc.baseName);
	}

	if (fp->errorCode != ENOENT)
	{
		FileError(fp, "History file \"%s\" cannot be opened",
			sc.historyName);
	}

	up = sc.sourceFile;

	if (!FioOpen(up, sc.sourceName, FIO_READ) == 0)
	{
		FioClose(up);

		Error("Source for module \"%s\" already exists as \"%s\"",
			sc.moduleName, sc.baseName);
	}

	if (up->errorCode != ENOENT)
	{
		FileError(up, "Source file \"%s\" cannot be opened",
			sc.sourceName);
	}

	ip = &inputFile;

	if (!FioOpen(ip, inputName, FIO_READ))
	{
		FileError(ip, "Cannot open initial source file \"%s\"",
			inputName);
	}

	/*
	 * Create any subdirectories that are required.
	 */
	CreateSubDirectories(sc.dirName, sc.moduleName);

	/*
	 * Try to create the history source file, and if successful, then
	 * get the initial remarks for the module.
	 */
	if (fc.verbosity)
	{
		printf("Creating new module \"%s\" from initial file \"%s\"\n",
			sc.moduleName, inputName);
	}

	if (!FioOpen(up, sc.sourceName, FIO_SEQWRITE))
	{
		FileError(up, "Cannot create source file \"%s\"",
			sc.sourceName);
	}

	remarks = GetRemarks(sc.moduleName, sc.remarkName, 1L);

	/*
	 * Copy the original file into the history latest source file,
	 * with the extra line at the front identifying it as edit 1.
	 */
	DisableSignals();

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf("[Copying \"%s\" to \"%s\"]\n", inputName,
			sc.sourceName);
	}

	if (!FioFormat(up, "%c 1\n", T_BEGIN_EDIT))
	{
		FileError(up, "Cannot write header line for source file \"%s\"",
			sc.sourceName);
	}

	if (!DoCopy(ip, up))
	{
		FileError(up, "Error during copy of \"%s\" to \"%s\"",
			inputName, sc.sourceName);
	}

	if (!FioClose(up))
		FileError(up, "Error closing \"%s\"", sc.sourceName);

	FioClose(ip);

	/*
	 * Write the initial history file data
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Creating history file \"%s\"]\n", sc.historyName);

	if (!FioOpen(fp, sc.historyName, FIO_WRITE))
	{
		FileError(fp, "Cannot create history file \"%s\"",
			sc.historyName);
	}

	FioFormat(fp, HEADER_FORMAT, T_HEADER, 1L, 1L, 0L, 0L);

	editPos = FioPos(fp);

	AddHistory(fp, remarks, 1L);

	posPos = FioPos(fp);

	FioFormat(fp, (editName ? "%c 1 %ld %s\n" : "%c 1 %ld\n"),
		T_POSITION, editPos, editName);

	FioFormat(fp, "%c %ld\n", T_EOF, FioPos(fp));

	FioSeek(fp, 0L);

	FioFormat(fp, HEADER_FORMAT, T_HEADER, 1L, 1L, 0L, posPos);

	if (!FioClose(fp))
	{
		FileError(fp, "Cannot write initial history file \"%s\"",
			sc.historyName);
	}

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf("[Creation complete]\n");
		fflush(stdout);
	}

	EnableSignals();
}


/*
 * Update the history file by appending the changes given by a file.
 */
void
UpdateHistory(
	const char *	inputName,	/* name of file to input */
	const char *	editName	/* name of edit (NULL if none) */
)
{
	FIO *		fp;		/* history file */
	const char *	cp;		/* current line of data */
	POS *		pp;		/* current entry in position table */
	POS *		posTable;	/* position table */
	char *		remarks;	/* remarks */
	long		newTablePos;	/* position of new position table */
	long		lastEditPos;	/* position of beginning of last edit */
	long		appendPos;	/* position to append new changes to */
	long		tempEdit;	/* edit number just scanned from line */
	long		newEdit;	/* number of new edit */
	long		edit;		/* current edit number */
	int		saveError;	/* saved error code on errors */

	/*
	 * Make sure that the input file contains the right version number
	 * before beginning the update, and if so, then get the remarks
	 * for the edit.
	 */
	fp = sc.historyFile;

	OpenHistoryFile(FIO_UPDATE);

	if (!sc.forceUpdateFlag)
	{
		tempEdit = FindEdit(inputName);

		if ((tempEdit > 0) && (tempEdit != sc.lastEdit))
		{
			Error(
	"Input file \"%s\" contains edit %ld instead of %ld for module \"%s\"",
			inputName, tempEdit, sc.lastEdit, sc.moduleName);
		}
	}

	newEdit = sc.lastEdit + 1;

	if (fc.verbosity)
	{
		printf("Installing \"%s\" as edit %ld of module \"%s\"\n",
			inputName, newEdit, sc.moduleName);
	}

	remarks = GetRemarks(sc.moduleName, sc.remarkName, newEdit);

	/*
	 * See if we are doing a straight update, or an update after a
	 * display of the differences.  If the latter, then we already have
	 * done the comparison, but we need to swap the differences.
	 */
	if (fc.snakeList)
		SwapFiles();
	else
		CompareFiles(inputName, sc.sourceName, 1L);

	posTable = ReadPosTable(fp);

	/*
	 * Now go back and verify that the last edit actually exists where it
	 * is supposed to be.  We will overwrite it with the changes.
	 * Skip over all remark lines to position to the place to append
	 * the new changes.  This will be at the location of the file line
	 * for the latest edit.
	 */
	lastEditPos = posTable[1].p_pos;

	cp = GetLine(fp, lastEditPos, T_BEGIN_EDIT);

	if (cp == NULL)
		Error("Cannot read beginning line of latest edit");

	cp = GetNumber(cp, &tempEdit);

	if ((cp == NULL) || (tempEdit != sc.lastEdit))
		Error("Bad begin edit line for latest edit");

	while (TRUE)
	{
		appendPos = FioPos(fp);

		cp = FioReadline(fp, (long *) NULL, FALSE);

		if (cp == NULL)
			Error("Error reading remark lines for latest edit");

		if (*cp == T_FILE)	/* the file line, all done */
			break;

		if (*cp != T_REMARK)	/* error if not remark line */
			Error("Unexpected line type in latest edit");
	}

	/*
	 * Now write the new version of the source file under a temporary name.
	 * This is to prevent problems due to quota errors.  So if this fails
	 * we have done nothing bad.
	 */
	DisableSignals();

	WriteSource(inputName);

	/*
	 * Now position back to the file line in the latest edit, and
	 * overwrite it with the changes for the new edit.  Errors will
	 * be checked for later and handled there, so we can recover from
	 * disk full or quota exceeded problems.
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Appending edits to history file]\n");

	if (!FioSeek(fp, appendPos))
		Error("Cannot seek to location to append new edit");

	WriteEdits(fp, sc.lastEdit);

	/*
	 * Now append the latest edit template to the file, and follow that
	 * with the saved position table, and then append a new end of file
	 * line.  Save the position of the beginning of the table for
	 * updating the header.  Errors are checked for later.
	 */
	pp = posTable;
	pp->p_pos = FioPos(fp);
	pp->p_names = editName;

	AddHistory(fp, remarks, newEdit);

	newTablePos = FioPos(fp);

	for (edit = newEdit; edit >= sc.firstEdit; edit--, pp++)
	{
		FioFormat(fp,
			(pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
			T_POSITION, edit, pp->p_pos, pp->p_names);
	}

	FioFormat(fp, "%c %ld\n", T_EOF, FioPos(fp));

	/*
	 * Force the last bit of the output to the file, then check for errors.
	 * If an error occurred during any of the above writes then we will try
	 * to recover by rewriting the old position table back to where it was.
	 * This should succeed since quota exceeded or disk full errors cannot
	 * occur since the position table was there previously!
	 */
	FioFlush(fp);

	if (FioError(fp) || !RenameFiles(EXT_SOURCE))
	{
		saveError = fp->errorCode;
		fp->errorCode = 0;

		if (!FioSeek(fp, appendPos))
		{
			FileError(fp,
		"Seek to last edit failed - history file \"%s\" damaged",
				sc.historyName);
		}

		FioFormat(fp, "%c 0\n", T_FILE);
		FioFormat(fp, "%c %ld 0 0\n", T_END_EDIT, sc.lastEdit);

		if (FioPos(fp) != sc.tablePos)
		{
			Error(
		"Position table moved - history file \"%s\" damaged",
				sc.historyName);
		}

		pp = &posTable[1];

		for (edit = sc.lastEdit; edit >= sc.firstEdit; edit--, pp++)
		{
			FioFormat(fp, (pp->p_names ?
				"%c %ld %ld %s\n" : "%c %ld %ld\n"),
				T_POSITION, edit, pp->p_pos, pp->p_names);
		}

		FioFormat(fp, "%c %ld\n", T_EOF, FioPos(fp));

		FioFlush(fp);

		if (FioError(fp))
		{
			FileError(fp,
	"Rewrite of position table failed - history file \"%s\" damaged",
				sc.historyName);
		}

		FioClose(fp);

		fp->errorCode = saveError;

		FileError(fp,
	"Update of history file \"%s\" failed (file still correct)",
			sc.historyName);
	}

	/*
	 * As the last step for inserting the new edit, rewrite the old header
	 * line to update the new last edit number and the new position of the
	 * beginning of the position table.  This line must be of fixed length.
	 * This write should succeed since the file is not being grown in size.
	 */
	if (!FioSeek(fp, 0L))
	{
		FileError(fp,
	"Seek to header line failed - history file \"%s\" damaged",
			sc.historyName);
	}

	FioFormat(fp, HEADER_FORMAT, T_HEADER, sc.firstEdit, newEdit, 0L,
		newTablePos);

	if (FioError(fp) || !FioFlush(fp))
	{
		FileError(fp,
	"Write of header line failed - history file \"%s\" damaged",
			sc.historyName);
	}

	FioClose(fp);

	if (fc.verbosity)
	{
		printf(
"Edit %ld of module \"%s\" saved (inserts %ld, deletes %ld, matches %ld)\n",
			newEdit, sc.moduleName, fc.deletes,
			fc.inserts, fc.matches);
	}

	fflush(stdout);

	EnableSignals();
}


/*
 * Add the lines in the history file for the specified new edit number.
 * This are the beginning and end lines, the file line, and the default
 * and user-supplied remark lines.  These lines are written to the current
 * position in the file.  Errors are remembered by the file structure.
 */
static void
AddHistory(
	FIO *		fp,		/* file to write to */
	const char *	remarks,	/* remarks to add to file */
	long		editNumber	/* edit number */
)
{
	const struct passwd *	pwd;
	const char *		userName;
	time_t			binTime;
	long			len;

	time(&binTime);

	pwd = getpwuid(getuid());

	if (pwd)
		userName = pwd->pw_name;
	else
		userName = "UNKNOWN";

	FioFormat(fp, "%c %ld\n", T_BEGIN_EDIT, editNumber);
	FioFormat(fp, "%c %s   %s", T_REMARK, userName, ctime(&binTime));

	while (*remarks)
	{
		len = strlen(remarks);

		FioFormat(fp, "%c ", T_REMARK);
		FioWrite(fp, remarks, len);

		remarks += (len + 1);
	}

	FioFormat(fp, "%c 0\n", T_FILE);
	FioFormat(fp, "%c %ld 0 0\n", T_END_EDIT, editNumber);
}


/*
 * Read remark lines from the user to describe the reason why an edit
 * was made.  This is made up of lines separated by nulls, with a double
 * null at the end.  The remarks can be from the terminal or from a file.
 */
static char *
GetRemarks(
	const char *	moduleName,	/* module name this is for */
	const char *	remarkName,	/* remarks file name (NULL if tty) */
	long		editNumber	/* edit number */
)
{
	char *		buf;		/* buffer for input */
	FIO *		fp;		/* remark file */
	char *		cp;		/* current line for retyping */
	long		avail;		/* total characters available */
	long		used;		/* number of characters used */
	long		line;
	int		ch;
	FIO		remarkFile;	/* remark file structure */

	if (!remarkName)		/* no remarks at all */
		return "";

	used = 0;
	avail = 400;

	buf = malloc(avail + 4);

	if (buf == NULL)
		Error("Cannot allocate memory for remarks text");

	/*
	 * If remark file name was supplied, read remarks from it.
	 */
	if (*remarkName)
	{
		fp = &remarkFile;

		if (!FioOpen(fp, remarkName, 0))
		{
			FileError(fp, "Cannot open remark file \"%s\"",
				remarkName);
		}

		while ((cp = FioReadline(fp, &line, FALSE)) != 0)
		{
			if ((used + line) >= avail)
			{
				avail = used + line + 200;

				buf = realloc(buf, avail + 4);

				if (buf == NULL)
				{
					Error(
				"Cannot reallocate memory for remarks text");
				}
			}

			strcpy(&buf[used], cp);
			used += (line + 1);
		}

		if (!FioClose(fp))
		{
			FileError(fp, "Error reading remark file \"%s\"",
				remarkName);
		}

		buf[used] = '\0';

		return buf;
	}

	/*
	 * No remark file specified, so get remarks from user.
	 */
	printf(
"\nPlease type remarks for edit %ld of \"%s\" (type \"!h\" for help)\n",
		editNumber, moduleName);

	line = 1;

	while (TRUE)
	{
		if ((used + 200) > avail)
		{
			avail += 200;

			buf = realloc(buf, avail + 4);

			if (buf == NULL)
			{
				Error(
				"Cannot reallocate memory for remark text");
			}
		}

		printf("%ld> ", line);
		fflush(stdout);

		if (fgets(&buf[used], 200, stdin) == NULL)
			break;

		ch = buf[used];

		if (ch == '\n')
			break;

		if (ch != '!')
		{
			/*
			 * Normal line
			 */
			used += strlen(&buf[used]) + 1;
			line++;

			continue;
		}

		buf[used] = '\0';

		ch = buf[used+1];		/* control line */

		if ((ch >= 'A') && (ch <= 'Z'))
			ch += ('a' - 'A');

		switch (ch)
		{
			case 'b':		/* back up to previous line */
				if (line <= 1)
					break;

				line--;

				cp = buf;

				for (ch = 1; ch < line; ch++)
					cp += (strlen(cp) + 1);

				used = (cp - buf);
				break;

			case 'c':		/* restart the remark */
				printf(
			"\nRemarks cleared, please type new remarks:\n");

				used = 0;
				line = 1;
				break;

			case 'h':		/* help */
				HelpRemarks();
				break;

			case 'q':		/* quit the command */
				Restart(TRUE);

			case 'r':		/* retype remarks */
				printf("\n");

				for (ch = 0, cp = buf; *cp;
					cp += (strlen(cp) + 1))
				{
					printf("%d> %s", ++ch, cp);
				}

				break;

			default:		/* unknown command */
				if (ch < ' ')
					ch = ' ';

				printf("Unknown command \"!%c\" - ignored\n",
					ch);
		}
	}

	buf[used] = '\0';		/* double null to end string */

	return buf;
}


/*
 * Describe help about entering remarks.
 */
static void
HelpRemarks()
{
printf("\nRemarks are lines of text which describe why an edit was made.\n");
printf("The lines of remarks are terminated by typing a single blank line.\n");
printf("Lines beginning with an exclamation mark are commands, as follows:\n");
printf("\n");
printf("!b   Backs up to the previous line so you can retype it.\n");
printf("!c   Clears all remarks entered so far so you can retype them.\n");
printf("!r   Redisplays all remarks entered so far.\n");
printf("!q   Aborts updating of the current module.\n\n");
}


/*
 * Write the new source file into a temporary file in the proper directory.
 * This is essentially a copy, except for the initial line indicating the
 * edit number of the new source.  This copy is done to guarantee that there
 * is enough disk space in order to finish the update.  If there is not, then
 * nothing harmful will have been done to the old source or history files.
 * It is assumed that the history file has already been opened.
 */
static void
WriteSource(const char * inputName)
{
	FIO *	ip;			/* input file */
	FIO *	up;			/* output file */
	FIO	inputFile;
	FIO	updateFile;
	char	updateName[PATH_MAX];	/* full name of new source file */

	strcpy(updateName, sc.baseName);
	strcat(updateName, EXT_NEW);

	ip = &inputFile;
	up = &updateFile;

	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		printf("[Copying \"%s\" to temporary file \"%s\"]\n",
			inputName, updateName);
	}

	if (!FioOpen(ip, inputName, FIO_READ))
		FileError(ip, "Cannot open new source file \"%s\"", inputName);

	if (!FioOpen(up, updateName, FIO_SEQWRITE))
	{
		FileError(up, "Cannot create temporary source file \"%s\"",
			updateName);
	}

	if (!FioFormat(up, "%c %ld\n", T_BEGIN_EDIT, sc.lastEdit + 1))
	{
		FileError(up,
	"Cannot write header line for temporary source file \"%s\"",
			updateName);
	}

	if (!DoCopy(ip, up))
	{
		FileError(up, "Error during copy of \"%s\" to \"%s\"",
			inputName, updateName);
	}

	if (!FioClose(up))
		FileError(up, "Error closing \"%s\"", updateName);

	if (!FioClose(ip))
		FileError(ip, "Error closing \"%s\"", inputName);
}


/*
 * Do the copy of the source file to the destination file, while
 * modifying the lines as necessary.  Returns true if successful.
 */
static BOOL
DoCopy(
	FIO *	ip,		/* input file */
	FIO *	op		/* output file */
)
{
	const char *	cp;		/* current line */
	long		lineLen;	/* current line length */
	long		i;		/* line count */

	for (i = sc.modifyLines; i > 0; i--)
	{
		cp = FioReadline(ip, &lineLen, FALSE);

		if (cp == NULL)
			return !FioError(ip);

		cp = ModifyLine(cp, &lineLen, (INFO *) NULL);

		if (!FioWrite(op, cp, lineLen))
			return FALSE;
	}

	return FioCopy(ip, op, -1L);
}


/*
 * Compare two files, and return the changes into a snake list.
 * This works by calling the fcomp module which is loaded with us.
 * Does not return if an error ocurred.
 */
static void
CompareFiles(
	const char *	file1,		/* first file to compare */
	const char *	file2,		/* second file to compare */
	long		skip2		/* lines to skip for second file */
)
{
	if (fc.verbosity > VERBOSE_DEFAULT)
		printf("[Comparing files \"%s\" and \"%s\"]\n", file1, file2);

	fc.fileBskip = skip2;
	fc.fileAmodify = sc.modifyLines;
	fc.fileBmodify = sc.modifyLines;
	fc.maxChanges = INFINITY;

	Fcomp(file1, file2);
}


/*
 * Swap the snake line numbers for two files to appear as if the comparison
 * was in the other order.  This is used when showing differences and then
 * updating, since showing differences needs the changes to the new version,
 * but updating needs the changes back to the older version.
 */
static void
SwapFiles()
{
	SNAKE *		sp;		/* current snake */
	long		temp;		/* temporary for swapping */
	FILEINFO	tempFile;	/* temporary for swapping */

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		temp = sp->line1;
		sp->line1 = sp->line2;
		sp->line2 = temp;
	}

	tempFile = fc.fileA;
	fc.fileA = fc.fileB;
	fc.fileB = tempFile;

	temp = fc.deletes;
	fc.deletes = fc.inserts;
	fc.inserts = temp;
}


/*
 * Write out the edit script corresponding to the results of the
 * comparison between two files.  Errors cause premature finishing,
 * and are accessible by using FioError afterwards.  This routine knows
 * the data structures defined by the fcomp module.  This routine is
 * also called after the begin line and remark lines have been
 * skipped over.
 */
static void
WriteEdits(
	FIO *	fp,		/* file to write to */
	long	editNumber	/* edit number to write */
)
{
	SNAKE *		sp;		/* current snake */
	const char *	cp;		/* line being inserted */
	long		line1;		/* current line numbers in files */
	long		line2;
	long		inserts;	/* inserts and deletes needed */
	long		deletes;
	long		totalInserts;	/* total lines inserted or deleted */
	long		totalDeletes;
	long		i;
	int		len;
	BOOL		eolMissing;

	totalInserts = 0;
	totalDeletes = 0;

	line1 = 0;
	line2 = 0;

	for (sp = fc.snakeList; sp; sp = sp->next)
	{
		deletes = sp->line1 - line1;
		inserts = sp->line2 - line2;

		eolMissing = FALSE;

		if (inserts > 0)
		{
			cp = fc.fileB.lines[line2 + inserts - 1]->data;
			len = strlen(cp);

			if ((len > 0) && (cp[len - 1] != '\n'))
				eolMissing = TRUE;
		}

		if (fc.debugFlag)
		{
			printf(
		"[line1 %ld  line2 %ld  count %ld  insert %ld  delete %ld]\n",
			sp->line1, sp->line2, sp->count, inserts, deletes);
		}

		if (deletes || inserts)
		{
			if (eolMissing)
			{
				FioFormat(fp, "%c %ld %ld %ld %c\n",
					T_CHANGE, line1 + 1, inserts,
					deletes, EOL_MISSING_FLAG);
			}
			else
			{
				FioFormat(fp, "%c %ld %ld %ld\n", T_CHANGE,
					line1 + 1, inserts, deletes);
			}

			if (FioError(fp))
				return;
		}

		for (i = 0; i < inserts; i++)
		{
			FioFormat(fp, "%c ", T_TEXT);

			cp = fc.fileB.lines[line2 + i]->data;

			FioWrite(fp, cp, (long) strlen(cp));

			if (FioError(fp))
				return;
		}

		if (eolMissing)
			FioChar(fp, '\n');

		totalInserts += inserts;
		totalDeletes += deletes;

		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}

	FioFormat(fp, "%c %ld %ld %ld\n", T_END_EDIT,
		editNumber, totalInserts, totalDeletes);
}

/* END CODE */
